# Installing Ops Director

The `ml-gradle` script in this directory will install Ops Director.

## Prerequisites

To install Ops Director, you must have:

1. Oracle Java JDK 1.8 on the machine where the Ops Director installer
   will be run.
2. A network connection with access to the internet.

## Configuration

There is a `gradle.properties` file in this directory. You can edit it
to change the Ops Director configuration.

1. You must update `mlHost` to the name of the host where you want to
   install Ops Director. It is best practice not to use localhost, but
   the actual hostname of the machine.
2. If the host your are installing onto is not in the group named
   “Default”, you must specify the `mlGroupName` property and give the
   correct group name.
3. If you are generating a new certificate authority for Ops Director
   to use, you must set the `opsdirCaCountry` and `opsdirCaOrgName`.
   For scripting purposes, or if you do not wish to edit the file, you
   can specify these on the command line with `-Pproperty=value`, for
   example
   
```
-PopsdirCaCountry=US
```

to specify US as the country for the certificate authority.

## Installation Instructions

In this directory, run the `gradlew` script with the `mlDeploy` target.
On Linux or the Mac:

```
./gradlew mlDeploy
```

On Windows:

```
.\gradlew.bat mlDeploy
```

Specify any additional `-P` options you wish before the target. For example:

```
./gradlew -PmlHost=myhost -PopsdirCaCountry=US -PopsdirCaOrgName=MarkLogic mlDeploy
```

# Uninstalling Ops Director

Recent installers also have the ability to un-install Ops Director.
This leverages the “undeploy” feature of `ml-gradle`. To un-install,
run

```
./gradlew -Pconfirm=true mlUndeploy
```

As with installation, you must specify `mlHost` and `mlGroupName` if
they differ from the defaults. (The `-Pconfirm=true` option just sets
the value of the `confirm` option to `true`. It’s just an extra step
that’s designed to help you avoid accidentally running uninstall.)

## Errors and warnings

The `ml-gradle` tool logs messages and warnings as it goes. If it
detects a situation that will prevent it from deploying (for example,
a missing configuration parameter), it will abort with an error.
