# Installing Ops Director

The `ml-gradle` script in this directory will install Ops Director on
a machine that does not currently have Ops Director installed (upgrade
is not supported yet).

## Prerequisites

To install Ops Director, you must have:

1. Oracle Java JDK 1.8.
2. A network connection with access to the internet.
3. A network connection to the machine where you want to install Ops Director.

The installer does not depend on anything else in the `develop`
branch. You can safely zip the installer directory up and move it to
another machine.

It will install most quickly if you run it on the machine where it
will install Ops Director (i.e. `localhost`), but you can do it from
any machine that has access.

### Windows prerequisites

Apparently, on Windows, you need to install `winutils.exe` for some part
of the Hadoop ecosystem (used by MLCP) to function. Google will
[find links](https://www.google.com/search?q=hadoop%20null%5Cbin%5Cwinutils)
for you. For example, [this one](http://teknosrc.com/spark-error-java-io-ioexception-could-not-locate-executable-null-bin-winutils-exe-hadoop-binaries/)

## Configuration

There is a `gradle.properties` file in this directory. You can edit it
to change the Ops Director configuration.

1. You must update `mlHost` to the name of the host where you want to
   install Ops Director.
2. If the host your are installing onto is not in the group named “Default”,
   you must specify the `mlGroupName` property and give the correct group
   name.
3. If you are generating a new certificate authority for Ops Director to use,
   you must set the `opsdirCaCountry` and `opsdirCaOrgName`.

For scripting purposes, or if you do not wish to edit the file, you can specify
these on the command line with `-Pproperty=value`, for example

```
-PopsdirCaCountry=US
```

to specify US as the country for the certificate authority.

## Installation Instructions

In this directory, run the `gradlew` script with the `mlDeploy` target.
On Linux or the Mac:

```
./gradlew mlDeploy
```

On Windows:

```
.\gradlew.bat mlDeploy
```

Specify any additional `-P` options you wish before the target. For example,
I usually run it this way:

```
./gradlew -PmlHost=builder -PopsdirCaCountry=US -PopsdirCaOrgName=MarkLogic mlDeploy
```

# Uninstalling Ops Director

Recent installers also have the ability to un-install Ops Director. This leverages
the “undeploy” feature of `ml-gradle`.

To un-install, run

```
./gradlew -Pconfirm=true mlUndeploy
```

As with installation, you must specify `mlHost` and `mlGroupName` if they differ
from the defaults. (The `-Pconfirm=true` option just sets the value of the
`confirm` option to `true`. It’s just an extra step that’s designed to help you
avoid accidentally running uninstall.)

## Errors and warnings

The `ml-gradle` tool logs messages and warnings as it goes. If it
detects a situation that will prevent it from deploying (for example,
a missing configuration parameter), it will abort with an error.



