xquery version "1.0-ml";

(: Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. :)

declare variable $error:errors as node()* external;

let $code   := $error:errors/error:code
let $status := xdmp:get-response-code()
return
  if ($status[1] lt 500)
  then
    (xdmp:set-response-content-type("application/json"),
     xdmp:set-response-code($status[1], $status[2]),
     object-node {
       "name": "EndpointError",
       "errorCode": string($status[1]),
       "message": xdmp:quote(string($error:errors/error:format-string)),
       "data": array-node { $error:errors/error:data/error:datum ! string(.) }
     })
  else
    if ($code = "OD-BADMETHOD")
    then
      (xdmp:set-response-content-type("application/json"),
       xdmp:set-response-code(405, "Method not allowed"),
       object-node {
         "name": "EndpointError",
         "errorCode": string($code),
         "message": xdmp:quote(string($error:errors/error:format-string)),
         "data": array-node { $error:errors/error:data/error:datum ! string(.) }
       })
    else if (starts-with($code, "OD-") and not($code eq 'OD-INTERR'))
    then
      (xdmp:set-response-content-type("application/json"),
       xdmp:set-response-code(400, "Bad Request"),
       object-node {
         "name": "EndpointError",
         "errorCode": string($code),
         "message": xdmp:quote(string($error:errors/error:format-string)),
         "data": array-node { $error:errors/error:data/error:datum ! string(.) }
       })
    else if ($code eq "SEC-PRIV")
    then
      (xdmp:set-response-content-type("application/json"),
       xdmp:set-response-code(403, "Forbidden"),
       object-node {
         "name": "EndpointError",
         "errorCode": string($code),
         "message": xdmp:quote(string($error:errors/error:format-string)),
         "data": array-node { $error:errors/error:data/error:datum ! string(.) }
       })
    else if ($code eq "XDMP-MODNOTFOUND")
    then
      let $uri := xdmp:get-original-url()
      let $_   := xdmp:set-response-content-type("application/json")
      let $_   := xdmp:set-response-code(404, "Not Found")
      return
       object-node {
         "name": "EndpointError",
         "errorCode": string($code),
         "message": "Not found: " || $uri,
         "data": array-node { $uri }
       }
    else
      (xdmp:set-response-content-type("application/json"),
       xdmp:set-response-code(500, "Internal Server Error"),
       object-node {
         "name": "EndpointError",
         "errorCode": string($status[1]),
         "message": xdmp:quote(string($error:errors/error:format-string)),
         "data": array-node { $error:errors/error:data/error:datum ! string(.) }
       })

