"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

function toXsDateTime(str) {
  if (str && str.split) {
    let [day, time] = str.split('T');
    return fn.dateTime(day, time);
  } else {
    return null;
  }
}

function parseParam(paramFieldDef, fieldDefs, paramDefs, rb) {
  let param = rb[paramFieldDef.name];
  if (param) {
    let fieldName = paramFieldDef.field || paramFieldDef.name;
    // this is crappy.
    let fieldDef = fieldDefs.find(field => field.name === fieldName);
    switch (paramFieldDef.type) {
    case 'jsonPropertyValueQuery': {
      let val = parseValue(fieldDef, param);
      return val && val.length ? [cts.jsonPropertyValueQuery(fieldName, val)] : [];
    }
    case 'jsonPropertyRangeQuery': {
      let val = parseValue(fieldDef, param);
      return val ? [cts.jsonPropertyRangeQuery(fieldName, paramFieldDef.dir, val)] : [];
    }
    default:
      throw new Error(`Unsupported parameter type, "${paramFieldDef.type}"`);
    }
  } else {
    return [];
  }
}

exports.parseParam = parseParam;
function parseValue(field, val) {
  switch (field.type) {
  case 'date':
    return toXsDateTime(val);
  case 'number':
    return val;
  default:
    return Array.isArray(val) ? val : val ? [val] : [];
  }
}

exports.parseValue = parseValue;
function parseParams(fieldDefs, paramDefs, rb, misc = []) {
  if (rb.direction) {
    if (rb.direction === 'desc' || rb.direction == 'descending') {
      rb.direction = "descending";
    } else {
      rb.direction = "ascending";
    }
  } else {
    rb.direction = paramDefs.defaultSortOrder;
  }

  let criteria = [];
  if (rb.text && rb.text.trim().length) {
    let text = [rb.text.trim() + "*"];
    let opts = ['case-insensitive', 'punctuation-insensitive'];
    if (paramDefs.fieldName) {
      criteria.push(cts.fieldWordQuery(paramDefs.fieldName, text, opts));
    } else if (paramDefs.wordField) {
      if (Array.isArray(paramDefs.wordField)) {
        criteria.push(cts.jsonPropertyWordQuery(paramDefs.wordField, text, opts));
      } else {
        if (paramDefs.wordField === '*') {
          criteria.push(cts.wordQuery(text, opts));
        } else {
          criteria.push(cts.jsonPropertyWordQuery(paramDefs.wordField, text, opts));
        }
      }
    }
  }
  if (paramDefs.fields) {
    let fieldsCriteria = paramDefs.fields.reduce((acc, paramFieldDef) => {
      acc = [...acc, ...parseParam(paramFieldDef, fieldDefs, paramDefs, rb)];
      return acc;
    }, []);
    criteria = [...criteria, ...fieldsCriteria];
  }

  if (rb.exclude && rb.exclude.length > 0) {
    const notQuery = cts.notQuery(cts.jsonPropertyValueQuery('opsdirId', rb.exclude));
    criteria = [...criteria, notQuery];
  }

  let slice = [];
  slice.push(parseInt(rb.start || 0));
  // this is a hard limit on pageLength, UI code must always send length < 1000
  // there is a bug which surfaces when UI send 9999 then estimate of the result is wrong but the length
  // of items list will be limited upto 50. ex: /v1/resource-group:list.
  // need to revist post v1
  if (rb.length && rb.length < 1000) {
    slice.push(slice[0] + parseInt(rb.length));
  } else {
    slice.push(slice[0] + 50);
  }

  let miscItems = {};
  for (let miscItem of misc) {
    miscItems[miscItem] = rb[miscItem];
  }
  return {
    criteria: criteria,
    sortField: rb.sortField || paramDefs.defaultSortField,
          direction: rb.direction,
          slice: slice,
          misc: miscItems
  };
}

exports.parseParams = parseParams;
