"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

var jsearch = require('/MarkLogic/jsearch.sjs');
var opsdir = require('/common/opsdir.xqy');
exports.rColl = 'http://marklogic.com/opsdirector/v1/resource';
exports.RESOURCE_TYPE = {
    resourceList: 'resource-list',
    clusters: 'clusters',
    databases: 'databases',
    hosts: 'hosts',
    forests: 'forests',
    groups: 'groups',
    servers: 'servers',
    databaseStatus: 'database-status',
    forestStatus: 'forest-status',
    hostStatus: 'host-status',
    serverStatus: 'server-status',
    tasks: 'tasks',
    counts: 'counts',
    privileges: 'privileges',
    roles: 'roles',
};
exports.FIELDS = {
    opsdirId: 'opsdirId',
    resourceType: 'opsdirResourceType',
    clusterId: 'opsdirClusterId',
    hostId: 'opsdirHostId',
    resourceName: 'opsdirName',
    ts: 'ts',
    group: 'group',
    groupName: 'groupName',
    forest: 'forest',
    forestName: 'forest',
};
exports.DOCUMENT_KIND = {
    resourceList: 'resource-list',
    resource: 'resource',
};
exports.OUT = {
    clusterId: 'clusterId',
    clusterName: 'clusterName',
    groups: 'groups',
    groupId: 'groupId',
    groupName: 'groupName',
    hosts: 'hosts',
    forests: 'forests',
    databases: 'datbases',
    databaseName: 'databaseName',
    servers: 'servers'
};
function and(...and) {
    return cts.andQuery(and);
}
exports.and = and;
function optionalAnd(optionalFieldName, optionalFieldVals, otherCriteria) {
    if (optionalFieldVals) {
        return and(cts.jsonPropertyValueQuery(optionalFieldName, optionalFieldVals), otherCriteria);
    }
    else {
        return otherCriteria;
    }
}
exports.optionalAnd = optionalAnd;
function clusterDoc(id) {
    return and([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.clusters]),
        cts.jsonPropertyValueQuery(exports.FIELDS.clusterId, [id])
    ]);
}
exports.clusterDoc = clusterDoc;
function hostsForCluster(id) {
    return cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.hosts]),
        cts.jsonPropertyValueQuery(exports.FIELDS.clusterId, [id])
    ]);
}
exports.hostsForCluster = hostsForCluster;
function groupsForCluster(id) {
    return cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.groups]),
        cts.jsonPropertyValueQuery(exports.FIELDS.clusterId, [id])
    ]);
}
exports.groupsForCluster = groupsForCluster;
function forestsForCluster(id) {
    return and([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.forests]),
        cts.jsonPropertyValueQuery(exports.FIELDS.clusterId, [id])
    ]);
}
exports.forestsForCluster = forestsForCluster;
function databasesForCluster(id) {
    return cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.databases]),
        cts.jsonPropertyValueQuery(exports.FIELDS.clusterId, [id])
    ]);
}
exports.databasesForCluster = databasesForCluster;
function serversForCluster(id) {
    return cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.servers]),
        cts.jsonPropertyValueQuery(exports.FIELDS.clusterId, [id])
    ]);
}
exports.serversForCluster = serversForCluster;
function clustersForHost(id) {
    return cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.clusters]),
        cts.jsonPropertyValueQuery(exports.FIELDS.hostId, [id])
    ]);
}
exports.clustersForHost = clustersForHost;
function groupNameForCluster(id) {
    return cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.groups]),
        cts.jsonPropertyValueQuery(exports.FIELDS.clusterId, [id])
    ]);
}
exports.groupNameForCluster = groupNameForCluster;
function configDocs(ids) {
    return optionalAnd(exports.FIELDS.opsdirId, ids, cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.orQuery([
            cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.groups]),
            cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.hosts]),
            cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.forests]),
            cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.databases]),
            cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.servers]),
            cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.clusters]),
        ])
    ]));
}
exports.configDocs = configDocs;
function clusterDocs(ids) {
    return optionalAnd(exports.FIELDS.opsdirId, ids, cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.clusters])
    ]));
}
exports.clusterDocs = clusterDocs;
function hostDocs(ids) {
    return optionalAnd(exports.FIELDS.opsdirId, ids, cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.hosts])
    ]));
}
exports.hostDocs = hostDocs;
function groupDocs(ids) {
    return optionalAnd(exports.FIELDS.opsdirId, ids, cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.groups])
    ]));
}
exports.groupDocs = groupDocs;
function serverDocs(ids) {
    return optionalAnd(exports.FIELDS.opsdirId, ids, cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.servers])
    ]));
}
exports.serverDocs = serverDocs;
function databaseDocs(ids) {
    return optionalAnd(exports.FIELDS.opsdirId, ids, cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.databases])
    ]));
}
exports.databaseDocs = databaseDocs;
function forestDocs(ids) {
    return optionalAnd(exports.FIELDS.opsdirId, ids, cts.andQuery([
        cts.collectionQuery([exports.rColl]),
        cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [exports.RESOURCE_TYPE.forests])
    ]));
}
exports.forestDocs = forestDocs;
function sortedTuples(sortField, tuples) {
    // arra.prototype.sort doesn't seem to work or maybe I'm nuts.
    // Anyway, this is better for our purposes.
    return tuples.reduce((acc, tup) => {
        if (tup === sortField) {
            acc.unshift(tup);
        }
        else {
            acc.push(tup);
        }
        return acc;
    }, []);
}
// function byResourceName(resourceNames: string[], otherCriteria: string) {
//   return cts.andQuery([
//     configDocs(),
//     cts.orQuery([
//       cts.jsonPropertyValueQuery( FIELDS.clusterName, resourceNames ),
//       cts.jsonPropertyValueQuery( FIELDS.groupName, resourceNames ),
//       cts.jsonPropertyValueQuery( FIELDS.forestName, resourceNames ),
//       cts.jsonPropertyValueQuery( FIELDS.databaseName, resourceNames ),
//       cts.jsonPropertyValueQuery( FIELDS.serverName, resourceNames ),
//       cts.jsonPropertyValueQuery( FIELDS.hostName, resourceNames ),
//     ])
//   ])
// }
// still some irregularities in how forests and groups are handled re: names of properties, using names vs. ids.
function clusterTuples(sortField, ids, resourceNames) {
    return jsearch.tuples(sortedTuples(sortField, [
        exports.FIELDS.opsdirId, exports.FIELDS.resourceName, exports.FIELDS.resourceType, exports.FIELDS.ts
    ]))
        .where(optionalAnd(exports.FIELDS.resourceName, resourceNames, clusterDocs(ids)));
}
exports.clusterTuples = clusterTuples;
function hostTuples(sortField, ids, resourceNames, clusterIds, groupNames) {
    return jsearch.tuples(sortedTuples(sortField, [
        exports.FIELDS.opsdirId, exports.FIELDS.resourceName, exports.FIELDS.clusterId, exports.FIELDS.group, exports.FIELDS.ts
    ])).where(optionalAnd(exports.FIELDS.resourceName, resourceNames, optionalAnd(exports.FIELDS.clusterId, clusterIds, optionalAnd(exports.FIELDS.group, groupNames, hostDocs(ids)))));
}
exports.hostTuples = hostTuples;
function databaseTuples(sortField, ids, resourceNames, clusterIds, hostIds, forestNames) {
    return jsearch.tuples(sortedTuples(sortField, [
        exports.FIELDS.opsdirId, exports.FIELDS.resourceName, exports.FIELDS.clusterId, exports.FIELDS.hostId, exports.FIELDS.forest, exports.FIELDS.ts
    ])).where(optionalAnd(exports.FIELDS.resourceName, resourceNames, optionalAnd(exports.FIELDS.clusterId, clusterIds, optionalAnd(exports.FIELDS.hostId, hostIds, optionalAnd(exports.FIELDS.forest, forestNames, databaseDocs(ids))))));
}
exports.databaseTuples = databaseTuples;
function serverTuples(sortField, ids, resourceNames, clusterIds, groupNames, hostIds) {
    return jsearch.tuples(sortedTuples(sortField, [
        exports.FIELDS.opsdirId, exports.FIELDS.resourceName, exports.FIELDS.clusterId, exports.FIELDS.groupName, exports.FIELDS.hostId, exports.FIELDS.ts
    ])).where(optionalAnd(exports.FIELDS.resourceName, resourceNames, optionalAnd(exports.FIELDS.clusterId, clusterIds, optionalAnd(exports.FIELDS.groupName, groupNames, optionalAnd(exports.FIELDS.hostId, hostIds, serverDocs(ids))))));
}
exports.serverTuples = serverTuples;
function forestTuples(sortField, ids, resourceNames, clusterIds, hostIds) {
    return jsearch.tuples(sortedTuples(sortField, [
        exports.FIELDS.opsdirId, exports.FIELDS.resourceName, exports.FIELDS.clusterId, exports.FIELDS.hostId, exports.FIELDS.ts
    ])).where(optionalAnd(exports.FIELDS.resourceName, resourceNames, optionalAnd(exports.FIELDS.clusterId, clusterIds, optionalAnd(exports.FIELDS.hostId, hostIds, serverDocs(ids)))));
}
exports.forestTuples = forestTuples;
function groupTuples(sortField, ids, resourceNames, clusterIds) {
    return jsearch.tuples(sortedTuples(sortField, [
        exports.FIELDS.opsdirId, exports.FIELDS.resourceName, exports.FIELDS.clusterId, exports.FIELDS.ts
    ])).where(optionalAnd(exports.FIELDS.resourceName, resourceNames, optionalAnd(exports.FIELDS.clusterId, clusterIds, groupDocs(ids))));
}
exports.groupTuples = groupTuples;
function resourceTuples(sortField, ids, resourceNames, resourceTypes, ...additionalCriteria) {
    return jsearch.tuples(sortedTuples(sortField, [
        exports.FIELDS.opsdirId,
        exports.FIELDS.resourceName,
        exports.FIELDS.resourceType
    ])).where(optionalAnd(exports.FIELDS.resourceName, resourceNames, optionalAnd(exports.FIELDS.resourceType, resourceTypes, and(configDocs(ids), ...additionalCriteria))));
}
exports.resourceTuples = resourceTuples;
function take(fieldIndex) {
    return (acc, item) => acc.push(item[fieldIndex]) && acc;
}
function resourcesForCluster(clusterId) {
    return jsearch.facets([
        jsearch.facet(exports.OUT.clusterName, exports.FIELDS.resourceName).thisWhere(clusterDoc(clusterId)),
        // don't this we've solved the group weirdness, yet
        jsearch.facet(exports.OUT.groups, exports.FIELDS.resourceName).thisWhere(groupsForCluster(clusterId)),
        jsearch.facet(exports.OUT.hosts, exports.FIELDS.opsdirId).thisWhere(hostsForCluster(clusterId)),
        jsearch.facet(exports.OUT.databases, exports.FIELDS.opsdirId).thisWhere(databasesForCluster(clusterId)),
        jsearch.facet(exports.OUT.forests, exports.FIELDS.opsdirId).thisWhere(forestsForCluster(clusterId)),
        jsearch.facet(exports.OUT.servers, exports.FIELDS.opsdirId).thisWhere(serversForCluster(clusterId)),
    ]).where(cts.andQuery([configDocs()])).result().facets;
}
exports.resourcesForCluster = resourcesForCluster;
function resourcesForHost(hostId) {
    return jsearch.facets([
        jsearch.facet(exports.OUT.clusterId, exports.FIELDS.opsdirId).thisWhere(cts.andQuery([
            clusterDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.hostId, [hostId])
        ])),
        jsearch.facet(exports.OUT.groupName, exports.FIELDS.group).thisWhere(cts.andQuery([
            hostDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.hostId, [hostId])
        ])),
        jsearch.facet(exports.OUT.databases, exports.FIELDS.opsdirId).thisWhere(cts.andQuery([
            databaseDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.hostId, [hostId])
        ])),
        jsearch.facet(exports.OUT.forests, exports.FIELDS.opsdirId).thisWhere(cts.andQuery([
            forestDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.hostId, [hostId])
        ])),
        jsearch.facet(exports.OUT.servers, exports.FIELDS.opsdirId).thisWhere(cts.andQuery([
            serverDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.hostId, [hostId])
        ]))
    ]).where(cts.andQuery([configDocs()])).result().facets;
}
exports.resourcesForHost = resourcesForHost;
function resourcesForGroup(groupName) {
    return jsearch.facets([
        jsearch.facet(exports.OUT.hosts, exports.FIELDS.opsdirId).thisWhere(cts.andQuery([
            hostDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.group, [groupName])
        ])),
        jsearch.facet(exports.OUT.groupId, exports.FIELDS.opsdirId).thisWhere(cts.andQuery([
            groupDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.groupName, [groupName])
        ])),
        jsearch.facet(exports.OUT.servers, exports.FIELDS.opsdirId).thisWhere(cts.andQuery([
            serverDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.groupName, [groupName])
        ])),
    ]).where(cts.andQuery([configDocs()])).result().facets;
}
exports.resourcesForGroup = resourcesForGroup;
function resourcesForDatabase(databaseId) {
    return jsearch.facets([
        jsearch.facet(exports.OUT.databaseName, exports.FIELDS.resourceName).thisWhere(cts.andQuery([
            databaseDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.opsdirId, [databaseId])
        ])),
        jsearch.facet(exports.OUT.clusterId, exports.FIELDS.opsdirId).thisWhere(cts.andQuery([
            databaseDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.opsdirId, [databaseId])
        ])),
        jsearch.facet(exports.OUT.forests, exports.FIELDS.forest).thisWhere(cts.andQuery([
            databaseDocs(),
            cts.jsonPropertyValueQuery(exports.FIELDS.opsdirId, [databaseId])
        ])),
    ]).where(cts.andQuery([configDocs()])).result().facets;
}
exports.resourcesForDatabase = resourcesForDatabase;
function groupsLookup(byField) {
    return groupTuples(byField, null, null, null).reduce(take(0), []);
}
exports.groupsLookup = groupsLookup;

function getResourceStats(type, id) {
    let rType = undefined;

    switch (type) {
    case "databases":
        rType = "database-status"
        break
    case "hosts":
        rType = "host-status"
        break
    case "forests":
        rType = "forest-status"
        break
    case "servers":
        rType = "server-status"
        break
    default:
        rType = "unknown"
        break
    }

    let criteria = [cts.collectionQuery([exports.rColl]),
                    cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, [rType]),
                           cts.jsonPropertyValueQuery(exports.FIELDS.opsdirId, [id])]
    let { results, estimate } = jsearch.documents()
        .where(cts.andQuery(criteria))
        .result()

    if (estimate === 1) {
        return results[0].document
    } else {
        return undefined
    }
}
exports.getResourceStats = getResourceStats

function getResourceCounts(type, id) {
    let criteria = [cts.collectionQuery([exports.rColl]),
                    cts.jsonPropertyValueQuery(exports.FIELDS.resourceType, ["counts"]),
                           cts.jsonPropertyValueQuery(exports.FIELDS.opsdirId, [id])]
    let { results, estimate } = jsearch.documents()
        .where(cts.andQuery(criteria))
        .result()

    if (estimate === 1) {
        console.log(results[0].document)
        return results[0].document
    } else {
        return undefined
    }
}
exports.getResourceCounts = getResourceCounts
