"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

var jsearch = require('/MarkLogic/jsearch');

function search(params) {
  fn.trace(params, "opsdir-search");
  let query = jsearch.documents()
      .where(cts.andQuery(params.criteria))
      .orderBy(params.sort.map(item => {
        return cts.indexOrder(cts.jsonPropertyReference(item[0]), item[1]);
      }))
      .slice(...params.slice)
      .reduce((accResults, { uri, document }) =>
              { let obj = document.toObject();
                accResults.push(params.fields.reduce((acc, field) => {
                  if (field.type !== 'projection') {
                    acc[field.name] = obj[field.name];
                  }
                  if (field.value){
                    acc[field.name] = field.value;
                  }
                  return acc;
                }, { uri: uri }));
                return accResults;
              }, []);

  let { results, estimate } = query.result();

  let actualDocs = results;
  if (actualDocs && actualDocs.length && params.projections) {
    params.projections.map(projection => projection(actualDocs));
  }

  return {
    items: actualDocs || [],
    estimate
  };
}
exports.search = search;

// The name rawSearch is a little odd. I'm tempted to call it
// unfilteredSearch except that it doesn't mean "filter" in the
// usual sense of filtering search results. Instead, it simply
// means that all of the properties are returned, not a filtered
// subset based on the fields.
function rawSearch(params) {
  let { results, estimate } = jsearch.documents()
      .where(cts.andQuery(params.criteria))
      .orderBy(params.sort.map(item => {
        return cts.indexOrder(cts.jsonPropertyReference(item[0]), item[1]);
      }))
      .slice(...params.slice)
      .reduce((accResults, { uri, document }) => {
        accResults.push(document.toObject());
        return accResults;
      }, [])
      .result();
  let actualDocs = results;
  if (actualDocs && actualDocs.length && params.projections) {
    params.projections.map(projection => projection(actualDocs));
  }
  return {
    items: actualDocs || [],
    estimate
  };
}
exports.rawSearch = rawSearch;

function facetedSearch(params) {
  let { facets, documents, estimate }
      = jsearch.facets(params.facets, jsearch.documents()
                       .where(cts.andQuery(params.criteria))
                       .orderBy(params.sort.map(item => {
                         return cts.indexOrder(cts.jsonPropertyReference(item[0]), item[1]);
                       }))
                       .slice(...params.slice)
                       .reduce((accResults, { uri, document }) => {
                         let obj = document.toObject();
                         accResults.push(params.fields.reduce((acc, field) => {
                           acc[field.name] = obj[field.name];
                           return acc;
                         }, {}));
                         return accResults;
                       }, []))
      .result();
  let actualDocs = documents;
  if (actualDocs && actualDocs.length && params.projections) {
    params.projections.map(projection => projection(actualDocs));
  }
  return {
    items: actualDocs || [],
    estimate,
    facets
  };
}
exports.facetedSearch = facetedSearch;
