"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// https://wiki.marklogic.com/x/uKQoAw

declareUpdate();

const endpoint = require("../lib/endpoint");
const jsearch = require('/MarkLogic/jsearch');
const opsec = require('/common/opsec.xqy');

const fieldDefs = {
  "id": "array",
  "ack": { "type": ["yes", "no"], "default": "yes" },
  "all": { "type": "boolean", "default": false }
};

class EP extends endpoint.Endpoint {
  POST() {
    return this.ack(this.getRequestBody());
  }

  GET() {
    throw this.clientError("Method Not Allowed", 405);
  }

  ack(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, fieldDefs);
    let ids = body.id;

    if (body.id && body.all && (body.all.toString() === "true")) {
      throw this.clientError("You cannot specify IDs and all", 400);
    }

    if (body.all && body.all.toString() === "true") {
      let findAck  = "no";
      if (body.ack === "no") {
        findAck = "yes";
      }
      // Find all the unacknowledged ones
      let qname    = fn.QName("", "opsdirId");
      let nakQuery = cts.jsonPropertyValueQuery("acknowledged", findAck);
      let aQuery   = cts.collectionQuery(["http://marklogic.com/opsdirector/v1/alerts"]);
      ids          = cts.elementValues(qname, null, null, cts.andQuery([nakQuery, aQuery]));
    }

    if (ids) {
      let ackQuery = jsearch.documents()
          .where([
            cts.jsonPropertyValueQuery('opsdirId', ids),
            cts.collectionQuery(['http://marklogic.com/opsdirector/v1/alerts'])
          ])
          .slice(0, Number.MAX_SAFE_INTEGER);

      if (ackQuery.result().results !== null) {
        for (let result of ackQuery.result().results) {
          try {
            // Must be amp'd so it's off in the opsec module
            opsec.acknowledgeAlert(result.uri, body.ack);
          } catch (error) {
            if (error.name === 'SEC-PERMDENIED') {
              throw new endpoint.EndpointError(403, error.message);
            } else {
              throw error;
            }
          }
        }
      }
    }

    xdmp.setResponseCode(204, "No content");
  }
}
endpoint.Endpoint.respond(EP);
