xquery version "1.0-ml";

(: Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. :)

import module namespace opsdir= "http://marklogic.com/v1/opsdirector"
at "/common/opsdir.xqy";

import module namespace jslib= "http://marklogic.com/v1/opsdirector/jslib"
at "/common/jslib.xqy";

import module namespace rg="http://marklogic.com/v1/opsdirector/resource-groups"
at "/common/resource-group.xqy";

declare namespace meter="http://marklogic.com/manage/meters";

declare variable $OneDay := xs:dayTimeDuration("P1D");

declare function local:period(
  $period as xs:string
) as xs:string
{
  if ($period = ("raw", "hour", "day"))
  then $period
  else error((), "Unexpected period: " || $period)
};

declare function local:dateTime(
  $dt as xs:string
) as xs:dateTime
{
  if ($dt castable as xs:dateTime)
  then $dt cast as xs:dateTime
  else error((), "Invalid datetime: " || $dt)
};

declare function local:ids(
  $ids as xs:string*
) as xs:unsignedLong*
{
  for $id in $ids
  return
    if ($id castable as xs:unsignedLong)
    then $id cast as xs:unsignedLong
    else error((), "Invalid ID: " || $id)
};

let $period   := xdmp:get-request-field("format","raw")
let $start    := local:dateTime((xdmp:get-request-field("start"),
                                 string(current-dateTime() - $OneDay))[1])
let $end      := local:dateTime((xdmp:get-request-field("end"),
                                 string(current-dateTime()))[1])
let $summary  := false()
let $detail   := true()
let $cid      := local:ids(xdmp:get-request-field("cluster") ! tokenize(.,','))
let $rid      := local:ids(xdmp:get-request-field("resource") ! tokenize(.,','))
let $length   := xs:integer( xdmp:get-request-field("length","10"))

let $resourceGroupIds := local:ids(xdmp:get-request-field("resourceGroupId") ! tokenize(.,','))

(: retrieve all server ids this user has access to :)
let $server-ids := opsdir:managed-resource-ids("servers")

(: if single rid, check it is accessible:)
let $rids:= if(exists($rid))
    then $rid[xs:string(.) = $server-ids]
    else $server-ids

(: if resourceGroupId supplied, match up resources :)
let $rids := if($resourceGroupIds)
    then let $resources := for $rgid in $resourceGroupIds
                           return rg:get(xs:string($rgid))/resource
         return distinct-values($rids[.=$resources])
    else $rids

return
    jslib:gzip-array-reply(
     array-node{
         subsequence(
             for $server-id in $rids
             let $resource := $server-id ||"/" || $period || "/request-rate"
             let $map := jslib:resource-time-series($resource,$start,$end,$period,xs:QName("meter:server-status"))
             let $map := map:get($map,$resource)
             let $map := map:get($map,map:keys($map))
             let $avg := avg($map)
             order by $avg descending
             return if(exists($map))
                 then
                   let $max := max($map)
                   let $server-resource := opsdir:managed-resources(xs:string($server-id))
                   let $server-name := $server-resource/opsdirName
                   let $cluster-name:= xs:string($server-resource/opsdirClusterName)
                   let $cluster-id:= xs:string($server-resource/opsdirClusterId)
                   return
                     object-node{
                         "serverName"    : string($server-name),
                         "serverId"      : string($server-id),
                         "clusterName"   : $cluster-name,
                         "clusterId"     : $cluster-id,
                         "avgRequestRate": $avg,
	                     "maxRequestRate": $max}
                 else (),1,$length)})
