"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

declareUpdate();

const endpoint = require("../lib/endpoint");

const fieldDefs = {
  "username": { type: "string", required: true },
  "password": { type: "string", required: true }
};

const CONFIGURI = '/config/email-config.json';

class EP extends endpoint.Endpoint {
  GET() {
    let _ = fn.trace({
      "method": xdmp.getRequestMethod(),
      "uri": xdmp.getOriginalUrl()
    }, "opsdir-app-endpoint");

    let headers = [ xdmp.addResponseHeader("Cache-Control", "no-cache"),
                    xdmp.addResponseHeader("Cache-Control", "no-store") ];
    let emailConfig = JSON.parse(cts.doc(CONFIGURI));
    if (emailConfig
        && emailConfig.email.authentication.username
        && emailConfig.email.authentication.password) {
      return {
        "email":
        {
          "authentication":
          {
            "username": emailConfig.email.authentication.username,
            "password": emailConfig.email.authentication.password
          }
        }
      };
    } else {
      return {
        "email":
        {
          "authentication": {}
        }
      };
    }
  }

  POST() {
    let body = this.getRequestBody();

    let _ = fn.trace({
      "method": xdmp.getRequestMethod(),
      "uri": xdmp.getOriginalUrl(),
      "body": body
    }, "opsdir-app-endpoint");

    body = this.validShape(body, fieldDefs);

    if (body.username === "") {
      throw this.clientError("Username must not be empty");
    }

    var emailConfig = cts.doc(CONFIGURI);

    if (emailConfig) {
      var emailConfigObj = emailConfig.toObject();
      emailConfigObj.email.authentication = body;
      xdmp.nodeReplace(emailConfig, emailConfigObj);
      xdmp.setResponseCode(204, "No content");
      return null;
    } else {
      xdmp.documentInsert(CONFIGURI, { "email": { "authentication": body } },
                          { permissions: [xdmp.permission('opsdir-admin', 'read'),
                                          xdmp.permission('opsdir-admin', 'update')] });
      xdmp.setResponseCode(201, "Created");
      return null;
    }
  }

  DELETE() {
    let _ = fn.trace({
      "method": xdmp.getRequestMethod(),
      "uri": xdmp.getOriginalUrl()
    }, "opsdir-app-endpoint");

    var emailConfig = cts.doc(CONFIGURI);

    if (emailConfig) {
      var emailConfigObj = emailConfig.toObject();
      emailConfigObj.email.authentication = {};
      xdmp.nodeReplace(emailConfig, emailConfigObj);
      xdmp.setResponseCode(204, "No content");
      return null;
    }

    xdmp.setResponseCode(204, "No content");
    return null;
  }
}
endpoint.Endpoint.respond(EP);
