"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// See https://wiki.marklogic.com/x/oIjqAw

const endpoint = require("../lib/endpoint");
const info     = require("/common/info.xqy");
const jslib    = require('/common/jslib.xqy');

const LOGS_COLLECTION = 'http://marklogic.com/opsdirector/v1/logs';

const fieldDefs = {
  "filter": "array",
  "query": "array",
  "hostId": "string",
  "clusterId": "string",
  "filename": { type: "string", default: "ErrorLog.txt" },
  "regex": "string",
  "regexFlags": "string",
  "tsFrom": "string",
  "tsTo": "string"
};

class EP extends endpoint.Endpoint {
  POST() {
    return this.query(this.getRequestBody());
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, fieldDefs);

    if (!body.hostId || !body.clusterId) {
      throw this.clientError("Cluster and host IDs must be specified");
    }

    let lines = [];
    try {
      lines = info.getLogsText(body.clusterId, body.hostId, body.filename,
                               body.tsFrom, body.tsTo, body.regex, body.regexFlags);
    } catch (err) {
      if (err && err.name === "OD-NOTMNGD") {
        lines = err.data;
      } else {
        throw err;
      }
    }

    if (lines === undefined) {
      lines = [];
    } else {
      lines = lines.toString().split("\n");
    }

    let result = { "types": [ "logLines" ],
                   "logLines": {
                     "estimate": lines.length,
                     "start": 0,
                     "count": lines.length,
                     "items": lines
                   }
                 };

    return jslib.gzipReply(result);
  }
}

endpoint.Endpoint.respond(EP);
