xquery version "1.0-ml";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace pki="http://marklogic.com/xdmp/pki"
    at "/MarkLogic/pki.xqy";

import module namespace admin="http://marklogic.com/xdmp/admin"
    at "/MarkLogic/admin.xqy";

import module namespace manage="http://marklogic.com/manage"
    at "/MarkLogic/manage/manage.xqy";

import module namespace config-lock="http://marklogic.com/config-lock"
    at "/MarkLogic/Admin/config-lock.xqy";

import module namespace odi="http://marklogic.com/v1/opsdirector/install"
    at "/common/opsdir-install.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

declare namespace group="http://marklogic.com/xdmp/group";
declare namespace x509="http://marklogic.com/xdmp/x509";

declare option xdmp:mapping "false";

declare variable $OPSDIRECTOR-CA-COMMON-NAME := "MarkLogic Ops Director Certificate Authority";

declare %private variable $iso-opts :=
  <options xmlns="xdmp:eval">
    <isolation>different-transaction</isolation>
    <transaction-mode>update-auto-commit</transaction-mode>
    <prevent-deadlocks>false</prevent-deadlocks>
  </options>;

declare %private variable $sec-opts :=
  <options xmlns="xdmp:eval">
    <database>{xdmp:security-database()}</database>
  </options>;

declare %private function local:install(
  $payload as object-node()
)
{
  let $installed := local:opsdir-ca-id($payload)
  let $opsdirCA  := if ($installed)
                    then
                      $installed
                    else
                      xdmp:invoke-function(function () {
                        local:install-opsdir-ca($payload)
                      }, $sec-opts)

  let $_ := if (exists($opsdirCA)
                and empty(pki:get-template-by-name($odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)))
            then
              (xdmp:invoke-function(
                 function () {
                   local:create-opsdir-certificate-template($payload)
                 }, $iso-opts),
               xdmp:invoke-function(
                 function () {
                   local:create-signed-host-cert()
                 }, $iso-opts))
            else
              ()

  let $_ := if (exists($opsdirCA))
            then
              xdmp:invoke-function(
                function () {
                  try {
                    xdmp:external-security($odi:OPSDIRECTOR-EXT-SEC-NAME)
                  } catch ($e) {
                    let $_ := trace("Creating external security", $t:trace-debug)
                    return
                       sec:create-external-security(
                          $odi:OPSDIRECTOR-EXT-SEC-NAME,
                          $odi:OPSDIRECTOR-EXT-SEC-DESC,
                          "certificate", 300, "internal",
                          (), (), $opsdirCA, fn:true())
                  }
                },
                <options xmlns="xdmp:eval">
                  { $iso-opts/*, $sec-opts/* }
                </options>)
            else
              ()

  let $_ := if ($opsdirCA)
            then
              xdmp:invoke-function(
                function () {
                  let $_ := trace("Configuring " || $odi:OPSDIRECTOR-DATA-SERVER
                                  || " for SSL.", $t:trace-status)
                  let $config    := admin:get-configuration()
                  let $group     := xdmp:group(local:select-group($payload, "datagroup"))
                  let $appserver := admin:appserver-get-id($config, $group,
                                                           $odi:OPSDIRECTOR-DATA-SERVER)
                  let $templ     := pki:get-template-by-name(
                                      $odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)
                                      /pki:template-id
                  let $config    := admin:appserver-set-external-security(
                                      $config, $appserver, ($odi:OPSDIRECTOR-EXT-SEC-NAME),
                                      fn:true(), "application-level")
                  let $config    := admin:appserver-set-ssl-certificate-template(
                                      $config, $appserver, $templ)
                  return
                    admin:save-configuration($config)
                  }, $iso-opts)
            else
              ()

  let $_ :=
    xdmp:invoke-function(
      function () {
        odi:save-opsdir-ca-id($opsdirCA)
      },
      $iso-opts)

  return
    if (empty($installed) and exists($opsdirCA))
    then
      pki:get-certificates($opsdirCA)
    else
      $installed
};

declare %private function local:opsdir-ca-id(
  $payload as object-node()
) as xs:unsignedLong?
{
  (: Is the user telling us to use a particular CA? :)
  if ($payload/ca castable as xs:unsignedLong)
  then
    $payload/ca cast as xs:unsignedLong
  else
    (: Have we been down this road before? :)
    let $dbid := try { xdmp:database($odi:OPSDIRECTOR-DATABASE) } catch ($e) { () }
    return
      if (exists($dbid))
      then
        let $id := xdmp:invoke-function(function () {
                     doc("/config/certificate-authority.txt")/text()
                   },
                   <options xmlns="xdmp:eval">
                     <database>{$dbid}</database>
                   </options>)
        return
          if (empty($id))
          then
            ()
          else
            $id cast as xs:unsignedLong
      else
        ()
};

declare function local:install-opsdir-ca(
  $payload as object-node()
) as xs:unsignedLong?
{
  let $_    := trace("Ops Director installer: installing Ops Director CA...",
                     $t:trace-debug)
  let $caid := if ($payload/ca castable as xs:unsignedLong)
               then $payload/ca cast as xs:unsignedLong
               else $payload/ca
  return
    if ($caid instance of object-node())
    then
      try {
        let $id := xdmp:credential-id("OpsDirector-Credential")
        return
          (xdmp:log("OpsDirector-Credential already exists"),
           $id)
      } catch ($e) {
        local:generate-opsdir-ca($caid)
      }
    else if ($caid = "EXTERNAL")
    then
      let $_ := trace("No credential: using externally signed certificates", $t:trace-debug)
      return
        ()
    else
      let $privkey := odi:certificate-authority-private-key($caid)
      let $certificate := pki:get-certificates($caid)
      let $_ := trace("Creating OpsDirector login credential", $t:trace-debug)
      let $credential-id := sec:create-credential(
                               $t:CRED-NAME,
                               "OpsDirector Login Credential",
                               (), (), $certificate/pki:pem, $privkey,
                               fn:true(), (),
                               (xdmp:permission("opsdir-admin-internal", "read")))
      let $_ := trace("Installed credential for existing ca:" || $credential-id,
                      $t:trace-debug)
      return
        $caid
};

declare function local:generate-opsdir-ca(
  $ca as object-node()
) as xs:unsignedLong?
{
  let $keys     := xdmp:rsa-generate()
  let $privkey  := $keys[1]
  let $pubkey   := $keys[2]

  let $country  := $ca/country/string()
  let $province := $ca/province/string()
  let $locality := $ca/locality/string()
  let $orgName  := $ca/organization/string()
  let $orgUnit  := $ca/unit/string()
  let $email    := $ca/email/string()

  let $x509     :=
    <cert xmlns="http://marklogic.com/xdmp/x509">
      <version>2</version>
      <serialNumber>{pki:integer-to-hex(xdmp:random())}</serialNumber>
      <issuer>
        <countryName>{$country}</countryName>
        <stateOrProvinceName>{$province}</stateOrProvinceName>
        <localityName>{$locality}</localityName>
        <organizationName>{$orgName}</organizationName>
        <organizationalUnitName>{$orgUnit}</organizationalUnitName>
        <emailAddress>{$email}</emailAddress>
        <commonName>{$OPSDIRECTOR-CA-COMMON-NAME}</commonName>
      </issuer>
      <validity>
        <notBefore>{fn:current-dateTime()}</notBefore>
        <notAfter>{fn:current-dateTime() + xs:dayTimeDuration("P3650D")}</notAfter>
      </validity>
      <subject>
        <countryName>{$country}</countryName>
        <stateOrProvinceName>{$province}</stateOrProvinceName>
        <localityName>{$locality}</localityName>
        <organizationName>{$orgName}</organizationName>
        <organizationalUnitName>{$orgUnit}</organizationalUnitName>
        <emailAddress>{$email}</emailAddress>
        <commonName>{$OPSDIRECTOR-CA-COMMON-NAME}</commonName>
      </subject>
      <publicKey>{$pubkey}</publicKey>
      <v3ext>
        <basicConstraints critical="false">CA:TRUE</basicConstraints>
        <keyUsage critical="false">Certificate Sign, CRL Sign</keyUsage>
        <subjectKeyIdentifier critical="false">
          {pki:integer-to-hex(xdmp:random())}
        </subjectKeyIdentifier>
      </v3ext>
    </cert>

  let $certificate :=
    xdmp:x509-certificate-generate($x509, $privkey, <options xmlns="ssl:options"/>)

  let $_ := trace("Installing OpsDirector certificate authority", $t:trace-debug)
  let $_ := trace("Creating OpsDirector login credential", $t:trace-debug)
  let $credential-id := sec:create-credential(
                           $t:CRED-NAME,
                           "OpsDirector Login Credential",
                           (), (), $certificate, $privkey,
                           fn:true(), (),
                           (xdmp:permission("opsdir-admin-internal", "read")))
  let $certificate-id := pki:insert-trusted-certificates($certificate)
  return
    $certificate-id
};

declare function local:create-opsdir-certificate-template(
  $payload as object-node()
) as xs:unsignedLong
{
  if (empty(pki:get-template-by-name($odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)))
  then
    let $_ := trace("Creating OpsDirector certificate template", $t:trace-debug)
    return
      pki:insert-template(
        pki:create-template(
          $odi:OPSDIRECTOR-SSL-TEMPLATE-NAME, $odi:OPSDIRECTOR-SSL-TEMPLATE-DESC,
          (), (),
          element x509:req {
            element x509:version {0},
            element x509:subject {
              element x509:organizationName { $odi:OPSDIRECTOR-SSL-TEMPLATE-ORG },
              $odi:OPSDIRECTOR-SSL-TEMPLATE-DEPT ! element x509:organizationalUnitName { . }
            },
            element x509:v3ext {
              element x509:nsCertType {
                attribute critical {"false"},
                "SSL Server"
              },
              element x509:subjectKeyIdentifier {
                attribute critical {"false"},
                pki:integer-to-hex(xdmp:random())
              }
            }
          }))
  else
    (),

  let $template-id := pki:get-template-by-name($odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)
                       /pki:template-id
  for $hostid in xdmp:group-hosts(xdmp:group(local:select-group($payload, "datagroup")))
  let $host := xdmp:host-name($hostid)
  let $_ := trace("Creating temporary certificates for " || $host || " if necessary.",
                  $t:trace-debug)
  return
    pki:generate-temporary-certificate-if-necessary(
           $template-id, 3650, $host, (), ())
};

declare function local:create-signed-host-cert()
{
  let $config := admin:get-configuration()
  for $host in admin:get-host-ids($config) ! xdmp:host-name(.)
  let $_ := trace("Creating signed host certificate for " || $host, "setup-cba")
  let $csr-pem :=
        pki:generate-certificate-request(
          pki:get-template-by-name($odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)/pki:template-id,
                                   $host, (), ())
  let $csr-xml := xdmp:x509-request-extract($csr-pem)
  let $cert-xml :=
    <cert xmlns="http://marklogic.com/xdmp/x509">
      <version>2</version>
      <serialNumber>{pki:integer-to-hex(xdmp:random())}</serialNumber>
      <validity>
        <notBefore>{fn:current-dateTime()}</notBefore>
        <notAfter>{fn:current-dateTime() + xs:dayTimeDuration("P3650D")}</notAfter>
      </validity>
      {$csr-xml/x509:subject}
      {$csr-xml/x509:publicKey}
      {$csr-xml/x509:v3ext}
    </cert>
  let $cert-pem :=
    xdmp:x509-certificate-generate(
      $cert-xml, (),
      <options xmlns="ssl:options">
        <credential-id>{xdmp:credential-id($t:CRED-NAME)}</credential-id>
      </options>)
  return
    pki:insert-signed-certificates($cert-pem)
};

declare function local:install-tasks(
  $group as xs:unsignedLong
) as empty-sequence()
{
  let $_        := trace("Ops Director installer: installing tasks...", $t:trace-debug)
  let $config   := admin:get-configuration()

  (: N.B. These aren't real config objects; the content and modules databases
     are computed without regard to what's configured (or not) here. :)
  let $tasks  := <tasks xmlns="http://marklogic.com/manage">
        <task>
          <task-path>/common/tasks/info.xqy</task-path>
          <task-root>/</task-root>
          <task-type>minutely</task-type>
          <task-period>1</task-period>
        </task>
        <task>
          <task-path>/common/tasks/running.xqy</task-path>
          <task-root>/</task-root>
          <task-type>minutely</task-type>
          <task-period>1</task-period>
        </task>
        <task>
          <task-path>/common/tasks/expire.xqy</task-path>
          <task-root>/</task-root>
          <task-type>daily</task-type>
          <task-period>1</task-period>
        </task>
        <task>
          <task-path>/common/tasks/health.xqy</task-path>
          <task-root>/</task-root>
          <task-type>minutely</task-type>
          <task-period>1</task-period>
        </task>
      </tasks>

  (: Delete any existing tasks that are left over from 1.x :)
  let $curtasks := admin:group-get-scheduled-tasks($config, $group)
  (: /MarkLogic || /opsdirector is intentional; the build process does a string repl :)
  let $odtasks  := $curtasks[starts-with(group:task-path, "/MarkLogic" || "/opsdirector/")]
  let $config   := admin:group-delete-scheduled-task($config, $group, $odtasks)

  (: Now (re)create the new ones :)
  let $config := fold-left(function($config, $task) {
                             local:task-create($config, $group, $task)
                           },
                           $config, $tasks/manage:task)
  return
    admin:save-configuration($config)
};

declare %private function local:get-admin-user-id(
) as xs:unsignedLong
{
  (: We don't know the name(s) of the admin user(s). Find all of the users
     with the admin role and return the first one. This is the same approach
     that CPF takes.
  :)
  let $ids := xdmp:invoke-function(function() {
                let $admin := sec:get-role-ids("admin")
                for $user in collection("http://marklogic.com/xdmp/users")/sec:user
                  where $user/sec:role-ids/sec:role-id = $admin
                  order by $user/sec:get-user
                  return
                    $user/sec:user-id/data()
              },
              <options xmlns="xdmp:eval">
                <database>{xdmp:security-database()}</database>
              </options>)
  return
    $ids[1]
};

declare %private function local:task-create(
  $config as element(configuration),
  $group as xs:unsignedLong,
  $task as element(manage:task)
) as element(configuration)
{
  (: FIXME: Data groups :)
  let $dbid   := admin:database-get-id($config, "OpsDirector")
  let $modid  := admin:database-get-id($config, "OpsDirectorModules")
  let $found  := for $gtask in admin:group-get-scheduled-tasks($config, $group)
                 where $gtask/group:task-path = $task/manage:task-path
                       and $gtask/group:task-root = $task/manage:task-root
                       and $gtask/group:task-database = $dbid
                       and $gtask/group:task-type = $task/manage:task-type
                 return
                   $gtask
  let $config := if (exists($found))
                 then
                   $config
                 else
                   let $_     := trace("Create task: " || $task/manage:task-path,
                                       $t:trace-status)
                   let $gtask := if ($task/manage:task-type = "minutely")
                                 then
                                   admin:group-minutely-scheduled-task(
                                     $task/manage:task-path,
                                     $task/manage:task-root,
                                     $task/manage:task-period cast as xs:positiveInteger,
                                     $dbid, $modid, local:get-admin-user-id(), (), ())
                                 else if ($task/manage:task-type = "daily")
                                      then
                                        admin:group-daily-scheduled-task(
                                        $task/manage:task-path,
                                        $task/manage:task-root,
                                        $task/manage:task-period cast as xs:positiveInteger,
                                        xs:time("23:00:00"),
                                        $dbid, $modid, local:get-admin-user-id(), (), ())
                                      else
                                        error((), "OD-INVTASK", "Invalid task type")
                   return
                     admin:group-add-scheduled-task($config, $group, $gtask)
  return
    $config
};

declare function local:select-group(
  $payload as object-node(),
  $field as xs:string?
) as xs:string
{
  let $group := $payload/*[local-name(.) = $field]
  return
    if (exists($group))
    then
      $group
    else
      xdmp:group-name(xdmp:group())
};

xdmp:security-assert("http://marklogic.com/xdmp/privileges/admin-ui", "execute"),

let $_       := if (xdmp:get-request-method() = "POST")
                then ()
                else error((), "Not a POST?")

let $payload := xdmp:get-request-body("json")/object-node()
let $_       := config-lock:lock-config(
                  function () {
                    (local:install($payload),
                     let $group := xdmp:group(local:select-group($payload, "datagroup"))
                     return
                       xdmp:invoke-function(function () { local:install-tasks($group) }, $iso-opts))
                  })
return
  "Hello world"
