"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

const search = require('../lib/search');
const endpoint = require("../lib/endpoint");
const jslib = require('/common/jslib.xqy');

const fieldDefs = {
  "filter": "array",
  "query": "array",
  "text": "string",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json"], default: "json" },

  "resourceType": "array",
  "id": "array",
  "clusterId": "array",
  "hostId": "array"
};

let baseQuery = [
  cts.collectionQuery(['http://marklogic.com/opsdirector/v1/resource']),
  cts.jsonPropertyValueQuery("opsdirKind", "resource")
];

class EP extends endpoint.Endpoint {
  POST() {
    return this.query(this.getRequestBody());
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, fieldDefs);

    if (body.resourceType) {
      baseQuery.push(cts.jsonPropertyValueQuery("opsdirResourceType", body.resourceType));
    }
    if (body.id) {
      baseQuery.push(cts.jsonPropertyValueQuery("opsdirId", body.id));
    }
    if (body.clusterId) {
      baseQuery.push(cts.jsonPropertyValueQuery("opsdirClusterId", body.clusterId));
    }
    if (body.hostId) {
      baseQuery.push(cts.jsonPropertyValueQuery("opsdirHostId", body.hostId));
    }

    let results = {
      "type": [ "properties" ],
      "properties": search.search(body, baseQuery)
    };

    return jslib.gzipReply(results);
  }
}

endpoint.Endpoint.respond(EP);
