"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

const endpoint = require("../lib/endpoint");
const opsdir   = require("/common/opsdir.xqy");

const fieldDefs = {
  "filter": "array",
  "view": { "type": "string" }
};

class EP extends endpoint.Endpoint {
  POST() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    let id = xdmp.getRequestField("id");
    if (!id) {
      throw this.clientError("Missing ID");
    }

    body = this.validShape(body, fieldDefs);

    let result = undefined;
    if (body.view == "security") {
      // Security only applies to clusters; make sure the ID supplied
      // is treated like a cluster id. (This will 404 for attempts to
      // use the security view on other resource types.
      result = opsdir.managedResource(id, body.view, id);
    } else {
      result = opsdir.managedResource(null, body.view, id);

      // handle server special case by providing IDs for the databases
      if (result && result.serverType && result.serverName && result.contentDatabase) {
        let serverResult = result.toObject();
        for (let prop of ["contentDatabase", "modulesDatabase", "lastLoginDatabase"]) {
          if (prop in serverResult) {
            let dbid = opsdir.managedResourceId(serverResult.opsdirClusterId,
                                                "databases",serverResult[prop]);
            if (dbid) {
              serverResult[prop + "Id"] = dbid;
              result=serverResult;
            }
          }
        }
      }
    }

    // Remove the keys we don't need
    if (body.filter) {
      for (let key in result) {
        if (body.filter.indexOf(key) == -1) {
          delete result[key];
        }
      }
    }

    if (result === null) {
      throw this.clientError("Not Found", 404);
    }

    return result;
  }
}
endpoint.Endpoint.respond(EP);
