"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// Sends a simple email message. Designed to test whether an email
// address works or not.

const endpoint = require("../lib/endpoint");
const cfg      = require('/common/config.xqy');

class EP extends endpoint.Endpoint {
  POST() {
    let body = this.getRequestBody();

    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                        "execute");

    try {
      let options = {
        "verifyCert": false
      };

      let emailConfig = cts.doc("/config/email-config.json");
      if (emailConfig) {
        emailConfig = emailConfig.toObject();
        if (emailConfig.email && emailConfig.email.authentication) {
          options.authentication = {
            "username": emailConfig.email.authentication.username,
            "password": emailConfig.email.authentication.password,
          }
        }
      }

      // Fix the from address using the values configured in the server
      body.from = { "name": cfg.getAtomicProperty("email-return-name",
                                                  "Do Not Reply"),
                    "address": cfg.getAtomicProperty("email-return-address",
                                                  "do-not-reply@marklogic.com") }

      xdmp.email(body, options);
      xdmp.setResponseCode(204, "No content");
    } catch (error) {
      console.log(error);
      throw this.clientError("Sending email failed; see logs");
    }

    return null;
  }
}

endpoint.Endpoint.respond(EP);
