"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// See https://wiki.marklogic.com/x/o4jqAw

const endpoint = require("../lib/endpoint");
const jslib    = require('/common/jslib.xqy');
const cfg      = require('/common/config.xqy');
const admin    = require("/MarkLogic/admin.xqy");

declareUpdate();

class SessionEndpoint extends endpoint.Endpoint {
  GET() {
    // N.B. There will *always* be a user logged in. The default
    // user on the application server if nothing else.

    this.checkAuth();

    const headers = [ xdmp.addResponseHeader("Cache-Control", "no-cache"),
                    xdmp.addResponseHeader("Cache-Control", "no-store") ];

    var uname = xdmp.getRequestUsername();
    var name = xdmp.getCurrentUser();
    var admin = false;
    var licenseAdmin = false;

    try {
      xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin","execute");
      admin = true;
    } catch (error) {
      // nop;
    }

    try {
      xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-license-admin",
                          "execute");
      licenseAdmin = true;
    } catch (error) {
      // nop;
    }

    let version = fn.head(cfg.getAtomicProperty("version")) || "1.1-1";

    let names = [];
    for (let name of xdmp.getCurrentRoles().toArray().map(
                        roleId => xdmp.roleName(roleId.toString()))) {
      if (!name.startsWith("opsdir-rg-internal")) {
        names.push(name);
      }
    }

    return {
      "server": jslib.clientSessionStatus(),
      "user": {
        "username": uname,
        "roles": names,
        "admin": admin,
        "licenseAdmin": licenseAdmin
      }
    };
  }

  POST() {
    let posted = this.getRequestBody();

    // Do we need to login?
    if (xdmp.getCurrentUser() !== posted.username) {
      if (xdmp.login(posted.username, posted.password)) {
        // check after login attempt if the user has
        // authorization to proceed further
        this.checkAuth();
      } else {
        throw new endpoint.EndpointError(401);
      }
    }

    // Are we in an upgrade scenario?
    let version = fn.head(cfg.getAtomicProperty("version")) || "1.0-1";
    if (version.substring(0, 4) !== "1.1-") {
      return { "redirect": "/v2/upgrade" };
    }

    // Ok, we're logged in (or we were already logged in) and we aren't
    // upgrading, just return the details.
    return {
      "server": jslib.clientSessionStatus(),
      "user": {
        "username": posted.username,
        "roles": xdmp.getCurrentRoles().toArray().map(
          roleId => xdmp.roleName(roleId.toString()))
      }
    };
  }

  DELETE() {
    return xdmp.logout();
  }

  // checks if the user has sufficient privileges to login to opsdirector app
  checkAuth() {
    let body = this.getRequestParams();
    let authchk = body["authchk"] === "true";

    let loggedIn = false;
    try {
      xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-user","execute");
      loggedIn = true;
    } catch (error) {
      if (error.name !== "SEC-PRIV") {
        throw error;
      }
    }

    if (!loggedIn && !authchk) {
      try {
        xdmp.logout();
      } catch (err) {
        console.log("Failed to logout: " + xdmp.getCurrentUser());
      }
      throw new endpoint.EndpointError(403);
    }
  }
}

endpoint.Endpoint.respond(SessionEndpoint);
