xquery version "1.0-ml";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace pki="http://marklogic.com/xdmp/pki"
    at "/MarkLogic/pki.xqy";

import module namespace admin="http://marklogic.com/xdmp/admin"
    at "/MarkLogic/admin.xqy";

import module namespace scmod="http://marklogic.com/manage/secure-credentials"
    at "/MarkLogic/manage/models/secure-credentials-model.xqy";

import module namespace manage="http://marklogic.com/manage"
    at "/MarkLogic/manage/manage.xqy";

import module namespace config-lock="http://marklogic.com/config-lock"
    at "/MarkLogic/Admin/config-lock.xqy";

import module namespace odi="http://marklogic.com/v1/opsdirector/install"
    at "/common/opsdir-install.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

declare namespace group="http://marklogic.com/xdmp/group";
declare namespace x509="http://marklogic.com/xdmp/x509";

declare option xdmp:mapping "false";

declare variable $OPSDIRECTOR-CA-COMMON-NAME := "MarkLogic Ops Director Certificate Authority";

declare %private variable $iso-opts :=
  <options xmlns="xdmp:eval">
    <isolation>different-transaction</isolation>
    <transaction-mode>update-auto-commit</transaction-mode>
    <prevent-deadlocks>false</prevent-deadlocks>
  </options>;

declare %private variable $sec-opts :=
  <options xmlns="xdmp:eval">
    <database>{xdmp:security-database()}</database>
  </options>;

declare %private function local:uninstall(
  $payload as object-node()
)
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/admin-ui", "execute"),
  local:teardown-server-config(),
  (: Delete all remaining opsdir-* users and credentials :)
  xdmp:invoke-function(function () {
      (/sec:user[matches(sec:user-name, "^opsdir-\d+$")],
        /sec:credential[matches(sec:credential-name, "^opsdir-\d+$")])
        ! xdmp:document-delete(xdmp:node-uri(.))
      }, $sec-opts)
};

declare function local:teardown-server-config(
)
{
  let $dbid     := xdmp:database($odi:OPSDIRECTOR-DATABASE)
  let $caid     := odi:opsdirector-ca-id()

  let $config   := admin:get-configuration()
  let $config := local:remove-servers($config)

  let $_      := xdmp:invoke-function(function () {
                     local:remove-tasks($config)
                   }, $iso-opts)

  let $_      := local:remove-amps()
  let $_      := local:remove-roles()
  let $_      := local:remove-privileges()

  let $_      := if ($caid)
                 then
                   xdmp:invoke-function(function () {
                       local:remove-security()
                     },
                     <options xmlns="xdmp:eval">
                       { $iso-opts/*, $sec-opts/* }
                     </options>)
                 else
                   ()
  return
    ()
};

declare function local:teardown-managed(
)
{
  let $config   := admin:get-configuration()
  let $endpoint := admin:cluster-get-opsdirector-session-endpoint($config)
  let $endpoint := substring-before(substring-after($endpoint, "//"), ":")
  return
    if ($endpoint = xdmp:host-name() or $endpoint = "localhost")
    then
      if (odi:externally-signed-certificate())
      then local:teardown-external-managed-config()
      else local:teardown-self-managed-config()
    else
      ()
};

declare function local:teardown-external-managed-config(
)
{
  let $_ := local:teardown-cluster-config()
  return
    ()
};

declare function local:teardown-self-managed-config(
)
{
  let $cert-id  := odi:opsdirector-ca-id()
  let $_  := local:teardown-cluster-config()
  let $_  := odi:teardown-opsdir-admin-user()
  let $ca := odi:external-security-certificate-authorities()
  let $_  := odi:teardown-external-security()
  let $_  := odi:uninstall-template()

  let $_  := try {
                xdmp:invoke-function(function () {
                   let $cred := scmod:identify("MarkLogic-OpsDirector")
                   return
                     scmod:delete-secure-credential(string($cred))
                  }, <options xmlns="xdmp:eval">
                       { $iso-opts/*, $sec-opts/* }
                     </options>)
             } catch ($e) {
               xdmp:log("Failed to remove secure credential: MarkLogic-OpsDirector")
             }

  return
    if (exists($ca) and empty($cert-id))
    then
      pki:delete-certificate($ca)
    else
      ()
};

declare function local:teardown-cluster-config(
)
{
  let $config    := admin:get-configuration()
  let $config    := fold-left(
                      function ($config, $group-id) {
                        let $id := try {
                                     admin:appserver-get-id($config, $group-id,
                                                            $odi:SECURE-MANAGEMENT-API)
                                   } catch ($e) {
                                     ()
                                   }
                        return
                          if (exists($id))
                          then
                            admin:appserver-delete($config, $id)
                          else
                            $config
                      }, $config, admin:get-group-ids($config))
  let $config    := admin:cluster-set-opsdirector-session-endpoint($config, "")
  let $config    := admin:cluster-set-opsdirector-log-level($config, "disabled")
  let $config    := admin:cluster-set-opsdirector-metering($config, "disabled")
  let $config    := admin:cluster-set-opsdirector-config($config, "disabled")
  return
    admin:save-configuration-without-restart($config)
};

declare %private function local:remove-servers(
  $config as element(configuration)
) as element(configuration)
{
  (: N.B. This function only removes the data a server; this is on purpose.
     The data server has to be removed so that the security artifacts can
     be removed. But I want ml-gradle to remove a server so that it causes
     a cluster restart. :)
  fold-left(
    function ($config, $app-id) {
      let $name := admin:appserver-get-name($config, $app-id)
      return
        if ($name eq $odi:OPSDIRECTOR-DATA-SERVER)
        then
          let $_ := xdmp:log("Deleting appserver: " || $name)
          return
            admin:appserver-delete($config, $app-id)
        else
          $config
    }, $config, admin:get-appserver-ids($config))
};

declare function local:remove-tasks(
  $config as element(configuration)
)
{
  let $_        := trace("Ops Director uninstaller: uninstalling tasks...", $t:trace-debug)
  let $modules  := xdmp:database("OpsDirectorModules")
  let $curtasks := admin:group-get-scheduled-tasks($config, xdmp:group())
  let $odtasks  := $curtasks[group:task-modules = $modules]
  let $config   := admin:group-delete-scheduled-task($config, xdmp:group(), $odtasks)
  return
    admin:save-configuration-without-restart($config)
};

declare %private function local:remove-amps(
)
{
  let $base-namespace := "http://marklogic.com/v1/opsdirector/"
  return
    xdmp:invoke-function(
      function () {
        for $amp in /sec:amp[starts-with(sec:namespace, $base-namespace)]
        return
          sec:remove-amp($amp/sec:namespace, $amp/sec:local-name,
                         $amp/sec:document-uri, $amp/sec:database)
      },
      <options xmlns="xdmp:eval">
        { $iso-opts/*, $sec-opts/* }
      </options>)
};

declare %private function local:remove-privileges(
)
{
  let $execute := ("http://marklogic.com/xdmp/privileges/opsdir-license-admin",
                   "http://marklogic.com/xdmp/privileges/opsdir-admin",
                   "http://marklogic.com/xdmp/privileges/opsdir-user",
                   "http://marklogic.com/xdmp/privileges/opsdir-data-internal")
  let $uri     := ("/certs/",
                   "/sessions/",
                   "/config-timestamps/",
                   "/logs/",
                   "/config/",
                   "/stats/",
                   "/resource/",
                   "/resource-scope/",
                   "/resource-group/",
                   "/meters/",
                   "/alerts/")
  return
    (for $p in $execute
     return
       xdmp:invoke-function(
         function () {
           try {
             sec:remove-privilege($p, "execute")
           } catch ($e) {
             xdmp:log("Error removing privilege: " || $p)
           }
         },
         <options xmlns="xdmp:eval">
           { $iso-opts/*, $sec-opts/* }
         </options>),

     for $p in $uri
     return
       xdmp:invoke-function(
         function () {
           try {
             sec:remove-privilege($p, "uri")
           } catch ($e) {
             xdmp:log("Error removing privilege: " || $p)
           }
         },
         <options xmlns="xdmp:eval">
           { $iso-opts/*, $sec-opts/* }
         </options>))
};

declare %private function local:remove-roles(
)
{
  let $rg-roles := xdmp:invoke-function(function () {
                        for $role in /sec:role/sec:role-name/string()
                        where matches($role, "^opsdir-rg-internal-[0-9a-f]+$")
                        return
                          $role
                     },
                     <options xmlns="xdmp:eval">
                       { $iso-opts/*, $sec-opts/* }
                     </options>)

  let $roles := ("opsdir-admin",
                 "opsdir-admin-internal",
                 "opsdir-data-internal",
                 "opsdir-license-admin",
                 "opsdir-user",
                 "opsdir-user-internal")
  return
    for $r in ($roles, $rg-roles)
    return
      xdmp:invoke-function(
        function () {
          try {
            sec:remove-role($r)
          } catch ($e) {
             xdmp:log(("Error removing role: " || $r, $e))
          }
        },
        <options xmlns="xdmp:eval">
          { $iso-opts/*, $sec-opts/* }
        </options>)
};

declare %private function local:remove-security(
)
{
  let $ca    := xdmp:invoke-function(
                  function () {
                    sec:external-security-get-ssl-client-certificate-authorities(
                        "OpsDirector-External-Security")
                  }, $sec-opts)

  let $_     := try {
                  if (exists($ca))
                  then
                    pki:delete-certificate($ca)
                  else
                    ()
                } catch ($e) {
                  xdmp:log(("Failed to remove external security", $e))
                }

  let $templ := pki:get-template-by-name("OpsDirector-SSL-Template")
  let $_     := if (exists($templ))
                then
                  pki:delete-template($templ/pki:template-id)
                else
                  ()

  let $_     := try {
                  sec:remove-credential($t:CRED-NAME)
                } catch ($e) {
                  (xdmp:log("Failed to remove secure credential: " || $t:CRED-NAME),
                   xdmp:log($e))
                }

  let $_     := try {
                  sec:remove-external-security("OpsDirector-External-Security")
                } catch ($e) {
                  (xdmp:log("Failed to remove external security: OpsDirector-External-Security"),
                   xdmp:log($e))
                }
  return
    ()
};

(: ============================================================ :)

declare %private function local:x-install(
  $payload as object-node()
)
{
  let $installed := local:opsdir-ca-id($payload)
  let $opsdirCA  := if ($installed)
                    then
                      $installed
                    else
                      xdmp:invoke-function(function () {
                        local:install-opsdir-ca($payload)
                      }, $sec-opts)

  let $_ := if (exists($opsdirCA)
                and empty(pki:get-template-by-name($odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)))
            then
              (xdmp:invoke-function(
                 function () {
                   local:create-opsdir-certificate-template($payload)
                 }, $iso-opts),
               xdmp:invoke-function(
                 function () {
                   local:create-signed-host-cert()
                 }, $iso-opts))
            else
              ()

  let $_ := if (exists($opsdirCA))
            then
              xdmp:invoke-function(
                function () {
                  try {
                    xdmp:external-security($odi:OPSDIRECTOR-EXT-SEC-NAME)
                  } catch ($e) {
                    let $_ := trace("Creating external security", $t:trace-debug)
                    return
                       sec:create-external-security(
                          $odi:OPSDIRECTOR-EXT-SEC-NAME,
                          $odi:OPSDIRECTOR-EXT-SEC-DESC,
                          "certificate", 300, "internal",
                          (), (), $opsdirCA, fn:true())
                  }
                },
                <options xmlns="xdmp:eval">
                  { $iso-opts/*, $sec-opts/* }
                </options>)
            else
              ()

  let $_ := if ($opsdirCA)
            then
              xdmp:invoke-function(
                function () {
                  let $_ := trace("Configuring " || $odi:OPSDIRECTOR-DATA-SERVER
                                  || " for SSL.", $t:trace-status)
                  let $config    := admin:get-configuration()
                  let $group     := xdmp:group(local:select-group($payload, "datagroup"))
                  let $appserver := admin:appserver-get-id($config, $group,
                                                           $odi:OPSDIRECTOR-DATA-SERVER)
                  let $templ     := pki:get-template-by-name(
                                      $odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)
                                      /pki:template-id
                  let $config    := admin:appserver-set-external-security(
                                      $config, $appserver, ($odi:OPSDIRECTOR-EXT-SEC-NAME),
                                      fn:true(), "application-level")
                  let $config    := admin:appserver-set-ssl-certificate-template(
                                      $config, $appserver, $templ)
                  return
                    admin:save-configuration($config)
                  }, $iso-opts)
            else
              ()

  let $_ :=
    xdmp:invoke-function(
      function () {
        odi:save-opsdir-ca-id($opsdirCA)
      },
      $iso-opts)

  return
    if (empty($installed) and exists($opsdirCA))
    then
      pki:get-certificates($opsdirCA)
    else
      $installed
};

declare %private function local:opsdir-ca-id(
  $payload as object-node()
) as xs:unsignedLong?
{
  (: Is the user telling us to use a particular CA? :)
  if ($payload/ca castable as xs:unsignedLong)
  then
    $payload/ca cast as xs:unsignedLong
  else
    (: Have we been down this road before? :)
    let $dbid := try { xdmp:database($odi:OPSDIRECTOR-DATABASE) } catch ($e) { () }
    return
      if (exists($dbid))
      then
        let $id := xdmp:invoke-function(function () {
                     doc("/config/certificate-authority.txt")/text()
                   },
                   <options xmlns="xdmp:eval">
                     <database>{$dbid}</database>
                   </options>)
        return
          if (empty($id))
          then
            ()
          else
            $id cast as xs:unsignedLong
      else
        ()
};

declare function local:install-opsdir-ca(
  $payload as object-node()
) as xs:unsignedLong?
{
  let $_    := trace("Ops Director installer: installing Ops Director CA...",
                     $t:trace-debug)
  let $caid := if ($payload/ca castable as xs:unsignedLong)
               then $payload/ca cast as xs:unsignedLong
               else $payload/ca
  return
    if ($caid instance of object-node())
    then
      try {
        let $id := xdmp:credential-id("OpsDirector-Credential")
        return
          (xdmp:log("OpsDirector-Credential already exists"),
           $id)
      } catch ($e) {
        local:generate-opsdir-ca($caid)
      }
    else if ($caid = "EXTERNAL")
    then
      let $_ := trace("No credential: using externally signed certificates", $t:trace-debug)
      return
        ()
    else
      let $privkey := odi:certificate-authority-private-key($caid)
      let $certificate := pki:get-certificates($caid)
      let $_ := trace("Creating OpsDirector login credential", $t:trace-debug)
      let $credential-id := sec:create-credential(
                               $t:CRED-NAME,
                               "OpsDirector Login Credential",
                               (), (), $certificate/pki:pem, $privkey,
                               fn:true(), (),
                               (xdmp:permission("opsdir-admin-internal", "read")))
      let $_ := trace("Installed credential for existing ca:" || $credential-id,
                      $t:trace-debug)
      return
        $caid
};

declare function local:generate-opsdir-ca(
  $ca as object-node()
) as xs:unsignedLong?
{
  let $keys     := xdmp:rsa-generate()
  let $privkey  := $keys[1]
  let $pubkey   := $keys[2]

  let $country  := $ca/country/string()
  let $province := $ca/province/string()
  let $locality := $ca/locality/string()
  let $orgName  := $ca/organization/string()
  let $orgUnit  := $ca/unit/string()
  let $email    := $ca/email/string()

  let $x509     :=
    <cert xmlns="http://marklogic.com/xdmp/x509">
      <version>2</version>
      <serialNumber>{pki:integer-to-hex(xdmp:random())}</serialNumber>
      <issuer>
        <countryName>{$country}</countryName>
        <stateOrProvinceName>{$province}</stateOrProvinceName>
        <localityName>{$locality}</localityName>
        <organizationName>{$orgName}</organizationName>
        <organizationalUnitName>{$orgUnit}</organizationalUnitName>
        <emailAddress>{$email}</emailAddress>
        <commonName>{$OPSDIRECTOR-CA-COMMON-NAME}</commonName>
      </issuer>
      <validity>
        <notBefore>{fn:current-dateTime()}</notBefore>
        <notAfter>{fn:current-dateTime() + xs:dayTimeDuration("P3650D")}</notAfter>
      </validity>
      <subject>
        <countryName>{$country}</countryName>
        <stateOrProvinceName>{$province}</stateOrProvinceName>
        <localityName>{$locality}</localityName>
        <organizationName>{$orgName}</organizationName>
        <organizationalUnitName>{$orgUnit}</organizationalUnitName>
        <emailAddress>{$email}</emailAddress>
        <commonName>{$OPSDIRECTOR-CA-COMMON-NAME}</commonName>
      </subject>
      <publicKey>{$pubkey}</publicKey>
      <v3ext>
        <basicConstraints critical="false">CA:TRUE</basicConstraints>
        <keyUsage critical="false">Certificate Sign, CRL Sign</keyUsage>
        <subjectKeyIdentifier critical="false">
          {pki:integer-to-hex(xdmp:random())}
        </subjectKeyIdentifier>
      </v3ext>
    </cert>

  let $certificate :=
    xdmp:x509-certificate-generate($x509, $privkey, <options xmlns="ssl:options"/>)

  let $_ := trace("Installing OpsDirector certificate authority", $t:trace-debug)
  let $_ := trace("Creating OpsDirector login credential", $t:trace-debug)
  let $credential-id := sec:create-credential(
                           $t:CRED-NAME,
                           "OpsDirector Login Credential",
                           (), (), $certificate, $privkey,
                           fn:true(), (),
                           (xdmp:permission("opsdir-admin-internal", "read")))
  let $certificate-id := pki:insert-trusted-certificates($certificate)
  return
    $certificate-id
};

declare function local:create-opsdir-certificate-template(
  $payload as object-node()
) as xs:unsignedLong
{
  if (empty(pki:get-template-by-name($odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)))
  then
    let $_ := trace("Creating OpsDirector certificate template", $t:trace-debug)
    return
      pki:insert-template(
        pki:create-template(
          $odi:OPSDIRECTOR-SSL-TEMPLATE-NAME, $odi:OPSDIRECTOR-SSL-TEMPLATE-DESC,
          (), (),
          element x509:req {
            element x509:version {0},
            element x509:subject {
              element x509:organizationName { $odi:OPSDIRECTOR-SSL-TEMPLATE-ORG },
              $odi:OPSDIRECTOR-SSL-TEMPLATE-DEPT ! element x509:organizationalUnitName { . }
            },
            element x509:v3ext {
              element x509:nsCertType {
                attribute critical {"false"},
                "SSL Server"
              },
              element x509:subjectKeyIdentifier {
                attribute critical {"false"},
                pki:integer-to-hex(xdmp:random())
              }
            }
          }))
  else
    (),

  let $template-id := pki:get-template-by-name($odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)
                       /pki:template-id
  for $hostid in xdmp:group-hosts(xdmp:group(local:select-group($payload, "datagroup")))
  let $host := xdmp:host-name($hostid)
  let $_ := trace("Creating temporary certificates for " || $host || " if necessary.",
                  $t:trace-debug)
  return
    pki:generate-temporary-certificate-if-necessary(
           $template-id, 3650, $host, (), ())
};

declare function local:create-signed-host-cert()
{
  let $config := admin:get-configuration()
  for $host in admin:get-host-ids($config) ! xdmp:host-name(.)
  let $_ := trace("Creating signed host certificate for " || $host, "setup-cba")
  let $csr-pem :=
        pki:generate-certificate-request(
          pki:get-template-by-name($odi:OPSDIRECTOR-SSL-TEMPLATE-NAME)/pki:template-id,
                                   $host, (), ())
  let $csr-xml := xdmp:x509-request-extract($csr-pem)
  let $cert-xml :=
    <cert xmlns="http://marklogic.com/xdmp/x509">
      <version>2</version>
      <serialNumber>{pki:integer-to-hex(xdmp:random())}</serialNumber>
      <validity>
        <notBefore>{fn:current-dateTime()}</notBefore>
        <notAfter>{fn:current-dateTime() + xs:dayTimeDuration("P3650D")}</notAfter>
      </validity>
      {$csr-xml/x509:subject}
      {$csr-xml/x509:publicKey}
      {$csr-xml/x509:v3ext}
    </cert>
  let $cert-pem :=
    xdmp:x509-certificate-generate(
      $cert-xml, (),
      <options xmlns="ssl:options">
        <credential-id>{xdmp:credential-id($t:CRED-NAME)}</credential-id>
      </options>)
  return
    pki:insert-signed-certificates($cert-pem)
};

declare %private function local:get-admin-user-id(
) as xs:unsignedLong
{
  (: We don't know the name(s) of the admin user(s). Find all of the users
     with the admin role and return the first one. This is the same approach
     that CPF takes.
  :)
  let $ids := xdmp:invoke-function(function() {
                let $admin := sec:get-role-ids("admin")
                for $user in collection("http://marklogic.com/xdmp/users")/sec:user
                  where $user/sec:role-ids/sec:role-id = $admin
                  order by $user/sec:get-user
                  return
                    $user/sec:user-id/data()
              },
              <options xmlns="xdmp:eval">
                <database>{xdmp:security-database()}</database>
              </options>)
  return
    $ids[1]
};

declare %private function local:task-create(
  $config as element(configuration),
  $group as xs:unsignedLong,
  $task as element(manage:task)
) as element(configuration)
{
  (: FIXME: Data groups :)
  let $dbid   := admin:database-get-id($config, "OpsDirector")
  let $modid  := admin:database-get-id($config, "OpsDirectorModules")
  let $found  := for $gtask in admin:group-get-scheduled-tasks($config, $group)
                 where $gtask/group:task-path = $task/manage:task-path
                       and $gtask/group:task-root = $task/manage:task-root
                       and $gtask/group:task-database = $dbid
                       and $gtask/group:task-type = $task/manage:task-type
                 return
                   $gtask
  let $config := if (exists($found))
                 then
                   $config
                 else
                   let $_     := trace("Create task: " || $task/manage:task-path,
                                       $t:trace-status)
                   let $gtask := if ($task/manage:task-type = "minutely")
                                 then
                                   admin:group-minutely-scheduled-task(
                                     $task/manage:task-path,
                                     $task/manage:task-root,
                                     $task/manage:task-period cast as xs:positiveInteger,
                                     $dbid, $modid, local:get-admin-user-id(), (), ())
                                 else if ($task/manage:task-type = "daily")
                                      then
                                        admin:group-daily-scheduled-task(
                                        $task/manage:task-path,
                                        $task/manage:task-root,
                                        $task/manage:task-period cast as xs:positiveInteger,
                                        xs:time("23:00:00"),
                                        $dbid, $modid, local:get-admin-user-id(), (), ())
                                      else
                                        error((), "OD-INVTASK", "Invalid task type")
                   return
                     admin:group-add-scheduled-task($config, $group, $gtask)
  return
    $config
};

declare function local:select-group(
  $payload as object-node(),
  $field as xs:string?
) as xs:string
{
  let $group := $payload/*[local-name(.) = $field]
  return
    if (exists($group))
    then
      $group
    else
      xdmp:group-name(xdmp:group())
};

xdmp:security-assert("http://marklogic.com/xdmp/privileges/admin-ui", "execute"),

let $_       := if (xdmp:get-request-method() = "POST")
                then ()
                else error((), "OD-NOTPOST", "Not a POST?")

let $payload := xdmp:get-request-body("json")/object-node()
let $_       := if ($payload/uninstall != true())
                then error((), "OPSDIR-NOTUNINSTALL", "Invalid uninstall payload")
                else ()
let $_       := xdmp:invoke-function(function () {
                  local:teardown-managed()
                }, $iso-opts)
let $_       := config-lock:lock-config(
                  function () {
                    local:uninstall($payload)
                  })
return
  "Hello world"
