xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

(: Note: This module has been written so that it can be shared between
   XQuery and JavaScript. Consequently it uses camelCase naming conventions
   and passes ID values around as strings instead of xs:unsignedLongs.
   Because Javascript. Also, the resource scope XML is not in any namespace.
   Because JSON.
:)

module namespace cfg="http://marklogic.com/v1/opsdirector/config";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $config-uri := "/config/opsdirector.xml";

declare function cfg:get-atomic-property(
  $name as xs:string
) as xs:anyAtomicType*
{
  cfg:get-atomic-property($name, ())
};

declare function cfg:get-atomic-property(
  $name as xs:string,
  $default as xs:anyAtomicType?
) as xs:anyAtomicType*
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  (: Always get properties in a non-update transaction :)
  xdmp:invoke-function(
    function () {
      let $config := doc($config-uri)/cfg:config
      let $qname  := fn:QName("http://marklogic.com/v1/opsdirector/config", $name)
      return
        if ($config/*[node-name(.) = $qname])
        then
          $config/*[node-name(.) = $qname] ! data(.)
        else
          $default
    },
    <options xmlns="xdmp:eval">
      <isolation>different-transaction</isolation>
      <update>false</update>
      <prevent-deadlocks>false</prevent-deadlocks>
    </options>)
};

declare function cfg:set-atomic-property(
  $name as xs:string,
  $value as xs:anyAtomicType?
) as xs:anyAtomicType*
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                       "execute"),

  let $config := doc($config-uri)/cfg:config
  let $qname  := fn:QName("http://marklogic.com/v1/opsdirector/config", $name)
  let $node   := $config/*[node-name(.) = $qname]
  return
    if (empty($node))
    then
      xdmp:node-insert-child($config,
                             element { $qname } { $value })
    else
      if (count($node) gt 1)
      then
        error((), "OD-CONFIG", "Property " || $name || " has multiple values")
      else
        xdmp:node-replace($node,
                          element { $qname } { $value })
};
