xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace raw="http://marklogic.com/v1/opsdirector/rawdata"
    at "/common/rawdata.xqy";

import module namespace ts="http://marklogic.com/v1/opsdirector/security"
    at "/common/opsec.xqy";

import module namespace j="http://marklogic.com/v1/opsdirector/joins"
    at "/common/joins.xqy";

import module namespace io="http://marklogic.com/v1/opsdirector/cluster-io"
    at "/common/cluster-io.xqy";

import module namespace cc="http://marklogic.com/v1/opsdirector/camel-case"
    at "/common/camel-case.xqy";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare %private variable $LOCK := "/config/joins.lock";

declare variable $json as object-node() external;
declare variable $cluster-id as xs:string external;
declare variable $host-id as xs:string external;

declare %private function local:config-uri(
  $timestamp as xs:string
) as xs:string
{
  local:config-uri((), (), $timestamp)
};

declare %private function local:config-uri(
  $type as xs:string?,
  $id as xs:unsignedLong?,
  $timestamp as xs:string
) as xs:string
{
  let $uritype := if (empty($type)) then "clusters" else $type
  let $idpart  := "/" || $cluster-id
                  || (if (empty($id)) then "" else "/" || $id)
                  || "/"
  return
    "/resource/" || $uritype || $idpart || $timestamp || ".json"
};

declare %private function local:config-list-uri(
  $type as xs:string?,
  $timestamp as xs:string
) as xs:string
{
  let $uritype := if (empty($type)) then "cluster" else $type
  let $idpart  := "/" || $cluster-id || "/list/"
  return
    "/resource/" || $uritype || $idpart || $timestamp || ".json"
};

declare function local:get-cluster-properties(
  $timestamp as xs:string,
  $hostname as xs:string
) as object-node()?
{
  let $coll    := $t:COLLROOT || "resource"
  return
    let $resp    := io:get($cluster-id, $hostname, "/manage/v2/properties")
    let $current := raw:resource($cluster-id, "clusters", $cluster-id)

    let $etag := if (string($resp[1]) eq "ERROR")
                 then ()
                 else $resp[1]/*:headers/*:etag
    let $j    := $resp[2]/node()
    let $j    := if (empty($j)) then object-node { } else $j

    where empty($j/error)
    return
      let $map  := map:new(
                     (map:entry("opsdirEtag", string($etag)),
                      map:entry("opsdirKind", "resource"),
                      map:entry("opsdirUid", concat(string($cluster-id),'-',xdmp:random())),
                      map:entry("opsdirClusterId", $cluster-id),
                      map:entry("opsdirHostId", $host-id),
                      map:entry("opsdirClusterName", string($j/cluster-name)),
                      map:entry("opsdirHostName", $hostname)))

      let $json := t:add-opsdir-properties($j, $current, "clusters", $cluster-id, $map)
      let $json := cc:to-camel-case($json)
      let $uri  := local:config-uri($timestamp)

      (: We can simply test for doc-available here
         because there's only one cluster resource! :)
      return
        if (exists($j) and not(doc-available($uri)))
        then
          let $_    := trace("Updating cluster " || $cluster-id, $t:trace-status)
          let $_    := raw:archive-resource($current)
          let $_    := trace("Creating  " || $uri, $t:trace-debug)
          return
            (local:store($uri, $json), $json)
        else
          doc($uri)/object-node()
};

declare function local:get-properties(
  $cluster as object-node(),
  $type as xs:string,
  $timestamp as xs:string,
  $hostname as xs:string
) as empty-sequence()
{
  let $coll    := $t:COLLROOT || "resource"
  let $cluster-name := string($cluster/opsdirName)

  return
    let $json := io:get($cluster-id, $hostname, "/manage/v2/" || $type)[2]/node()
    where exists($json) and empty($json/error)
    return
      let $_       := local:process-deleted-resources($json, $type, $timestamp)
      let $names   := $json/*/list-items/list-item[nameref]

      let $map     := map:new(
                        (map:entry("opsdirKind", "resource-list"),
                         map:entry("opsdirClusterId", $cluster-id),
                         map:entry("opsdirClusterName", $cluster-name),
                         map:entry("opsdirHostId", $host-id),
                         map:entry("opsdirClusterName", string($cluster/opsdirName)),
                         map:entry("opsdirHostName", $hostname)))

      let $listid  := string($cluster-id) || "-" || $type || "-list"
      let $json    := cc:to-camel-case(object-node { "resources": array-node { $names } })
      let $json    := t:add-opsdir-properties($json, (), "resource-list", $listid, $map)

      (: The list URIs serve as an indication of whether or
         not the resources have been updated :)
      let $curlist := raw:resource($cluster-id, "resource-list",
                                   $cluster-id || "-" || $type || "-list")
      let $uri     := local:config-list-uri($type, $timestamp)
      let $_       := if (not(empty($curlist)))
                      then (raw:archive-resource($curlist),
                            trace("Creating  " || $uri, $t:trace-debug))

                      else ()
      let $_       := local:store($uri, $json)

      for $li in $names
      let $id       := $li/idref
      let $group    := $li/groupnameref
      let $name     := $li/nameref
      let $params   := if (exists($group))
                       then "?group-id=" || $group
                       else ""
      let $resp     := io:get($cluster-id, $hostname, "/manage/v2/" || $type
                              || "/" || $id || "/properties" || $params)

      let $current  := raw:resource($cluster-id, $type, string($id))

      let $etag   := if (string($resp[1]) eq "ERROR")
                     then ()
                          (: on 13 Jan 2018 I saw a response with multiple, identical
                             etag headers; I can't reproduce it, but I can defend
                             against it. :)
                     else ($resp[1]/*:headers/*:etag)[1]
      let $j      := $resp[2]/node()

      where (empty($j/error))
      return
        let $map    := map:new(
                         (map:entry("opsdirEtag", string($etag)),
                          map:entry("opsdirKind", "resource"),
                          map:entry("opsdirUid", concat(string($id),'-',xdmp:random())),
                          map:entry("opsdirClusterId", $cluster-id),
                          map:entry("opsdirHostId", $host-id),
                          map:entry("opsdirClusterName", string($cluster/opsdirName)),
                          map:entry("opsdirHostName", $hostname),
                          if ($type eq 'hosts')
                          then map:entry("opsdirGroupName", $j/group)
                          else (),
                          if ($type eq 'servers')
                          then map:entry("opsdirGroupName", $j/group-name)
                          else (),
                          if ($type eq 'forests')
                          then map:entry("hostName", $j/host) (: normalize for indexing :)
                          else ()))

        let $j  := t:add-opsdir-properties($j, $current, $type, string($id), $map)

        let $j  := xdmp:to-json(
                       map:new(
                         (
                          for $property in ($j/node() except $j/id)
                          return
                            map:entry(local-name-from-QName(node-name($property)), $property)
                         )))

        let $new      := cc:to-camel-case($j/node())

        (: Hosts are a special case. In order to know what to poll, we need
           host status documents for each host. :)
        let $_        := if ($type eq 'hosts')
                         then
                           local:make-host-status($new)
                         else
                           ()
        return
          if (local:config-changed($current, $new))
          then
            let $_ := raw:archive-resource($current)
            let $_ := trace("Creating  " || $uri, $t:trace-debug)
            let $_ := if (empty($current))
                      then trace("Updating " || $type || " " || $id || " (" || $name || ")",
                                 $t:trace-status)
                      else ()
            return
              local:store(local:config-uri($type, $id, $timestamp), $new)
          else
            let $_ := trace("Unchanged " || $type || " " || $id || " (" || $name || ")",
                            $t:trace-upload)
            return
              ()
};

declare %private function local:make-host-status(
  $host as object-node()
) as empty-sequence()
{
  let $uri := "/config/host-status/" || $host/opsdirClusterId || "/"
              || $host/opsdirId || "/snapshot.json"
  (: If the document already exists, leave it alone; what we're aiming to
     do here is make sure that a status documents gets created for every
     newly added host :)
  where not(doc-available($uri))
  return
    (: Yes the timestamp is absurdly in the past; we want to make sure
       that the new ones are considered the oldest so that they get
       scanned first. :)
    let $obj := object-node {
        "message": "unknown",
        "ts": "1970-01-01T00:00:00Z",
        "opsdirHostId": $host/opsdirId,
        "status": "unknown",
        "opsdirClusterId": $host/opsdirClusterId
      }
    return
      xdmp:document-insert($uri, $obj, $t:DBPERMS, ())
};

declare %private function local:config-changed(
  $current as object-node()?,
  $new as object-node()
) as xs:boolean
{
  (: Hack to remove the opsdirUid and ts properties because they always change :)
  let $map := if (empty($current)) then map:map() else xdmp:from-json($current)
  let $_   := map:delete($map, "opsdirUid")
  let $_   := map:delete($map, "ts")
  let $compare_cur := if (empty($current)) then () else xdmp:to-json($map)
  let $map := xdmp:from-json($new)
  let $_   := map:delete($map, "opsdirUid")
  let $_   := map:delete($map, "ts")
  let $compare_new := xdmp:to-json($map)
  let $same := deep-equal($compare_cur, $compare_new)
  return
    not($same)
};

declare function local:process-deleted-resources(
  $json as object-node(),
  $type as xs:string,
  $timestamp as xs:string
) as empty-sequence()
{
  let $rsrcids := raw:resource-ids($cluster-id, $type)
  let $ids     := $json/*/list-items/list-item[nameref]/idref/string()
  let $deleted := for $id in $rsrcids
                  where not($id = $ids)
                  return
                    $id
  let $rsrcs   := $deleted ! raw:resource($cluster-id, $type, .)
  let $_       := $rsrcs ! raw:archive-resource(.)
  return
    if (empty($deleted))
    then ()
    else
      let $_ := trace("Deleted " || $type || ": " || string-join($deleted, ", "),
                      $t:trace-status)
      let $_ := ts:remove-resources-from-resource-group($type, $deleted)
      let $map := map:new(
                    (map:entry("opsdirKind", "deleted"),
                     map:entry("resources", array-node { $deleted })))
      let $doc := t:add-opsdir-properties(object-node { }, (), $type, (), $map)
      let $uri := "/resource/deleted/" || $type || "/" || $cluster-id
                  || "/" || $timestamp || ".json"
      return
        local:store($uri, $doc, $t:COLLROOT || "resource/deleted")
};

declare function local:get-cluster(
  $timestamp as xs:string,
  $hostname as xs:string
)
{
  (: Simply test for the config URI because there's only one cluster resource! :)
  let $uri  := local:config-uri($timestamp)
  return
    if (doc-available($uri))
    then
      (false(), doc($uri)/object-node())
    else
      (true(), local:get-cluster-properties($timestamp, $hostname))
};

declare function local:get(
  $cluster as object-node(),
  $type as xs:string?,
  $timestamp as xs:string,
  $hostname as xs:string
) as xs:boolean
{
  local:get($cluster, $type, $timestamp, $hostname, false())
};

declare function local:get(
  $cluster as object-node(),
  $type as xs:string?,
  $timestamp as xs:string,
  $hostname as xs:string,
  $force as xs:boolean
) as xs:boolean
{
  let $uri    := local:config-list-uri($type, $timestamp)
  let $update := not(doc-available($uri))

  (: If we're forcing, then the underlying config timestamp may not have
     changed. But if that's the case, we'll attempt to archive the current
     one and then save a new one with the same name. That won't work!
     So fudge the new timestamp. :)
  let $timestamp := if ($force)
                    then string(current-dateTime())
                    else $timestamp

  let $props  := if ($update or $force)
                 then local:get-properties($cluster, $type, $timestamp, $hostname)
                 else ()
  return
    $update or $force
};

declare function local:store(
  $uri as xs:string,
  $json as object-node()
) as empty-sequence()
{
  local:store($uri, $json,
              if (contains($uri, "/list/"))
              then $t:COLLROOT || "resource"
              else $t:COLLROOT || "resource")
};

declare function local:store(
  $uri as xs:string,
  $json as object-node(),
  $collection as xs:string?
) as empty-sequence()
{
  if ($json/errorResponse)
  then
    let $_ := trace(("ERROR: " || $uri, $json), $t:trace-debug-configs)
    return
      ()
  else
    let $_ := trace("Ins: " || $uri, $t:trace-debug-uri)
    let $_ := trace("Insert: " || $uri, $t:trace-debug-insert)
    let $_ := xdmp:document-insert($uri, $json, $t:DBPERMS, $collection)
    return
      ()
};

declare function local:get-configs(
  $timestamps as object-node(),
  $hostname as xs:string
)
{
  (: Force update access to be sequential :)
  xdmp:lock-for-update($LOCK),

  let $clusterts   := $timestamps/configFiles[name='clusters.xml']/timestamp
  let $clusterts   := t:timestamp(xs:dateTime($clusterts))

  let $cl-update   := local:get-cluster($clusterts, $hostname)
  let $force       := $cl-update[1]
  let $cluster     := $cl-update[2]
  return
    if (empty($cluster))
    then
      xdmp:log("Failed to get cluster status for " || $hostname || "; giving up.")
    else
      (: If the cluster timestamp has changed, update everything; this assures
         that if the cluster name changed, the opsdirClusterName property in
         each resource will be updated. Technically, this is probably only necessary
         if the cluster name changes, but cluster changes are sufficiently uncommon
         that it doesn't seem necessary to be that precise. :)

      let $forests     := let $config := $timestamps/configFiles[name = "assignments.xml"]
                          let $ts := $config/timestamp
                          return local:get($cluster, "forests", $ts, $hostname, $force)

      (: If the forests have changed, update the databases in case a forest name changed. :)

      let $databases   := let $config := $timestamps/configFiles[name = "databases.xml"]
                          let $ts := $config/timestamp
                          return local:get($cluster, "databases", $ts, $hostname,
                                           ($force or $forests))

      (: If the databases have changed but we didn't update the forests, update the
         forests as well in case a database name changed :)
      let $forests     := if ($databases and not($forests))
                          then
                            let $config := $timestamps/configFiles[name = "assignments.xml"]
                            let $ts := $config/timestamp
                            return local:get($cluster, "forests", $ts, $hostname, true())
                          else
                            $forests

      let $groups      := let $config := $timestamps/configFiles[name = "groups.xml"]
                          let $ts := $config/timestamp
                          return
                            local:get($cluster, "groups", $ts, $hostname, $force)

      (: If the databases have changed, update the servers in case a database name changed. :)

      let $servers     := let $config := $timestamps/configFiles[name = "groups.xml"]
                          let $ts := $config/timestamp
                          return
                            local:get($cluster, "servers", $ts, $hostname,
                                      $force or $databases)

      let $hosts       := let $config := $timestamps/configFiles[name = "hosts.xml"]
                          let $ts := $config/timestamp
                          return local:get($cluster, "hosts", $ts, $hostname, $force)

      for $config in $timestamps/configFiles
      let $filename := $config/name
      return
        switch ($filename)
        case "assignments.xml" return ()
        case "clusters.xml" return ()
        case "databases.xml" return ()
        case "groups.xml" return ()
        case "hosts.xml" return ()
        default return xdmp:log("Update configs: Unexpected filename: " || $filename)
};

let $timestamps := $json
let $hostname   := $timestamps/hostName/data()
let $_          := trace("Updating configuration from " || $hostname,  $t:trace-debug)
return
  local:get-configs($timestamps, $hostname),

xdmp:invoke-function(
  function () {
    (xdmp:sleep(4000),
     j:update-joins())
  },
  <options xmlns="xdmp:eval">
    <isolation>different-transaction</isolation>
    <transaction-mode>update-auto-commit</transaction-mode>
  </options>)

