xquery version "1.0-ml";

(: Copyright 2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace j="http://marklogic.com/v1/opsdirector/joins";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace raw="http://marklogic.com/v1/opsdirector/rawdata"
    at "/common/rawdata.xqy";

import module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes"
    at "/common/rscopes.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

import module namespace p="http://marklogic.com/v1/opsdirector/polling"
    at "/common/polling.xqy";

declare namespace host="http://marklogic.com/xdmp/status/host";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $COLL := $t:COLLROOT || "join";

declare %private variable $QOPTS
  := ("collation=http://marklogic.com/collation/");

declare %private variable $LOCK := "/config/joins.lock";

declare function j:clusters-table(
  $qhost as object-node()?
)
{
  let $_ := trace("Computing clusters joins", $t:trace-joins)

  let $current := cts:search(collection(),
                     cts:directory-query("/joins/clusters/", "infinity"))/opsdirId
  let $all     := cts:search(collection($t:COLLROOT||"resource"),
                     cts:json-property-value-query("opsdirResourceType", "clusters", "exact"))
                     /object-node()
  let $update  := $all/opsdirId
  let $_       := $current[not(. = $update)]
                    ! xdmp:document-delete("/joins/clusters/" || . || ".json")

  for $cid in raw:resource-ids("clusters")
  let $cluster := raw:resources("clusters", $cid)
  let $cluster-name := $cluster/opsdirName
  let $group-count := count(raw:resource-ids($cid, "groups"))
  let $hosts := raw:resource-ids($cid, "hosts")
  let $host-count := count($hosts)
  let $database-count := count(raw:resource-ids($cid, "databases"))
  let $forest-count := count(raw:resource-ids($cid, "forests"))
  let $server-count := count(raw:resource-ids($cid, "servers"))

  let $security := raw:resource($cid, "security", $cid)
  let $details  := map:map()
  let $_        := for $prop in $security/keystore/*
                   (: Fuss with property names :)
                   let $name  := local-name-from-QName(node-name($prop))
                   let $first := upper-case(substring($name, 1, 1))
                   let $name  := "keystore" || $first || substring($name, 2)
                   return
                     map:put($details, $name, $prop/data())

  let $status   := raw:resource($cid, "local-cluster-status", $cid)
  let $sinfo    := map:new()
  let $_        := if (exists($status))
                   then map:put($sinfo, "statusProperties", $status/statusProperties)
                   else ()
  let $_        := if (map:keys($sinfo))
                   then map:put($details, "statusInfo", $sinfo)
                   else ()

  let $map := j:host-details($cid, $hosts)

  let $effectiveVersion := min(map:get($map, "effectiveVersion"))
  let $effectiveVersion := if (empty($effectiveVersion))
                           then "Unknown"
                           else $effectiveVersion

  let $version          := array-node {
                             distinct-values(map:get($map, "version"))
                           }

  let $lastStartup      := max(map:get($map, "lastStartup"))
  let $lastStartup      := if (empty($lastStartup))
                           then "Unknown"
                           else $lastStartup

  let $osVersion        := array-node {
                             distinct-values(map:get($map, "osVersion"))
                           }

  let $obj := xdmp:to-json($details
                + object-node {
                    "ts": string(current-dateTime()),
                    "opsdirResourceType": "clusters",
                    "opsdirId": $cid,
                    "opsdirClusterId": $cid,
                    "opsdirName": $cluster-name,
                    "groupCount": $group-count,
                    "groupIds": array-node { raw:resource-ids($cid, "groups") },
                    "hostCount": $host-count,
                    "hostIds": array-node { raw:resource-ids($cid, "hosts") },
                    "databaseCount": $database-count,
                    "databaseIds": array-node { raw:resource-ids($cid, "databases") },
                    "forestCount": $forest-count,
                    "forestIds": array-node { raw:resource-ids($cid, "forests") },
                    "serverCount": $server-count,
                    "serverIds": array-node { raw:resource-ids($cid, "servers") },
                    "version": $version,
                    "effectiveVersion": $effectiveVersion,
                    "lastStartup": $lastStartup,
                    "encryption": if (empty($security/keystore/dataEncryption))
                                  then "Unknown"
                                  else $security/keystore/dataEncryption,
                    "osVersion": $osVersion
                  })

  let $perms := rs:effective-resource-permissions($cluster)
  return
    if (empty($qhost) or $qhost/opsdirClusterId = $cid)
    then
      let $_ := trace("Store /joins/clusters/" || $cid || ".json " || $cluster-name, $t:trace-joins-uri)
      return
        xdmp:document-insert("/joins/clusters/" || $cid || ".json", $obj, $perms, $COLL)
    else
      ()
};

declare function j:hosts-table(
  $qhost as object-node()?
)
{
  let $_ := trace("Computing hosts joins", $t:trace-joins)
  for $cid in raw:resource-ids("clusters")
  let $cluster := raw:resources("clusters", $cid)
  let $cluster-name := $cluster/opsdirName
  let $current := cts:search(collection(),
                     cts:and-query((
                        cts:directory-query("/joins/hosts/", "infinity"),
                        cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))
                    /opsdirId
  let $hosts   := cts:search(collection($t:COLLROOT||"resource"),
                     cts:and-query((
                        cts:json-property-value-query("opsdirResourceType", "hosts", "exact"),
                        cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))
                    /object-node()
  let $update  := $hosts/opsdirId
  let $_       := j:delete("/joins/hosts/", $cid, $current[not(. = $update)])

  let $forests := raw:resources-of-type($cid, "forests")
  let $dbs     := raw:resources-of-type($cid, "databases")
  let $groups  := raw:resources-of-type($cid, "groups")
  let $servers := raw:resources-of-type($cid, "servers")

  for $host in j:limit("/joins/hosts/", $cid, $hosts)
  let $details := j:host-details($cid, $host/opsdirId)
  let $_       := map:delete($details, "encryption")
  let $_       := map:delete($details, "lastStartup")

  let $forests := $forests[hostName = $host/opsdirName]

  let $dbs     := $dbs[$forests/forestName = forest]

  let $groups  := $groups[opsdirName = $host/opsdirGroupName]

  let $servers := $servers[opsdirGroupName = $host/opsdirGroupName]

  let $status  := raw:resource($cid, "host-status", $host/opsdirId)
  let $detail  := $status/statusProperties/statusDetail
  let $disksp  := sum(($detail/hostSize/value,
                       $detail/hostLargeDataSize/value,
                       $detail/dataDirSpace/value))

  let $sinfo   := map:new()
  let $_       := if (exists($status))
                  then map:put($sinfo, "statusProperties", $status/statusProperties)
                  else ()
  let $_       := if (map:keys($sinfo))
                  then map:put($details, "statusInfo", $sinfo)
                  else ()

  let $obj     := xdmp:to-json($details
                    + object-node {
                        "ts": string(current-dateTime()),
                        "opsdirResourceType": "hosts",
                        "opsdirId": string($host/opsdirId),
                        "opsdirName": string($host/opsdirName),
                        "opsdirClusterId": $cid,
                        "opsdirClusterName": $cluster-name,
                        "groupId": string($groups/opsdirId),
                        "groupName": string($host/opsdirGroupName),
                        "forestCount": string(count($forests)),
                        "forestIds": array-node { $forests ! string(./opsdirId) },
                        "forestNames": array-node { $forests ! string(./opsdirName) },
                        "databaseCount": string(count($dbs)),
                        "databaseIds": array-node { $dbs ! string(./opsdirId) },
                        "databaseNames": array-node { $dbs ! string(./opsdirName) },
                        "serverCount": count($servers),
                        "serverIds": array-node { $servers ! string(./opsdirId) },
                        "serverNames": array-node { $servers ! string(./opsdirName) },
                        "diskSpace": string($disksp),
                        "maintenanceMode": string($host/hostMode),
                        "zone": string($host/zone)
                     })

  let $perms := rs:effective-resource-permissions($host)

  return
    if (empty($qhost) or $qhost/opsdirClusterId = $cid)
    then
      let $_ := trace("Store /joins/hosts/" || $cid || "/" || $host/opsdirId
                      || ".json " || $host/opsdirName, $t:trace-joins-uri)
      return
        xdmp:document-insert("/joins/hosts/" || $cid || "/" || $host/opsdirId || ".json",
                             $obj, $perms, $COLL)
    else
      ()
};

declare function j:databases-table(
  $qhost as object-node()?
)
{
  let $_ := trace("Computing databases joins", $t:trace-joins)
  for $cid in raw:resource-ids("clusters")
  let $cluster   := raw:resources("clusters", $cid)
  let $cluster-name := $cluster/opsdirName
  let $current   := cts:search(collection(),
                       cts:and-query((
                          cts:directory-query("/joins/databases/", "infinity"),
                          cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))
                      /opsdirId
  let $databases := cts:search(collection($t:COLLROOT||"resource"),
                      cts:and-query((
                        cts:json-property-value-query
                           ("opsdirResourceType", "databases", "exact"),
                        cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))
                      /object-node()
  let $update  := $databases/opsdirId/string()

  let $_       := j:delete("/joins/databases/", $cid, $current[not(. = $update)])

  let $forests := raw:resources-of-type($cid, "forests")
  let $hosts   := raw:resources-of-type($cid, "hosts")

  for $database in j:limit("/joins/databases/", $cid, $databases)
  let $details := map:map()
  let $fids    := $forests[opsdirName = $database/forest]/opsdirId
  let $fstats  := $fids ! raw:resource($cid, "forest-status", .)
  let $stands  := $fstats/statusProperties/stand[standKind/value = 'Active']
  let $disksp  := sum($stands/diskSize/value)

  let $fhosts  := $forests[opsdirId = $fids]/hostName
  let $fhids   := $hosts[opsdirName = $fhosts]/opsdirId

  let $counts  := raw:resource($cid, "counts", $database/opsdirId)
  let $dCount  := $counts/countProperties/documents/value

  let $status  := raw:resource($cid, "database-status", $database/opsdirId)

  let $flexrep := (string($status/statusProperties/flexibleReplicationEnabled/value) = "true")

  let $sprop   := $status/statusProperties
  let $ldf := (string($sprop/localDiskFailover/localDiskFailoverActive/value) = "true")
  let $sdf := (string($sprop/sharedDiskFailover/sharedDiskFailoverActive/value) = "true")

  let $lastfull := $status/statusProperties/lastBackup
  let $lastincr := $status/statusProperties/lastIncrBackup
  let $last     := if ($lastfull and $lastincr)
                   then if ($lastfull/value gt $lastincr/value)
                        then $lastfull/value
                        else $lastincr/value
                   else ($lastfull/value, $lastincr/value)
  let $lasttype := if ($lastfull/value = $last)
                   then "full"
                   else "incr"

  let $lasttype := if (exists($last)) then string($lasttype) else null-node { }
  let $last     := if (exists($last)) then string($last) else null-node { }

  let $repl     := $database/databaseReplication
  let $repl     := if (exists($repl/foreignMaster))
                   then "Replica"
                   else if (exists($repl/foreignReplica))
                        then "Master"
                        else "No"

  let $dbname   := string($database/opsdirName)
  let $servers  := cts:search(collection($t:COLLROOT||"resource"),
                     cts:and-query((
                       cts:json-property-value-query("opsdirClusterId", $cid, "exact"),
                       cts:json-property-value-query("opsdirResourceType", "servers", "exact"),
                       cts:or-query((
                         cts:json-property-value-query("contentDatabase", $dbname, "exact"),
                         cts:json-property-value-query("lastLoginDatabase", $dbname, "exact"),
                         cts:json-property-value-query("modulesDatabase", $dbname, "exact"))))))
                     /object-node()

  let $sinfo   := map:new()
  let $_       := if (exists($counts))
                  then map:put($sinfo, "countProperties", $counts/countProperties)
                  else ()
  let $_       := if (exists($status))
                  then map:put($sinfo, "statusProperties", $status/statusProperties)
                  else ()
  let $_       := if (map:keys($sinfo))
                  then map:put($details, "statusInfo", $sinfo)
                  else ()

  let $obj     := xdmp:to-json($details
                    + object-node {
                        "ts": string(current-dateTime()),
                        "opsdirResourceType": "databases",
                        "opsdirId": string($database/opsdirId),
                        "opsdirName": string($database/opsdirName),
                        "opsdirClusterId": $cid,
                        "opsdirClusterName": $cluster-name,
                        "forestNames": array-node { $database/forest },
                        "forestIds": array-node { $fids },
                        "hostNames": array-node { distinct-values($fhosts) },
                        "hostIds": array-node { distinct-values($fhids) },
                        "forestCount": string(count($database/forest)),
                        "diskSize": string($disksp),
                        "documentCount": string($dCount),
                        "lastBackup": $last,
                        "lastBackupType": $lasttype,
                        "encryption": string($database/dataEncryption),
                        "ha": if ($ldf or $sdf) then "Yes" else "No",
                        "databaseReplication": $repl,
                        "flexibleReplication": if ($flexrep) then "Yes" else "No",
                        "serverIds": array-node { $servers ! string(./opsdirId) },
                        "serverNames": array-node { $servers ! string(./opsdirName) },
                        "securityDatabase": string($database/securityDatabase),
                        "schemaDatabase": string($database/schemaDatabase),
                        "triggersDatabase": string($database/triggersDatabase)
                     } )

  let $perms := rs:effective-resource-permissions($database)
  return
    if (empty($qhost) or $qhost/opsdirClusterId = $cid)
    then
      let $_ := trace("Store /joins/databases/" || $cid || "/" || $database/opsdirId
                      || ".json " || $database/opsdirName,
                      $t:trace-joins-uri)
      return
        xdmp:document-insert("/joins/databases/" || $cid || "/" || $database/opsdirId || ".json",
                             $obj, $perms, $COLL)
    else
      ()
};

declare function j:tasks-table(
  $qhost as object-node()?
)
{
  let $_ := trace("Computing tasks joins", $t:trace-joins)
  for $cid in raw:resource-ids("clusters")
  let $cluster := raw:resources("clusters", $cid)
  let $cluster-name := $cluster/opsdirName
  let $current := cts:search(collection(),
                    cts:and-query((
                      cts:directory-query("/joins/tasks/", "infinity"),
                      cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))
                    /opsdirId
  let $tasks   := cts:search(collection($t:COLLROOT||"resource"),
                    cts:and-query((
                      cts:json-property-value-query("opsdirResourceType", "tasks", "exact"),
                      cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))
                    /object-node()
  let $update  := $tasks/taskId
  let $_       := j:delete("/joins/tasks/", $cid, $current[not(. = $update)])

  let $hosts   := raw:resources-of-type($cid, "hosts")
  let $dbs     := raw:resources-of-type($cid, "databases")

  for $task in j:limit("/joins/tasks/", $cid, $tasks)
  let $taskHostNames := j:task-hosts($task)
  let $taskHosts := $hosts[opsdirName = $taskHostNames]
  let $databases := ($task/taskDatabase,
                     if ($task/taskModules = "")
                     then ()
                     else $task/taskModules)
  let $groupId := raw:resource-id($cid, "groups", $task/opsdirGroupName)
  let $obj := object-node {
        "ts": string(current-dateTime()),
        "opsdirResourceType": "tasks",
        "opsdirId": string($task/taskId),
        "opsdirClusterId": $cid,
        "opsdirClusterName": $cluster-name,
        "groupId": $groupId,
        "groupName": $task/opsdirGroupName,
        "taskHost": $task/taskHost,
        "hostNames": array-node { $taskHostNames },
        "hostIds": array-node { $taskHosts/opsdirId },
        "databaseNames": array-node { $databases },
        "databaseIds": array-node { $dbs[opsdirName = $databases]/opsdirId },
        (: explicitly a string because we want to index it for sorting :)
        "enabled": string($task/taskEnabled),
        "root": string($task/taskRoot),
        "path": string($task/taskPath),
        "type": string($task/taskType),
        "period": $task/taskPeriod,
        "created": string($task/taskTimestamp),
        "contentDatabase": string($task/taskDatabase),
        "modulesDatabase": if ($task/taskModules = "")
                       then "(filesystem)"
                       else string($task/taskModules),
        "user": string($task/taskUser),
        "priority": string($task/taskPriority)
    }

  (: The task files have to have the permissions of the corresponding host :)
  let $perms := rs:effective-resource-permissions($taskHosts[1])

  return
    if (empty($qhost) or $qhost/opsdirClusterId = $cid)
    then
      let $_ := trace("Store /joins/tasks/" || $cid || "/" || $task/taskId
                      || ".json " || $task/taskId, $t:trace-joins-uri)
      return
        xdmp:document-insert("/joins/tasks/" || $cid || "/" || $task/taskId || ".json",
                             $obj, $perms, $COLL)
    else
      ()
};

declare function j:servers-table(
  $qhost as object-node()?
)
{
  let $_ := trace("Computing servers joins", $t:trace-joins)
  for $cid in raw:resource-ids("clusters")
  let $cluster := raw:resources("clusters", $cid)
  let $cluster-name := $cluster/opsdirName
  let $current := cts:search(collection(),
                  cts:and-query((
                  cts:directory-query("/joins/servers/", "infinity"),
                  cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))/opsdirId
  let $servers := cts:search(collection($t:COLLROOT||"resource"),
                  cts:and-query((
                  cts:json-property-value-query("opsdirResourceType", "servers", "exact"),
                  cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))/object-node()
  let $update  := $servers/opsdirId
  let $_       := j:delete("/joins/servers/", $cid, $current[not(. = $update)])

  let $groups  := raw:resources-of-type($cid, "groups")
  let $hosts   := raw:resources-of-type($cid, "hosts")
  let $dbs     := raw:resources-of-type($cid, "databases")

  for $server in j:limit("/joins/servers/", $cid, $servers)
  let $details := map:map()
  let $groups  := $groups[opsdirName = $server/opsdirGroupName]
  let $hosts   := $hosts[opsdirGroupName = $server/opsdirGroupName]
  let $dbnames := distinct-values(($server/contentDatabase, $server/lastLoginDatabase,
                                   $server/modulesDatabase))
  let $dbids   := $dbs[opsdirName = $dbnames]/opsdirId

  (: For reasons unknown, webdav servers have an empty content database
     and use the modules database for content. That's confusing so we'
     "fix" that here. :)
  let $contDB  := if ($server/serverType = "webdav")
                  then string($server/modulesDatabase)
                  else string($server/contentDatabase)
  let $modDB   := if ($server/serverType = "webdav")
                  then ""
                  else if (empty($server/modulesDatabase))
                       then "(file system)"
                       else string($server/modulesDatabase)

  let $status  := raw:resource($cid, "server-status", $server/opsdirId)
  let $sinfo   := map:new()
  let $_       := if (exists($status))
                  then map:put($sinfo, "statusProperties", $status/statusProperties)
                  else ()
  let $_       := if (map:keys($sinfo))
                  then map:put($details, "statusInfo", $sinfo)
                  else ()

  let $obj     := xdmp:to-json($details
                    + object-node {
                        "ts": string(current-dateTime()),
                        "opsdirResourceType": "servers",
                        "opsdirName": $server/opsdirName,
                        "opsdirId": $server/opsdirId,
                        "opsdirClusterName": $cluster-name,
                        "opsdirClusterId": $cid,
                        "serverType": $server/serverType,
                        "contentDatabase": $contDB,
                        "port": $server/port,
                        "ssl": exists($server/sslCertificateTemplate),
                        "groupId": $groups/opsdirId,
                        "groupName": $server/opsdirGroupName,
                        "hostIds": array-node { $hosts ! string(./opsdirId) },
                        "hostNames": array-node { $hosts ! string(./opsdirName) },
                        "databaseIds": array-node { $dbids },
                        "databasenames": array-node { $dbnames },
                        "modulesDatabase": $modDB,
                        "root": $server/root,
                        "security": if (string($server/internalSecurity) = "true")
                                    then "internal" else "external"
                      } )

  let $perms := rs:effective-resource-permissions($server)
  return
    if (empty($qhost) or $qhost/opsdirClusterId = $cid)
    then
      let $_     := trace("Store /joins/servers/" || $cid || "/" || $server/opsdirId
                          || ".json " || $server/opsdirName,
                          $t:trace-joins-uri)
      return
        xdmp:document-insert("/joins/servers/" || $cid || "/" || $server/opsdirId || ".json",
                             $obj, $perms, $COLL)
    else
      ()
};

declare function j:forests-table(
  $qhost as object-node()?
)
{
  let $_ := trace("Computing forests joins", $t:trace-joins)
  for $cid in raw:resource-ids("clusters")
  let $cluster := raw:resources("clusters", $cid)
  let $cluster-name := $cluster/opsdirName
  let $current := cts:search(collection(),
                  cts:and-query((
                  cts:directory-query("/joins/forests/", "infinity"),
                  cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))/opsdirId
  let $forests := cts:search(collection($t:COLLROOT||"resource"),
                  cts:and-query((
                  cts:json-property-value-query("opsdirResourceType", "forests", "exact"),
                  cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))/object-node()
  let $update  := $forests/opsdirId
  let $_       := j:delete("/joins/forests/", $cid, $current[not(. = $update)])

  let $dbs     := raw:resources-of-type($cid, "databases")

  for $forest in j:limit("/joins/forests/", $cid, $forests)
  let $details := map:map()

  let $_       := map:put($details, "fastDataMaxSize", $forest/fastDataMaxSize)
  let $_       := map:put($details, "rebalancerEnable", $forest/rebalancerEnable)
  let $_       := map:put($details, "range", $forest/range)
  let $_       := map:put($details, "databaseReplication", $forest/databaseReplication)

  let $counts  := raw:resource($cid, "counts", $forest/opsdirId)/countProperties
  let $_       := for $key in $counts/*
                  return
                    map:put($details, local-name-from-QName(node-name($key)), $key/value)

  let $storage := raw:resource($cid, "forest-storage", $forest/opsdirId)
  let $_       := if ($storage//remainingSpace)
                  then map:put($details, "freeSpace", sum($storage//remainingSpace/value))
                  else ()
  let $_       := if ($storage//forestSize)
                  then map:put($details, "size", sum($storage//forestSize/value))
                  else ()
  let $_       := if ($storage//forestReserve)
                  then map:put($details, "forestReserve", sum($storage//forestReserve/value))
                  else ()
  let $_       := if ($storage//deviceSpace)
                  then map:put($details, "deviceSpace", sum($storage//deviceSpace/value))
                  else ()
  let $_       := if ($storage//capacity)
                  then map:put($details, "capacity", sum($storage//capacity/value))
                  else ()
  let $_       := if ($storage//largeDataSize)
                  then map:put($details, "largeDataSize", sum($storage//largeDataSize/value))
                  else ()
  let $_       := if (string($forest/fastDataDirectory) != "")
                                                             (: Really!? :)
                  then map:put($details, "fastDataSize", sum($storage//forestSize/value))
                  else ()

  let $status  := raw:resource($cid, "forest-status", $forest/opsdirId)/statusProperties

  let $_       := map:put($details, "state", $status/state/value)
  let $_       := map:put($details, "updatesAllowed", $status/updatesAllowed/value)
  let $_       := map:put($details, "dataDirectory", $status/dataDir)
  let $_       := map:put($details, "largeDataDirectory", $status/largeDataDir)
  let $_       := map:put($details, "fastDataDirectory", $status/fastDataDir)
  let $_       := map:put($details, "encryptionEnabled", $status/encryption/value)
  let $_       := map:put($details, "standCount", count($status/stand))
  let $_       := map:put($details, "encryptedDiskSize",
                          sum($status/stand/encryptedDiskSize/value))


  let $_       := if ($forest/failoverEnable)
                  then if ($status/replicaForests)
                       then (map:put($details, "localDiskFailoverActive", true()),
                             map:put($details, "failoverMasterReplica", "master"))
                       else map:put($details, "localDiskFailoverActive", false())
                  else ()

  let $_       := if ($status/masterForest != $forest/opdirId)
                  then (map:put($details, "localDiskFailoverActive", true()),
                        map:put($details, "failoverMasterReplica", "replica"))
                  else map:put($details, "localDiskFailoverActive", false())


  let $_       := if ($status/sharedDiskFailoverActive/value)
                  then map:put($details, "sharedDiskFailoverActive",
                               $status/sharedDiskFailoverActive/value)
                  else map:put($details, "sharedDiskFailoverActive", false())

  let $sinfo   := map:new()
  let $_       := if (exists($counts))
                  then map:put($sinfo, "countProperties", $counts)
                  else ()
  let $_       := if (exists($status))
                  then map:put($sinfo, "statusProperties", $status)
                  else ()
  let $_       := if (exists($status))
                  then map:put($sinfo, "storageListItems", $storage/storageListItems)
                  else ()
  let $_       := if (map:keys($sinfo))
                  then map:put($details, "statusInfo", $sinfo)
                  else ()

  let $obj     := xdmp:to-json($details
                    + object-node {
                        "ts": string(current-dateTime()),
                        "opsdirResourceType": "forests",
                        "opsdirName": $forest/opsdirName,
                        "opsdirId": $forest/opsdirId,
                        "opsdirClusterName": $cluster-name,
                        "opsdirClusterId": $cid,
                        "hostNames": array-node { $forest/hostName },
                        "hostIds": array-node { raw:resource-id($cid, "hosts", $forest/hostName) },
                        "databaseNames": array-node { $dbs[forest = $forest/opsdirName]/opsdirName },
                        "databaseIds": array-node { $dbs[forest = $forest/opsdirName]/opsdirId },
                        "status": if (string($forest/enabled) = "true")
                                  then "enabled" else "disabled",
                        "availability": $forest/availability,
                        "failoverEnabled": $forest/failoverEnable })/object-node()

  (: At the moment, forests aren't first-class in Ops Director.
     Consequently, we store them with the permissions of the database
     that contains them. :)
  let $perms := if (empty($dbs[forest = $forest/opsdirName]))
                then $t:DBPERMS
                else rs:effective-resource-permissions(($dbs[forest = $forest/opsdirName])[1])
  return
    if (empty($qhost) or $qhost/opsdirClusterId = $cid)
    then
      let $_     := trace("Store /joins/forests/" || $cid || "/" || $forest/opsdirId
                          || ".json " || $forest/opsdirName,
                          $t:trace-joins-uri)
      return
        xdmp:document-insert("/joins/forests/" || $cid || "/" || $forest/opsdirId || ".json",
                             $obj, $perms, $COLL)
    else
      ()
};

declare function j:groups-table(
  $qhost as object-node()?
)
{
  let $_ := trace("Computing groups joins", $t:trace-joins)
  for $cid in raw:resource-ids("clusters")
  let $cluster   := raw:resources("clusters", $cid)
  let $cluster-name := $cluster/opsdirName
  let $current   := cts:search(collection(),
                       cts:and-query((
                          cts:directory-query("/joins/groups/", "infinity"),
                          cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))/opsdirId
  let $groups    := cts:search(collection($t:COLLROOT||"resource"),
                      cts:and-query((
                        cts:json-property-value-query("opsdirResourceType", "groups", "exact"),
                        cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))/object-node()
  let $update  := $groups/opsdirId/string()
  let $_       := j:delete("/joins/groups/", $cid, $current[not(. = $update)])

  let $servers := raw:resources-of-type($cid, "servers")
  let $hosts   := raw:resources-of-type($cid, "hosts")

  for $group in j:limit("/joins/groups/", $cid, $groups)

  let $hosts   := $hosts[opsdirGroupName = $group/opsdirName]
  let $servers := $servers[opsdirGroupName = $group/opsdirName]

  let $obj     := object-node {
                        "ts": string(current-dateTime()),
                        "opsdirResourceType": "groups",
                        "opsdirId": string($group/opsdirId),
                        "opsdirName": string($group/opsdirName),
                        "opsdirClusterId": $cid,
                        "opsdirClusterName": $cluster-name,
                        "hostIds": array-node { $hosts ! ./opsdirId },
                        "hostNames": array-node { $hosts ! ./opsdirName },
                        "hostCount": count($hosts),
                        "serverIds": array-node { $servers ! ./opsdirId },
                        "serverNames": array-node { $servers ! ./opsdirName },
                        "serverCount": count($servers)
                     }

  let $perms := rs:effective-resource-permissions($group)
  return
    if (empty($qhost) or $qhost/opsdirClusterId = $cid)
    then
      let $_ := trace("Store /joins/groups/" || $cid || "/" || $group/opsdirId
                      || ".json " || $group/opsdirName, $t:trace-joins-uri)
      return
        xdmp:document-insert("/joins/groups/" || $cid || "/" || $group/opsdirId || ".json",
                             $obj, $perms, $COLL)
    else
      ()
};

declare function j:update-joins(
) as empty-sequence()
{
  j:update-joins(())
};

declare function j:update-joins(
  $host as object-node()?
) as empty-sequence()
{
  (: Remove all join artifacts for any cluster that's been removed :)
  let $cids := raw:resource-ids("clusters")
  let $leftover := cts:uris((), ("any"),
                     cts:and-query((
                       cts:directory-query("/joins/", "infinity"),
                       cts:not-query(
                         cts:json-property-value-query("opsdirClusterId", $cids, "exact")))))
  let $_ := for $uri in $leftover
            let $_ := trace("Deleting " || $uri, $t:trace-joins)
            return
              xdmp:document-delete($uri)
  return
    if (j:healed-dup-resources())
    then xdmp:log("Healed duplicate resources...")
    else j:do-update-joins($host)
};

declare %private function j:healed-dup-resources(
) as xs:boolean
{
  let $healed := for $cid in raw:resource-ids("clusters")
                 for $kind in cts:values(cts:json-property-reference("opsdirResourceType"))
                 let $dupids := cts:values(cts:json-property-reference("opsdirId"), (), (),
                    cts:and-query((
                       cts:collection-query($t:COLLROOT||"resource"),
                       cts:json-property-value-query("opsdirResourceType", $kind, "exact"),
                       cts:json-property-value-query("opsdirClusterId", $cid, "exact")))
                 )[cts:frequency(.) gt 1]
                 where $dupids
                 return
                   j:heal-dup-resources($cid, $kind, $dupids)
  return
    exists($healed)
};

declare %private function j:heal-dup-resources(
  $cid as xs:string,
  $kind as xs:string,
  $ids as xs:string+
) as xs:string+
{
  for $id in $ids
  let $uris := cts:uris((), ("descending"),
                    cts:and-query((
                       cts:collection-query($t:COLLROOT||"resource"),
                       cts:json-property-value-query("opsdirResourceType", $kind, "exact"),
                       cts:json-property-value-query("opsdirId", $id, "exact"),
                       cts:json-property-value-query("opsdirClusterId", $cid, "exact"))))
  let $_    := for $uri in subsequence($uris, 2)
               let $_ := xdmp:log("Removing duplicate resource: " || $uri)
               return
                 xdmp:document-set-collections($uri, ())
  return
    $ids
};

declare %private function j:do-update-joins(
  $host as object-node()?
) as empty-sequence()
{
  let $txn-name    := "opsdir-update-joins-"
                      || (if (exists($host)) then $host/opsdirClusterName else "all")
  let $concurrency := xs:unsignedLong(config:get-atomic-property("joins-concurrency", 2))
  let $time-limit  := xs:unsignedLong(config:get-atomic-property("joins-time-limit", 300))
  let $txnlist     := xdmp:transaction($txn-name)
  return
    if (count($txnlist) ge $concurrency)
    then
      let $_    := trace("Host " || xdmp:host-name()
                         || " update joins concurrency too high: "
                         || count($txnlist), $t:trace-polling)
      let $txns := xdmp:host-status(xdmp:host())/host:transactions
                      /host:transaction[host:transaction-name = $txn-name]
      for $txn in $txns
      let $time := current-dateTime() - $txn/host:start-time
      let $_    := trace("Txn " || $txn/host:transaction-id
                         || " polling update joins for " || $time, $t:trace-polling)
      return
        ()
    else
      xdmp:spawn-function(
        function () {
          j:do-update-joins-txn($host, $txn-name, $time-limit)
        },
        <options xmlns="xdmp:eval">
          <time-limit>{$time-limit}</time-limit>
          <update>true</update>
          <commit>auto</commit>
        </options>)
};

declare %private function j:do-update-joins-txn(
  $host as object-node()?,
  $txn-name as xs:string,
  $time-limit as xs:unsignedLong
) as empty-sequence()
{
  (: Force update access to be sequential; in order to check for deleted resources,
     each attempt to make joins does queries across many of the join documents. That
     can lead to deadlock if they're allowed to run in parallel.
  :)
  xdmp:lock-for-update($LOCK),

  xdmp:set-transaction-name($txn-name),
  j:clusters-table($host),
  j:groups-table($host),
  j:hosts-table($host),
  j:databases-table($host),
  j:tasks-table($host),
  j:servers-table($host),
  j:forests-table($host)
};

(: ============================================================ :)

declare %private function j:delete(
  $path as xs:string,
  $cid as xs:string,
  $ids as xs:string*
) as empty-sequence()
{
  for $id in $ids
  let $uri := $path || $cid || "/" || $id || ".json"
  let $_   := trace("Deleting " || $uri, $t:trace-joins)
  return
    xdmp:document-delete($uri)
};

declare %private function j:limit(
  $path as xs:string,
  $cid as xs:string,
  $configs as object-node()*
) as object-node()*
{
  let $force := xdmp:trace-enabled($t:trace-force-joins)
  for $obj in $configs
  let $id  := $obj/opsdirId
  let $uri := $path || $cid || "/" || $id || ".json"
  return
    if ($force or not(doc-available($uri)))
    then
      $obj
    else
      let $ts    := doc($uri)/object-node()/ts
      let $query := cts:and-query((
                       cts:json-property-value-query("opsdirId", $id, "exact"),
                       cts:json-property-value-query("opsdirClusterId", $cid, "exact"),
                       cts:json-property-value-query("opsdirKind", ("resource", "stats"), "exact"),
                       cts:json-property-range-query("ts", ">", $ts)))
      let $other := xdmp:estimate(cts:search(collection(), $query))

      let $_     := trace("Update join? " || $uri || " (" || $other || ")",
                          $t:trace-debug-joins)
      let $_     := if (xdmp:trace-enabled($t:trace-debug-joins))
                    then
                      for $doc in cts:search(collection(), $query)
                      return
                        trace("   " || xdmp:node-uri($doc), $t:trace-debug-joins)
                    else
                      ()

      let $upd   := empty($ts) or ($other gt 0)
      let $_     := if ($upd)
                    then trace("Update " || $uri, $t:trace-debug-joins)
                    else trace("Skip   " || $uri, $t:trace-debug-joins)
      where $upd
      return
        $obj
};

declare %private function j:host-details(
  $cid as xs:string,
  $hosts as xs:string+
) as map:map
{
  let $map := map:new()
  let $_   := for $host in $hosts
              let $hostinfo := raw:resource($cid, "host-status", $hosts[1])
              let $security := raw:resource($cid, "security", $cid)
              let $detail := $hostinfo/statusProperties/statusDetail

              let $_   := if ($detail/version)
                          then map:put($map, "version",
                                       (map:get($map, "version"),
                                        string($detail/version)))
                          else
                            ()
              let $_   := if ($detail/effectiveVersion)
                          then map:put($map, "effectiveVersion",
                                       (map:get($map, "effectiveVersion"),
                                        string($detail/effectiveVersion/value)))
                          else
                            ()
              let $_   := if ($detail/lastStartup)
                          then
                            let $dtstart := xs:dateTime($detail/lastStartup/value)
                            let $uptime  := current-dateTime() - $dtstart
                            return
                              (map:put($map, "lastStartup",
                                       (map:get($map, "lastStartup"),
                                        string($detail/lastStartup/value))),
                               map:put($map, "uptime",
                                       (map:get($map, "uptime"),
                                        string($uptime))))
                          else
                            ()
              let $_   := if ($detail/osVersion)
                          then map:put($map, "osVersion",
                                       (map:get($map, "osVersion"),
                                        string($detail/osVersion)))
                          else
                            ()
              let $_   := if ($security/keystore/dataEncryption)
                          then map:put($map, "encryption",
                                       (map:get($map, "encryption"),
                                        string($security/keystore/dataEncryption)))
                          else
                            ()
              return
                ()

  (: Some values won't be known until the scheduled tasks have queried
     for them; stick in "Unknown" in the meantime so that the keys
     have some value.
  :)
  let $_   := if (map:contains($map, "version"))
              then ()
              else map:put($map, "version", "Unknown")

  let $_   := if (map:contains($map, "effectiveVersion"))
              then ()
              else map:put($map, "effectiveVersion", "Unknown")

  let $_   := if (map:contains($map, "osVersion"))
              then ()
              else map:put($map, "osVersion", "Unknown")

  let $_   := if (map:contains($map, "encryption"))
              then ()
              else map:put($map, "encryption", "Unknown")
  return
    $map
};

declare %private function j:task-hosts(
  $task as object-node()
) as xs:string*
{
  if (exists($task/null-node("taskHost")))
  then
    let $group := $task/opsdirGroupName
    let $hosts := raw:resources-of-type($task/opsdirClusterId, "hosts")
    for $host in $hosts
    where $host/opsdirGroupName = $group
    return
      string($host/opsdirName)
  else
    string($task/taskHost)
};

declare function j:poll-update-joins(
)
{
  p:poll-clusters("joins", j:poll-update#1)
};

declare function j:poll-update(
  $host as object-node()
) as empty-sequence()
{
  j:update-joins($host)
};
