xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

(: Logs: forest states;
         server startup/restart;
         look for "Database"?

2015-04-12 01:53:35.443 Warning: Forest XXX is at 92% of stand limit

:)

(: Make it possible for host startup messages to close whole sets of messages? :)

module namespace la="http://marklogic.com/v1/opsdirector/log/alerts";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes"
    at "/common/rscopes.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $SERVER-FIELD-NAME := "opsdir-logalerts-map";
declare variable $CACHE-DURATION := xs:dayTimeDuration("PT1M");

declare function la:augment(
  $log as object-node()
) as object-node()
{
  let $_    := trace(upper-case($log/level) || " " || $log/message,
                     $t:trace-log-filter-debug)
  let $map  := la:get-map()
  let $augs := map:get($map, $log/level)
  return
    if (empty($augs))
    then
      let $_ := trace("No augs for " || $log/level || " level", $t:trace-log-filter-debug)
      return
        $log
    else
      (: Could we use $1 to find the aug to avoid the second call to matches()? :)
      let $regex := string-join($augs ! map:get(., "regex") ! concat("(", ., ")"), "|")
      return
        if (matches($log/message, $regex))
        then
          let $augs  := for $aug in $augs
                        where matches($log/message, map:get($aug, "regex"))
                        return
                          $aug
          return
            if (exists($augs))
            then
              la:augment-with-augs($log, $augs)
            else
              let $_ := trace("No augs for " || $log/level || " " || $log/message,
                              $t:trace-log-filter-debug)
              return
                $log
        else
          let $_ := trace("No match: " || $log/message, $t:trace-log-filter-debug)
          return
            $log
};

declare %private function la:get-map(
) as map:map
{
  (: In the case where a system has been updated from Ops Director 1.0 to 1.1,
     the xdmp:get/set server field privileges may be missing from the
     opsdir-data-internal role; so just do it the slow way.
  :)
  let $map := try {
                xdmp:get-server-field($SERVER-FIELD-NAME)
              } catch ($e) {
                if ($e/error:code eq 'SEC-PRIV')
                then
                  ()
                else
                  xdmp:rethrow()
              }
  return
    if (empty($map)
        or (current-dateTime() - map:get($map, "timestamp") > $CACHE-DURATION))
    then
      let $_   := trace("Building server field from sources", $t:trace-log-filter)
      let $map := la:build-map-from-sources()
      let $_   := try {
                    xdmp:set-server-field($SERVER-FIELD-NAME, $map)
                  } catch ($e) {
                    ()
                  }
      return
        $map
    else
      $map
};

declare function la:build-map-from-sources(
) as map:map
{
  let $map  := map:new(map:entry("timestamp", current-dateTime()))
  let $augs := collection($t:LOG-AUGMENTERS)//la:augment
  let $_    := trace("Loaded " || count($augs) || " log alert augmenters",
                     $t:trace-log-filter)
  let $_    := for $aug in $augs
               let $level := $aug/@level/string()
               let $regex := $aug/@regex/string()
               let $_     := if (empty($level) or empty($regex))
                             then
                               xdmp:log(("Error: augmenting a log requires level and regex",
                                         $aug))
                             else
                               ()
               where exists($level) and exists($regex)
              return
                let $augmap := map:map()
                let $pmap   := map:map()
                let $_      := map:put($augmap, "regex", $regex)
                let $_      := for $elem in $aug/la:*
                               let $map := map:new((map:entry("value", string($elem)),
                                                    if ($elem/@as)
                                                    then map:entry("type", string($elem/@as))
                                                    else ()))
                               return
                                 map:put($pmap, local-name($elem), $map)
                let $_      := if ($aug/la:alertCode)
                               then
                                 let $type := $aug/la:resourceType/string()
                                 let $name := $aug/la:resourceName/string()
                                 return
                                   if (empty($type) or empty($name))
                                   then
                                     (map:delete($pmap, "alertCode"),
                                      xdmp:log(("Error: augmented log with alertCode "
                                                || "requires resource type and name", $aug)))
                                   else
                                     ()
                               else
                                 ()
                let $_      := map:put($augmap, "props", $pmap)
                return
                  if (map:contains($map, $level))
                  then
                    map:put($map, $level, (map:get($map, $level), $augmap))
                  else
                    map:put($map, $level, $augmap)
  return
    $map
};

declare %private function la:augment-with-augs(
    $log as object-node(),
    $augs as map:map+
) as object-node()
{
  let $props := $augs ! la:props($log/message, .)
  let $map   := map:map()
  let $_     := for $name in map:keys($props)
                return
                  la:update-map($map, $log, $name, $props)
  let $alert := if (map:contains($map, "alertCode")
                    and (map:contains($map, "resourceType"))
                    and (map:contains($map, "resourceName")))
                then
                  (: The log messages generated for debugging have a distracting
                     event:id at the beginning; just remove that for now. :)
                  let $msg    := string($log/message)
                  let $msg    := if (starts-with($msg, "[Event:id=Telemetry Log] "))
                                 then substring-after($msg, "] ")
                                 else $msg
                  let $rtype  := map:get($map, "resourceType")
                  let $rid    := map:get($map, "resourceId")
                  let $rname  := map:get($map, "resourceName")
                  let $state  := (map:get($map, "state"), "info")[1]
                  let $status := if (map:contains($map, "opens"))
                                 then "open"
                                 else "reported"
                  return
                    map:new((
                      map:entry("acknowledged", "no"),
                      map:entry("code", string(map:get($map, "alertCode"))),
                      map:entry("duration", "PT0S"),
                      map:entry("logId", string($log/uid)),
                      map:entry("fromLog", true()),
                      map:entry("local", true()),
                      map:entry("message", $msg),
                      map:entry("opsdirClusterId", string($log/opsdirClusterId)),
                      map:entry("opsdirClusterName", string($log/opsdirClusterName)),
                      map:entry("opsdirDtstart", string(current-dateTime())),
                      map:entry("opsdirHostId", string($log/opsdirHostId)),
                      map:entry("opsdirId", concat(string($log/uid), '-', xdmp:random())),
                      map:entry("resourceId", string($rid)),
                      map:entry("resourceName", string($rname)),
                      map:entry("resourceType", $rtype),
                      map:entry("opsdirResourceType", "alerts"),
                      map:entry("state", $state),
                      map:entry("logLevel", string($log/level)),
                      map:entry("status", $status),
                      map:entry("ts", string($log/ts)),
                      map:entry("uid", string(xdmp:random()))
                    ))
                else
                  ()
  let $alert := if (map:contains($map, "opens"))
                then
                  (: Make sure open/close messages include the HostId :)
                  let $msg := $log/opsdirHostId || " " || map:get($map, "opens")
                  return
                    $alert + map:entry("logMsgId", xdmp:sha1($msg, "hex"))
                else if (map:contains($map, "closes"))
                     then
                       (: Make sure open/close messages include the HostId :)
                       let $msg := $log/opsdirHostId || " " || map:get($map, "closes")
                       return
                         $alert + map:entry("closesMsgId", xdmp:sha1($msg))
                     else
                       $alert
  let $_ := trace("Aug " || $log/uid || ": " || $log/message, $t:trace-log-filter)
  let $_ := if (empty($alert))
            then ()
            else la:process-alert(xdmp:to-json($alert)/object-node())
  let $_ := map:delete($map, "opens")
  let $_ := map:delete($map, "closes")
  return
    xdmp:to-json($log + $map)/object-node()
};

declare %private function la:props(
  $message as xs:string,
  $augment as map:map
) as map:map
{
  let $replmap := map:map()
  let $pmap    := map:get($augment, "props")
  let $_       := for $name in map:keys($pmap)
                  let $prop := map:get($pmap, $name)
                  return
                    map:put($replmap, $name, map:new((
                             if (map:contains($prop, "type"))
                             then map:entry("type", map:get($prop, "type"))
                             else (),
                             map:entry("value",
                                       replace($message,
                                               map:get($augment, "regex"),
                                               map:get($prop, "value"))))))
  return
    $replmap
};

declare %private function la:value(
  $prop as map:map
) as xs:anyAtomicType
{
  let $type := map:get($prop, "type")
  let $value := map:get($prop, "value")
  return
    if (exists($type))
    then
      switch ($type)
      case "timestamp" return xdmp:timestamp-to-wallclock(xs:unsignedLong($value))
      default return string($value)
    else
      $value
};

declare %private function la:update-map(
  $map as map:map,
  $log as object-node(),
  $name as xs:string,
  $props as map:map
) as empty-sequence()
{
  let $value := la:value(map:get($props, $name))
  let $_     := map:put($map, $name, $value)
  return
    if ($name = "resourceName" and map:contains($props, "resourceType"))
    then
      let $rtype := map:get(map:get($props, "resourceType"), "value")
      let $id := la:get-resource-id($log/opsdirClusterId, $log/opsdirHostId,
                                    $rtype || "s", $value)
      where count($id) = 1
      return
        map:put($map, "resourceId", $id)
    else
      ()
};

declare %private function la:get-resource-id(
  $clusterId as xs:string,
  $hostId as xs:string,
  $rtype as xs:string,
  $value as xs:string
) as xs:string?
{
  let $id := t:managed-resource-id($clusterId, $rtype, $value)
  return
    if (count($id) gt 1)
    then
      (: This can happen for servers that appear in multiple groups.
         Find the one that's running on the host that sent the log message.
      :)
      for $rsrc in t:managed-resources($id)
      where $hostId = $rsrc/hostIds
      return
        $rsrc/opsdirId
    else
      $id
};

declare %private function la:process-alert(
  $alert as object-node()
) as empty-sequence()
{
  if ($alert/closesMsgId and la:close-alert($alert))
  then
    ()
  else
    if ($alert/logMsgId and la:alert-exists($alert))
    then
      let $_ := trace("Skipping alert: " || $alert/logMsgId,
                      $t:trace-log-filter)
      return
        ()
    else
      let $_ := trace("Creating alert: " || $alert/logMsgId || ": " || $alert/message,
                      $t:trace-log-filter)
      let $cluster-id := $alert/opsdirClusterId
      let $host-id    := $alert/opsdirHostId
      let $timestamp  := $alert/ts

      let $coll  := ("http://marklogic.com/opsdirector/v1/alerts/" || $cluster-id,
                     "http://marklogic.com/opsdirector/v1/alerts")

      let $id    := $alert/uid
      let $uri   := "/alerts/" || $cluster-id || "/" || $id || ".json"

      let $p     := if ($alert/resourceId)
                    then
                      let $r := t:managed-resource((), (), $alert/resourceId)
                      return
                        if (exists($r))
                        then
                          rs:effective-resource-permissions($r)
                        else
                          $t:DBPERMS
                    else
                      $t:DBPERMS
      return
        xdmp:document-insert($uri, $alert, $p, $coll)
};

declare %private function la:close-alert(
  $alert as object-node()
) as xs:boolean
{
  let $cluster-id := $alert/opsdirClusterId
  let $timestamp  := $alert/ts
  let $msgId := $alert/closesMsgId
  let $coll  := "http://marklogic.com/opsdirector/v1/alerts/" || $cluster-id
  let $open  := cts:search(collection($coll),
                           cts:and-query((
                             cts:json-property-value-query("logMsgId", $msgId),
                             cts:json-property-value-query("status", "open"))))/object-node()
  return
    if (count($open) = 1)
    then
      let $_ := trace("Closing alert: " || $msgId, $t:trace-log-filter)
      (: Oh for the love of ... this is a hack :)
      let $map := map:new(for $key in $open/* except ($open/status)
                          return map:entry(
                                   local-name-from-QName(node-name($key)),
                                   $open/*[node-name(.) = node-name($key)]))
      let $duration := xs:dateTime($timestamp) - xs:dateTime($open/opsdirDtstart)
      let $obj := object-node { "opsdirDtend": $timestamp,
                                "duration": string($duration),
                                "status": "closed" }
      let $_ := xdmp:node-replace($open, xdmp:to-json($map + $obj)/node())
      return
        true()
    else
      let $_ := trace("Alerts with " || $msgId || ": " || count($open), $t:trace-log-filter)
      return
        false()
};

declare %private function la:alert-exists(
  $alert as object-node()
) as xs:boolean
{
  let $cluster-id := $alert/opsdirClusterId
  let $timestamp  := $alert/ts
  let $msgId := $alert/logMsgId
  let $coll  := "http://marklogic.com/opsdirector/v1/alerts/" || $cluster-id
  let $open  := xdmp:estimate(
                  cts:search(collection($coll),
                             cts:and-query((
                               cts:json-property-value-query("logMsgId", $msgId),
                               cts:json-property-value-query("status", "open")))))
  return
    $open gt 0
};
