xquery version "1.0-ml";

(: Copyright 2002-2018 MarkLogic Corporation.  All Rights Reserved. :)

(: Install-time library functions. :)

module namespace odi="http://marklogic.com/v1/opsdirector/install";

import module namespace admin="http://marklogic.com/xdmp/admin"
    at "/MarkLogic/admin.xqy";

import module namespace sec = "http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace pki="http://marklogic.com/xdmp/pki"
    at "/MarkLogic/pki.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

declare namespace x509="http://marklogic.com/xdmp/x509";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $odi:OPSDIRECTOR-DATABASE    := "OpsDirector";
declare variable $odi:OPSDIRECTOR-UI-SERVER   := "OpsDirectorApplication";
declare variable $odi:OPSDIRECTOR-DATA-SERVER := "OpsDirectorSystem";
declare variable $odi:OPSDIRECTOR-EXTERNAL-SECURITY := "OpsDirectorSystem";
declare variable $odi:OPSDIRECTOR-SYSTEM-USER := "opsdirector-system-user";
declare variable $odi:SECURE-MANAGEMENT-API   := "SecureManage";
declare variable $odi:OPSDIRECTOR-CERTIFICATE-TEMPLATE := "OpsDirector Template";
declare variable $odi:OPSDIRECTOR-CA-COMMON-NAME := "MarkLogic Ops Director Certificate Authority";
declare variable $odi:OPSDIRECTOR-SSL-TEMPLATE-NAME := "OpsDirector-SSL-Template";
declare variable $odi:OPSDIRECTOR-SSL-TEMPLATE-DESC := "OpsDirector SSL template for managed clusters";
declare variable $odi:OPSDIRECTOR-SSL-TEMPLATE-ORG  := "MarkLogic Corporation";
declare variable $odi:OPSDIRECTOR-SSL-TEMPLATE-DEPT := ();

declare variable $odi:OPSDIRECTOR-EXT-SEC-NAME := "OpsDirector-External-Security";
declare variable $odi:OPSDIRECTOR-EXT-SEC-DESC := "OpsDirector Certificate Authentication of Clients";

declare %private variable $iso-opts :=
  <options xmlns="xdmp:eval">
    <isolation>different-transaction</isolation>
    <transaction-mode>update-auto-commit</transaction-mode>
    <prevent-deadlocks>true</prevent-deadlocks>
  </options>;

declare function odi:managed-clusters(
  $opsdirdb as xs:unsignedLong
) as element()*
{
  let $objects
    := xdmp:invoke-function(
         function () {
           try {
             cts:search(collection("http://marklogic.com/opsdirector/v1/resource"),
                        cts:directory-query("/resource/clusters/",
                                            "infinity"))/object-node()
           } catch ($e) {
             object-node {
               "error": string($e/error:code),
               "message": string($e/error:message),
               "detail": xdmp:quote($e)
             }
           }
         },
         <options xmlns="xdmp:eval">
           <database>{$opsdirdb}</database>
         </options>)
  for $object in $objects
  let $id   := $object/opsdirId
  let $name := $object/opsdirName
  return
    if ($object/error)
    then
      <span class="error" xmlns="http://www.w3.org/1999/xhtml">
        { "ERROR: " || $object/*:message }
      </span>
    else
      <span class="cluster" xmlns="http://www.w3.org/1999/xhtml"
            data-cluster="{$id}">
        { $name }
      </span>
};

declare function odi:managed-hosts(
  $opsdirdb as xs:unsignedLong,
  $cluster-id as xs:string
) as xs:unsignedLong
{
  let $ids := xdmp:invoke-function(
                    function () {
                      t:managed-resource-ids($cluster-id, "hosts")
                    },
                    <options xmlns="xdmp:eval">
                      <database>{$opsdirdb}</database>
                    </options>)
  return
    count($ids)
};

declare function odi:opsdirector-ca-id(
) as xs:unsignedLong?
{
  try {
    xdmp:invoke-function(
      function () {
        let $id := doc("/config/certificate-authority.txt")/text()
        return
          if ($id eq "")
          then
            ()
          else
            $id cast as xs:unsignedLong
      },
      <options xmlns="xdmp:eval">
        <database>{xdmp:database($odi:OPSDIRECTOR-DATABASE)}</database>
      </options>)
  } catch ($e) {
    ()
  }
};

declare function odi:usable-certificate-authorities(
) as xs:unsignedLong*
{
  xdmp:invoke-function(
    function () {
      for $ca-id in pki:get-certificates(pki:get-trusted-certificate-ids())
                       [pki:authority = fn:true()]/pki:certificate-id
      where /pki:certificate-private-key[pki:certificate-id = $ca-id]
      return
        $ca-id
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>
  )
};

declare function odi:certificate-authority-private-key(
  $ca-id as xs:unsignedLong
) as xs:string
{
  xdmp:invoke-function(
    function () {
      /pki:certificate-private-key[pki:certificate-id = $ca-id]/pki:private-key
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>
  )
};

declare function odi:certificate-authority(
  $ca-id as xs:unsignedLong
) as element()?
{
  xdmp:invoke-function(
    function () {
      /pki:certificate[pki:certificate-id = $ca-id]
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>
  )
};

declare function odi:opsdirector-ca(
) as xs:string?
{
  let $id := odi:opsdirector-ca-id()
  where exists($id)
  return
    (pki:get-certificates($id)/x509:cert/x509:subject/x509:commonName)[1]
};

declare function odi:save-opsdir-ca-id(
  $id as xs:unsignedLong?
) as empty-sequence()
{
  xdmp:invoke-function(
    function () {
      xdmp:document-insert("/config/certificate-authority.txt",
                           text { string($id) }, $t:ADMIN-DBPERMS, ())
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:database($odi:OPSDIRECTOR-DATABASE)}</database>
    </options>)
};

declare function odi:external-security-certificate-authorities(
) as xs:unsignedLong*
{
  xdmp:invoke-function(
    function () {
      sec:external-security-get-ssl-client-certificate-authorities(
          $odi:OPSDIRECTOR-EXTERNAL-SECURITY)
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>
  )
};

declare function odi:teardown-external-security()
{
  xdmp:invoke-function(
    function () {
      sec:remove-external-security($odi:OPSDIRECTOR-EXTERNAL-SECURITY)
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>
  )
};

declare function odi:teardown-opsdir-admin-user()
{
  try {
    xdmp:invoke-function(
      function () {
        sec:remove-user($odi:OPSDIRECTOR-SYSTEM-USER)
      },
      <options xmlns="xdmp:eval">
        <database>{xdmp:security-database()}</database>
      </options>
    )
  } catch ($e) {
    if ($e/error:code = "SEC-USERDNE")
    then ()
    else xdmp:rethrow()
  }
};

declare function odi:teardown-cluster-config()
{
  xdmp:invoke-function(
    function () {
      let $config    := admin:get-configuration()
      let $config    := fold-left(
                          function ($config, $group-id) {
                            let $id := try {
                                         admin:appserver-get-id($config, $group-id,
                                                                $odi:SECURE-MANAGEMENT-API)
                                       } catch ($e) {
                                         ()
                                       }
                            return
                              if (exists($id))
                              then
                                admin:appserver-delete($config, $id)
                              else
                                $config
                          }, $config, admin:get-group-ids($config))
      let $config    := admin:cluster-set-opsdirector-session-endpoint($config, "")
      let $config    := admin:cluster-set-opsdirector-log-level($config, "disabled")
      let $config    := admin:cluster-set-opsdirector-metering($config, "disabled")
      let $config    := admin:cluster-set-opsdirector-config($config, "disabled")
      return
        admin:save-configuration-without-restart($config)
    }, $iso-opts
  )
};

declare function odi:uninstall-template()
{
  let $template := pki:get-template-by-name($odi:OPSDIRECTOR-CERTIFICATE-TEMPLATE)
  return
    pki:delete-template($template/pki:template-id)
};

declare function odi:externally-signed-certificate(
) as xs:boolean
{
  xdmp:invoke-function(
    function () {
      let $cred := sec:get-credential("MarkLogic-OpsDirector")
      let $cert := $cred/sec:credential-certificate
      return
        not(xdmp:x509-certificate-extract($cert)/x509:issuer/x509:commonName/string()
              = $odi:OPSDIRECTOR-CA-COMMON-NAME)
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>
  )
};

declare function odi:disconnect-cluster(
) as xs:boolean
{
  let $config     := admin:get-configuration()
  let $session    := admin:cluster-get-opsdirector-session-endpoint($config)
  let $disconnect := substring-before($session, '/v1/') || "/disconnect"

  let $hostname   := xdmp:host-name()
  let $cluster-id := string(xdmp:cluster())

  let $hash       := xdmp:hmac-sha256($t:SHARED-SECRET, $hostname||$cluster-id, "hex")
  let $body       := <payload>
                       <cluster-id>{$cluster-id}</cluster-id>
                       <hostname>{$hostname}</hostname>
                       <hash>{$hash}</hash>
                       <version>{xdmp:software-version()}</version>
                     </payload>
  let $opts       := <options xmlns="xdmp:http">
                       <headers>
                         <accept>application/xml</accept>
                       </headers>
                       <credential-id>
                         { xdmp:credential-id("MarkLogic-OpsDirector") }
                       </credential-id>
                       <verify-cert>false</verify-cert>
                       <timeout>30</timeout>
                     </options>
  let $resp       := try {
                       xdmp:http-post($disconnect, $opts, $body)
                     } catch ($e) {
                       ()
                     }
  return
    $resp[1]/*:code = 200 and string($resp[2]) = "ok"
};

declare function odi:find-management-api(
) as xs:unsignedLong*
{
  let $config := admin:get-configuration()
  let $found  := for $group in xdmp:groups()
                 for $id in admin:group-get-appserver-ids($config, $group)
                 return
                     if (admin:appserver-get-type($config,$id) eq 'http')
                     then let $rewriter := admin:appserver-get-url-rewriter($config, $id)
                          let $globally := admin:appserver-get-rewrite-resolves-globally($config, $id)
                          where $globally and $rewriter = "/MarkLogic/manage/rewriter.xml"
                          return
                              ($group, $id)
                     else ()
  return
    subsequence($found, 1, 2)
};

declare function odi:find-available-port(
) as xs:unsignedLong
{
  odi:find-available-port(8000, 1)
};

declare function odi:find-available-port(
  $pref as xs:unsignedLong,
  $limit as xs:unsignedLong
) as xs:unsignedLong*
{
  let $config := admin:get-configuration()
  let $ports  := for $group in xdmp:groups()
                 for $id in admin:group-get-appserver-ids($config, $group)
                 return
                   admin:appserver-get-port($config, $id)
  let $ports  := for $port in $ports
                 order by $port
                 return $port
  let $unused := fold-left(function ($available, $port) {
                             let $next := $port + 1
                             return
                               if (($next = $ports) or ($next lt 8000))
                               then
                                 $available
                               else
                                 ($available, $next)
                           }, (), ($pref - 1 to $pref + 100))
  return
    $unused[1 to $limit]
};
