xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

(: Note: This module has been written so that it can be shared between
   XQuery and JavaScript. Consequently it uses camelCase naming conventions
   and passes ID values around as strings instead of xs:unsignedLongs.
   Because Javascript. Also, the resource scope XML is not in any namespace.
   Because JSON.
:)

module namespace t="http://marklogic.com/v1/opsdirector";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace ts="http://marklogic.com/v1/opsdirector/security"
    at "/common/opsec.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $t:OPSDIRECTOR-VERSION := "2.0.0";

declare variable $t:CA-NAME := "OpsDirector-CA";
declare variable $t:CRED-NAME := "OpsDirector-Credential";

declare variable $t:DBPERMS :=
  (xdmp:permission("opsdir-admin", "read"),
   xdmp:permission("opsdir-admin", "update"),
   xdmp:permission("opsdir-data-internal", "read"),
   xdmp:permission("opsdir-data-internal", "update"));

declare variable $t:ADMIN-DBPERMS :=
  (xdmp:permission("opsdir-admin", "read"),
   xdmp:permission("opsdir-admin", "update"));

declare variable $t:COLLROOT := "http://marklogic.com/opsdirector/v1/";

declare variable $t:trace-status       := "opsdir-status";
declare variable $t:trace-upload       := "opsdir-upload";
declare variable $t:trace-debug        := "opsdir-debug";
declare variable $t:trace-debug-io     := "opsdir-debug-io";
declare variable $t:trace-debug-insert := "opsdir-insert";
declare variable $t:trace-debug-uri    := "opsdir-uri";
declare variable $t:trace-debug-perm   := "opsdir-debug-perm";
declare variable $t:trace-debug-priv   := "opsdir-debug-priv";
declare variable $t:trace-debug-role   := "opsdir-debug-role";
declare variable $t:trace-debug-rgrp   := "opsdir-debug-rgrp";
declare variable $t:trace-debug-http   := "opsdir-debug-http";
declare variable $t:trace-sessions     := "opsdir-sessions";
declare variable $t:trace-session-expires    := "opsdir-session-expires";
declare variable $t:trace-debug-scopes       := "opsdir-debug-scopes";
declare variable $t:trace-debug-log-messages := "opsdir-debug-log-messages";
declare variable $t:trace-debug-alerts       := "opsdir-debug-alerts";
declare variable $t:trace-email-alerts       := "opsdir-email-alerts";
declare variable $t:trace-debug-cluster-offline := "opsdir-debug-cluster-offline";
declare variable $t:trace-debug-configs := "opsdir-debug-configs";
declare variable $t:trace-disconnect   := "opsdir-disconnect";
declare variable $t:trace-polling      := "opsdir-polling";
declare variable $t:trace-pselect      := "opsdir-polling-selection";
declare variable $t:trace-poll-info    := "opsdir-polling-info";
declare variable $t:trace-expires      := "opsdir-expires";
declare variable $t:trace-joins        := "opsdir-joins";
declare variable $t:trace-joins-uri    := "opsdir-joins-uri";
declare variable $t:trace-info-uri     := "opsdir-info-uri";
declare variable $t:trace-debug-joins  := "opsdir-debug-joins";
declare variable $t:trace-force-joins  := "opsdir-force-joins";
declare variable $t:trace-debug-csv    := "opsdir-debug-csv";
declare variable $t:trace-search       := "opsdir-search";
declare variable $t:trace-app-endpoint := "opsdir-app-endpoint";

declare variable $t:trace-log-filter       := "opsdir-log-filter";
declare variable $t:trace-log-filter-debug := "opsdir-log-filter-debug";
declare variable $t:trace-alert-disable    := "opsdir-alert-disable";

(: This trace event enables unit test code. :)
declare variable $t:trace-joins-test   := "opsdir-joins-test";

declare variable $t:RESOURCE-COLLECTION
  := $t:COLLROOT || "resource";

declare variable $t:JOINS-COLLECTION
  := if (xdmp:trace-enabled("opsdir-joins-test"))
     then "http://marklogic.com/opsdirector/v1/join/test/data"
     else "http://marklogic.com/opsdirector/v1/join";

declare variable $t:ALERTS-COLLECTION
  := if (xdmp:trace-enabled("opsdir-joins-test"))
     then "http://marklogic.com/opsdirector/v1/alerts/test/data"
     else "http://marklogic.com/opsdirector/v1/alerts";

declare variable $t:LOG-AUGMENTERS
  := "http://marklogic.com/opsdirector/v1/log/alerts";

declare variable $t:DEFAULT-RESOURCE-SCOPE := "default-resource-scope";
declare variable $t:RESOURCE-SCOPE         := "resource-scope";
declare variable $t:RESOURCE-GROUP         := "resource-group";
declare variable $t:RESOURCE               := "resource";
declare variable $t:RESOURCE-SCOPES-LIST   := ($t:DEFAULT-RESOURCE-SCOPE,
                                               $t:RESOURCE-SCOPE,
                                               $t:RESOURCE-GROUP);

declare variable $t:SHARED-SECRET
     := "It doesn't matter as long as it's shared with OpsDirector";

(: I've allowed this to become polluted with types of resource files in the db :)
declare variable $t:RESOURCE-TYPES
  := ("clusters", "groups", "hosts", "databases", "servers", "forests", "tasks",
      "counts", "roles", "privileges", "security",
      "host-status", "forest-status", "forest-storage", "database-status", "server-status","local-cluster-status",
      "resource-list", "alerts");

declare variable $t:SERVER-RESOURCE-TYPES
  := ("clusters", "groups", "hosts", "databases", "servers", "forests", "tasks");

declare %private variable $Z := xs:dayTimeDuration("PT0S");
declare %private variable $QOPTS
  := ("collation=http://marklogic.com/collation/");

(: ====================================================================== :)

declare function t:timestamp(
) as xs:string
{
  t:timestamp(current-dateTime())
};

declare function t:timestamp(
  $dt as xs:dateTime
) as xs:string
{
  let $stamp := adjust-dateTime-to-timezone($dt, $Z)
  return
    format-dateTime($stamp, "[Y0001][M01][D01]-[H01][m01][s01].[f000001]")
};

(: ====================================================================== :)

declare function t:add-opsdir-properties(
  $obj as object-node(),
  $rsrc as object-node()?,
  $type as xs:string,
  $id as xs:string?
) as object-node()
{
  t:add-opsdir-properties($obj, $rsrc, $type, $id, ())
};

declare function t:add-opsdir-properties(
  $obj as object-node(),
  $rsrc as object-node()?,
  $type as xs:string,
  $id as xs:string?,
  $other as map:map?
) as object-node()
{
  t:add-opsdir-properties($obj, $rsrc, $type, $id, $other, ())
};

declare function t:add-opsdir-properties(
  $obj as object-node(),
  $rsrc as object-node()?,
  $type as xs:string,
  $id as xs:string?,
  $other as map:map?,
  $exclude as node()*
) as object-node()
{
  (: Get the name property; beware that there are two possible forms depending
     on whether the JSON has been camel cased or not. :)
  let $name  := switch ($type)
                case "forests"    return ($obj/forestName, $obj/forest-name)[1]
                case "clusters"   return ($obj/clusterName, $obj/cluster-name)[1]
                case "databases"  return ($obj/databaseName, $obj/database-name)[1]
                case "groups"     return ($obj/groupName, $obj/group-name)[1]
                case "servers"    return ($obj/serverName, $obj/server-name)[1]
                case "hosts"      return ($obj/hostName, $obj/host-name)[1]
                case "roles"      return ($obj/roleName, $obj/role-name)[1]
                case "privileges" return ($obj/privilegeName, $obj/privilege-name)[1]
                case "tasks"      return $id
                case "host-status"     return ($obj/hostStatus/name,
                                               $obj/host-status/name)[1]
                case "database-status" return ($obj/databaseStatus/name,
                                               $obj/database-status/name)[1]
                case "forest-status"   return ($obj/forestStatus/name,
                                               $obj/forest-status/name)[1]
                case "server-status"   return ($obj/serverStatus/name,
                                               $obj/server-status/name)[1]
                case "counts"          return ($obj/databaseCounts/name,
                                               $obj/database-counts/name)[1]
                default return ()

  let $other := if (empty($other)) then map:map() else $other

  let $props := map:new(
                  ($other,

                   if (map:contains($other, 'ts'))
                   then ()
                   else map:entry("ts", string(current-dateTime())),

                   if (map:contains($other, 'opsdirName') or empty($name))
                   then ()
                   else map:entry("opsdirName", $name),

                   if (map:contains($other, 'opsdirResourceType'))
                   then ()
                   else map:entry("opsdirResourceType", $type),

                   if (map:contains($other, 'opsdirId') or empty($id))
                   then ()
                   else map:entry("opsdirId", $id),

                   if (map:contains($other, 'opsdirDtstart'))
                   then ()
                   else
                     if ($rsrc/opsdirDtstart)
                     then map:entry("opsdirDtstart", $rsrc/opsdirDtstart)
                     else map:entry("opsdirDtstart", string(current-dateTime())),

                   if (map:contains($other, 'opsdirDtend'))
                   then ()
                   else map:entry("opsdirDtend", "9999-12-31T23:59:59Z"),

                   for $property in $obj/node() except $exclude
                   let $lname := local-name-from-QName(node-name($property))
                   where not(map:contains($other, $lname))
                   return
                     map:entry($lname, $property)))
  return
    xdmp:to-json($props)/object-node()
};

(: ====================================================================== :)

declare function t:replace-property(
  $obj as object-node(),
  $property as xs:string,
  $value as item()*
) as object-node()
{
  xdmp:to-json(
    map:new(
      (map:entry($property, $value),
       for $prop in $obj/node()
       let $name := local-name-from-QName(node-name($prop))
       where $name ne $property
       return
         map:entry($name, $prop)
    ))
  )/object-node()
};

(: ====================================================================== :)

declare function t:managed-resource-ids(
  $resource-type as xs:string?
) as xs:string*
{
  t:managed-resource-ids((), $resource-type)
};

declare function t:managed-resource-ids(
  $cluster-id as xs:string?,
  $resource-type as xs:string?
) as xs:string*
{
  t:managed-resource-ids($cluster-id, $resource-type, ())
};

declare function t:managed-resource-ids(
  $cluster-id as xs:string?,
  $resource-type as xs:string?,
  $updated-since as xs:string?
) as xs:string*
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  if (empty($updated-since) or ($updated-since castable as xs:dateTime))
  then ()
  else error((), "OD-INVDATE", ("Invalid date: " || $updated-since)),

  let $type-query := if (empty($resource-type))
                     then ()
                     else cts:element-range-query(
                              xs:QName("opsdirResourceType"), "=", $resource-type, $QOPTS)
  let $clus-query := if (empty($cluster-id))
                     then
                       ()
                     else
                       cts:element-range-query(
                           xs:QName("opsdirClusterId"), "=", $cluster-id, $QOPTS)
  let $coll-query := cts:collection-query($t:JOINS-COLLECTION)
  let $time-query := if (empty($updated-since))
                     then ()
                     else cts:element-range-query(
                              xs:QName("ts"), ">", xs:dateTime($updated-since))
  return
    cts:element-values(
      xs:QName("opsdirId"), (), (),
      cts:and-query(($type-query, $clus-query, $coll-query, $time-query)))
};

(: ====================================================================== :)

declare function t:managed-resources(
  $ids as xs:string*
) as object-node()*
{
  t:managed-resources((), $ids)
};

declare function t:managed-resources(
  $resource-type as xs:string?,
  $ids as xs:string*
) as object-node()*
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  $ids ! t:managed-resource((), $resource-type, .)
};

declare function t:managed-resource(
  $cluster-id as xs:string?,
  $resource-type as xs:string?,
  $id as xs:string
) as object-node()?
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  let $type-query := if (empty($resource-type))
                     then
                       (: exclude stats, counts, etc. by default :)
                       cts:or-query((
                         for $type in $t:SERVER-RESOURCE-TYPES
                         return
                           cts:json-property-range-query(
                              "opsdirResourceType", "=", $type, $QOPTS)))
                     else
                       cts:json-property-range-query(
                           "opsdirResourceType", "=", $resource-type, $QOPTS)

  let $c-query   := if (empty($cluster-id))
                    then
                      ()
                    else
                      cts:json-property-range-query(
                          "opsdirClusterId", "=", $cluster-id, $QOPTS)

  let $id-query  := cts:json-property-range-query("opsdirId", "=", $id, $QOPTS)

  let $coll      := if (empty($resource-type) or ($resource-type = $t:SERVER-RESOURCE-TYPES))
                    then $t:JOINS-COLLECTION
                    else if (contains($resource-type, "status") or $resource-type = "counts"
                             or $resource-type = "forest-storage")
                         then $t:COLLROOT||"stats"
                         else $t:COLLROOT||"resource"

  let $doc       := cts:search(collection($coll),
                               cts:and-query(($type-query, $c-query, $id-query)),
                               "unfiltered")
  let $_ := if (count($doc) gt 1)
            then xdmp:log(("Error: multiple resource documents for ('"
                          || $cluster-id || "','" || $resource-type || "','"
                          || $id || " (" || count($doc) || ")",
                          $doc ! xdmp:node-uri(.)))
            else ()
  return
    ($doc)[1]/node()
};

(: ====================================================================== :)

declare function t:managed-resource-id(
  $cluster-id as xs:string?,
  $resource-type as xs:string,
  $resource-name as xs:string
) as xs:string*
{
  (: N.B. There can be several servers with the same name in different groups :)
  if ($resource-type = $t:RESOURCE-TYPES)
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  for $id in t:managed-resource-ids($cluster-id, $resource-type)
  let $obj  := t:managed-resource($cluster-id, $resource-type, $id)
  where $obj/opsdirName = $resource-name
  return
    $obj/opsdirId
};

(: ====================================================================== :)

declare function t:resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string,
  $resource-id as xs:string
) as object-node()
{
  t:resource-scope($cluster, $resource-type, $resource-id, true())
};

declare function t:resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string,
  $resource-id as xs:string,
  $create as xs:boolean
) as object-node()?
{
  if (empty($cluster) or ($cluster = t:managed-resource-ids("clusters")))
  then ()
  else error((), "OD-NOCLUSTER", ("No such cluster: " || $cluster)),

  let $obj  := t:managed-resource($cluster, $resource-type, $resource-id)
  let $_    := if (empty($obj))
               then error((), "OD-NORESOURCE", ("Invalid scope, no such resource ("
                              || $resource-type || "/" || $resource-id || ")"))
               else ()

  let $cid  := string($obj/opsdirClusterId)
  let $rid  := string($obj/opsdirId)

  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:json-property-range-query("opsdirClusterId", "=", $cid, $QOPTS)
  let $id-query  := cts:json-property-range-query("opsdirId", "=", $rid, $QOPTS)
  let $k-query   := cts:json-property-value-query("opsdirKind", "resource")

  let $doc       := cts:search(collection($t:COLLROOT || "resource-scope"),
                               cts:and-query(($dt-query, $c-query, $id-query, $k-query)),
                               "unfiltered")

  let $_    := if (count($doc) gt 1)
               then error((), "OD-INTERR" ("Internal error: multiple resources for "
                              || "resource-scope(" || $cluster || ", "
                              || $resource-type || ", " || $resource-id || ")"))
               else ()
  return
    if (empty($doc) and $create)
    then
      object-node {
        "opsdirKind": $t:RESOURCE-SCOPE,
        "opsdirClusterId": $cid,
        "opsdirResourceType": $resource-type,
        "opsdirId": $rid,
        "opsdirDtstart": string(current-dateTime()),
        "ts": string(current-dateTime()),
        "opsdirDtend": "9999-12-31T23:59:59Z"
      }
    else
      $doc/object-node()
};

declare function t:default-resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string?
) as object-node()
{
  t:default-resource-scope($cluster, $resource-type, true())
};

declare %private function t:default-resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string?,
  $create as xs:boolean
) as object-node()?
{
  if (empty($resource-type) or ($resource-type = ("all", $t:RESOURCE-TYPES)))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  if (empty($cluster) or ($cluster = ("all", t:managed-resource-ids("clusters"))))
  then ()
  else error((), "OD-NOCLUSTER", ("No such cluster: " || $cluster)),

  let $cid   := if (empty($cluster)) then "all" else $cluster
  let $rtype := if (empty($resource-type)) then "all" else $resource-type

  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:element-range-query(
                        xs:QName("opsdirClusterId"), "=", $cid, $QOPTS)
  let $t-query   := cts:element-range-query(
                        xs:QName("opsdirResourceType"), "=", $rtype, $QOPTS)
  let $id-query  := cts:element-range-query(xs:QName("opsdirId"), "=", "all", $QOPTS)
  let $doc       := cts:search(collection($t:COLLROOT || "resource-scope"),
                               cts:and-query(($dt-query, $c-query, $t-query, $id-query)),
                               "unfiltered")

  let $_    := if (count($doc) gt 1)
               then error((), "OD-INTERR", ("Internal error: multiple resources for "
                              || "default-resource-scope(" || $cluster || ", "
                              || $resource-type || ")"))
               else ()
  return
    if (empty($doc) and $create)
    then
      object-node {
        "opsdirKind": $t:DEFAULT-RESOURCE-SCOPE,
        "opsdirClusterId": $cid,
        "opsdirResourceType": $resource-type,
        "opsdirId": "all",
        "opsdirDtstart": string(current-dateTime()),
        "ts": string(current-dateTime()),
        "opsdirDtend": "9999-12-31T23:59:59Z"
      }
    else
      $doc/object-node()
};

(: ====================================================================== :)

declare %private function t:assert-roles(
  $roles as xs:string*
) as xs:string*
{
  for $role in $roles
  return
    try {
       string(xdmp:role($role))
    } catch ($e) {
      if ($role castable as xs:unsignedLong)
      then
        if (empty(ts:get-role-names($role cast as xs:unsignedLong)))
        then error((), "OD-BADROLE", ("Bad role: " || $role))
        else $role
      else
        error((), "OD-BADROLE", ("Bad role: " || $role))
   }
};

declare function t:assert-scope(
  $scope as object-node()
) as object-node()
{
  t:assert-scope($scope, $t:RESOURCE-SCOPES-LIST)
};

declare function t:assert-scope(
  $scope as object-node(),
  $scopes as xs:string+
) as object-node()
{
  if ($scope/opsdirKind = $scopes)
  then
    $scope
  else
    error((), "OD-BADSCOPE", ("Unexpected scope: " || $scope/opsdirKind))
};

declare function t:group-id-for-host(
  $host as object-node()
)
{
  let $cid   := $host/opsdirClusterId
  let $name  := $host/opsdirGroupName
  let $qtype := cts:json-property-value-query("opsdirResourceType", "groups")
  let $qname := cts:json-property-value-query("opsdirName", $name)
  let $qclus := cts:json-property-value-query("opsdirClusterId", $cid)
  let $group := cts:search(collection($t:COLLROOT||"resource"),
                           cts:and-query(($qtype, $qname, $qclus)))[1]/object-node()
  return
    $group/opsdirId
};

(: ====================================================================== :)

declare %private function t:resource-get-config-uri(
    $scope as element(resourceScope)
) as xs:string
{
  let $coll := $t:COLLROOT || "config/" || $scope/opsdirClusterId
               || "/" || $scope/opsdirResourceType || "/" || $scope/opsdirId
  let $doc  := collection($coll)
  let $_    := if (count($doc) ne 1)
               then error((), "OD-BADRSRC", ("Resource doesn't exactly identify a single config: " || $coll))
               else ()
  return
    xdmp:node-uri($doc)
};

(: ====================================================================== :)

declare function t:cluster-resource-timestamps(
  $cluster-id as xs:string,
  $host-id as xs:string
) as object-node()?
{
  let $t-query   := cts:element-range-query(
                        xs:QName("opsdirResourceType"), "=", "timestamps", $QOPTS)
  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:element-range-query(
                        xs:QName("opsdirClusterId"), "=", $cluster-id, $QOPTS)
  let $h-query   := cts:element-range-query(
                        xs:QName("opsdirHostId"), "=", $host-id, $QOPTS)
  let $doc       := cts:search(collection(),
                               cts:and-query(($t-query, $dt-query, $c-query, $h-query)),
                               ("unfiltered",
                                cts:index-order(
                                  cts:element-reference(fn:QName("","ts")), "descending")))

  let $_ := if (count($doc) gt 1)
            then xdmp:log(("Multiple timestamps:", $doc ! xdmp:node-uri(.)))
            else ()

  return
    ($doc/node())[1]
};

(: ====================================================================== :)

declare function t:assert-resource-type(
  $resources as xs:string*,
  $type as xs:string
) as xs:string*
{
  let $_ := if ($type = $SERVER-RESOURCE-TYPES)
            then
              ()
            else
              error((), "OD-BADRTYPE", ("Bad resource type: " || $type))
  return
    if (empty($resources))
    then
      ()
    else
      let $ids       := t:managed-resource-ids((), $type)
      let $resources := for $resource in $resources
                        return
                          if ($resource = $ids)
                          then $resource
                          else
                            let $_ := trace($resource || " is not a " || $type || ": "
                                            || string-join($ids ! string(.), ", "),
                                            $t:trace-debug-rgrp)
                            return
                              error((), "OD-BADRSRC", ("Bad resource; not a " || $type || ": " || $resource))
      return
        $resources
};

(: ====================================================================== :)
(: N.B. These are common, they're called from info and disconn, but the
   aren't amped so they're not likely to work from random places.
:)

declare function t:open-local-alert(
  $cid as xs:string,
  $hid as xs:string,
  $code as xs:string,
  $state as xs:string,
  $resourceType as xs:string,
  $resourceId as xs:string,
  $resourceName as xs:string,
  $message as xs:string
) as empty-sequence()
{
  (: Is there an existing open alert with this $cid and $code?
     There should never be more than one, but ... :)
  let $colls := ("http://marklogic.com/opsdirector/v1/alerts/" || $cid,
                 "http://marklogic.com/opsdirector/v1/alerts")
  let $qclus := cts:json-property-value-query("opsdirClusterId", $cid)
  let $qrsrc := cts:json-property-value-query("opsdirId", $resourceId)
  let $qstat := cts:json-property-value-query("status", "open")
  let $alert := cts:search(collection($colls[1]),
                           cts:and-query(($qclus, $qrsrc, $qstat)))/object-node()

  let $alert := $alert[code = $code]
  return
    if (empty($alert))
    then
      (: We're going to generate a log message; let's make sure it has the
         same permissions as the relevant resource... :)
      let $rsrc  := t:managed-resource($cid, $resourceType||"s", $resourceId)
      let $perms := if (exists($rsrc))
                    then xdmp:document-get-permissions(xdmp:node-uri($rsrc))
                    else
                      (xdmp:log("Generating local alert for non-existant resource: "
                                || $resourceId || " (" || $resourceType
                                || " on " || $cid || ")"),
                       $t:DBPERMS)
      let $cluster := t:managed-resource($cid, "clusters", $cid)
      let $host    := t:managed-resource($cid, "hosts", $hid)
      (: Generate a log message :)
      let $_       := xdmp:log($code || ": " || $host/*:opsdirName || ": " || $message)
      let $uid     := string(xdmp:random())
      let $alert   := object-node {
        "opsdirClusterId": $cid,
        "opsdirClusterName": $cluster/*:opsdirName,
        "opsdirHostId": $hid,
        "duration": "PT0S",
        "opsdirResourceType": "alerts",
        "resourceType": $resourceType,
        "opsdirDtstart": string(current-dateTime()),
        "code": $code,
        "message": $message,
        "state": $state,
        "local": true(),
        "resourceName": $resourceName,
        "status": "open",
        "id": string(xdmp:random()),
        "uid": $uid,
        "ts": string(current-dateTime()),
        "acknowledged": "no",
        "opsdirId": $resourceId,
        "resourceId": $resourceId }
      return
        xdmp:document-insert("/alerts/" || $cid || "/" || $uid || ".json",
                             $alert, $perms, $colls)
    else
      let $uri := xdmp:node-uri($alert)
      (: Oh for the love of ... this is a hack :)
      let $map := map:new(for $key in $alert/* except ($alert/duration)
                          return map:entry(
                                    local-name-from-QName(node-name($key)),
                                    $alert/*[node-name(.) = node-name($key)]))
      let $duration := current-dateTime() - xs:dateTime($alert/dtstart)
      let $obj      := object-node { "duration": string($duration) }
      let $newalert := xdmp:to-json($map + $obj)/node()
      return
        xdmp:node-replace($alert, $newalert)
};

declare function t:close-local-alert(
  $cid as xs:string,
  $resourceId as xs:string,
  $code as xs:string
) as empty-sequence()
{
  let $colls := ("http://marklogic.com/opsdirector/v1/alerts/" || $cid,
                 "http://marklogic.com/opsdirector/v1/alerts")
  let $qclus := cts:json-property-value-query("opsdirClusterId", $cid)
  let $qrsrc := cts:json-property-value-query("opsdirId", $resourceId)
  let $qstat := cts:json-property-value-query("status", "open")
  let $alert := cts:search(collection($colls[1]),
                           cts:and-query(($qclus, $qrsrc, $qstat)))/object-node()
  for $alert in $alert[code = $code]
  return
    (: Oh for the love of ... this is a hack :)
    let $map := map:new(for $key in $alert/*
                        except ($alert/duration|$alert/opsdirDtend|$alert/status)
                        return map:entry(local-name-from-QName(node-name($key)),
                                         $alert/*[node-name(.) = node-name($key)]))
    let $duration := current-dateTime() - xs:dateTime($alert/dtstart)
    let $obj      := object-node { "duration": string($duration),
                                   "opsdirDtend": string(current-dateTime()),
                                   "status": "closed" }
    let $newalert := xdmp:to-json($map + $obj)/node()
    return
      xdmp:node-replace($alert, $newalert)
};

(: ====================================================================== :)

declare function t:get-role-name(
  $roleid as xs:string
) as xs:string
{
  ts:get-role-names(xs:unsignedLong($roleid))
};

(: ====================================================================== :)

declare function t:disconnect-cluster(
  $cid as xs:string
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                       "execute"),

  let $hids    := t:managed-resource-ids($cid, "hosts")
  let $errors  := for $hid in $hids
                  let $uri := "/config/host-status/" || $cid || "/" || $hid ||"/snapshot.json"
                  let $obj := doc($uri)/object-node()
                  where $obj/status = "error" or $obj/status = "wrongid"
                  return
                    $obj
  let $_       := if (count($hids) != count($errors))
                  then error((), "OD-NOTDISC",
                             "Some hosts can still be reached: unmanage from one of them")
                  else ()

  let $alluris := cts:uris((), (),
                    cts:and-query(
                      (cts:collection-query(($RESOURCE-COLLECTION, $JOINS-COLLECTION)),
                       cts:element-range-query(
                          xs:QName("opsdirClusterId"), "=", $cid,
                                   ("collation=http://marklogic.com/collation/")))))

  return
    ts:reset-collections-on-uris($alluris)
};

