xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace ts="http://marklogic.com/v1/opsdirector/security";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes"
    at "/common/rscopes.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare function ts:get-role-names(
  $roles as xs:unsignedLong*
) as xs:string*
{
  ts:get-role-names($roles, ())
};

declare function ts:get-role-names(
  $requested-roles as xs:unsignedLong*,
  $regex as xs:string?
) as xs:string*
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  (: If we're opsdir-admin, then it's ok to show all the roles,
     otherwise you only get the roles that you have :)
  let $roles := try {
                  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                                       "execute"),
                  $requested-roles
                } catch($e) {
                  let $allowed-roles := xdmp:user-roles(xdmp:get-current-user())
                  return
                    if (empty($requested-roles))
                    then
                      $allowed-roles
                    else
                      for $role in $requested-roles
                      where $role = $allowed-roles
                      return
                        $role
                }

  return
    xdmp:invoke-function(
      function () {
        if (empty($roles))
        then
          for $name in /sec:role ! string(./sec:role-name)
          where empty($regex) or matches($name, $regex)
          return
            $name
        else
          (: What about roles that no longer exist? :)
          for $role in $roles
          return
            try {
              sec:get-role-names($role)
            } catch ($e) {
              if ($e/error:code = "SEC-ROLEDNE")
              then ()
              else xdmp:rethrow()
            }
      },
      <options xmlns="xdmp:eval">
        <database>{xdmp:security-database()}</database>
      </options>)
};

declare function ts:get-role-description(
  $name as xs:string
) as xs:string
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  xdmp:invoke-function(
    function () {
      sec:role-get-description($name)
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>)
};

declare function ts:set-role-description(
  $name as xs:string,
  $description as xs:string
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  xdmp:invoke-function(
    function () {
      sec:role-set-description($name, $description)
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>)
};

declare function ts:get-role-subroles(
  $name as xs:string*
) as xs:string*
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  xdmp:role-roles($name) ! ts:get-role-names(.)
};

declare function ts:set-role-subroles(
  $name as xs:string,
  $subroles as xs:string*
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  xdmp:invoke-function(
    function () {
      sec:role-set-roles($name, $subroles)
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>)
};

declare function ts:role-exists(
  $role as xs:string
) as xs:boolean
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  xdmp:invoke-function(
    function () { sec:role-exists($role) },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
    </options>)
};

declare function ts:create-role(
  $role as xs:string
) as xs:unsignedLong
{
  ts:create-role($role, (), "Internal role for OpsDirector resource group")
};

declare function ts:create-role(
  $role as xs:string,
  $subroles as xs:string*,
  $description as xs:string
) as xs:unsignedLong
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  xdmp:invoke-function(
    function () { sec:create-role($role, $description, $subroles, (), ()) },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
      <isolation>different-transaction</isolation>
      <transaction-mode>update-auto-commit</transaction-mode>
      <prevent-deadlocks>false</prevent-deadlocks>
    </options>)
};

declare function ts:remove-role(
  $role as xs:string
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  xdmp:invoke-function(
    function () {
      try {
        sec:remove-role($role)
      } catch ($e) {
        if ($e/error:code = "SEC-ROLEDNE")
        then ()
        else xdmp:rethrow()
      }
    },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
      <isolation>different-transaction</isolation>
      <transaction-mode>update-auto-commit</transaction-mode>
      <prevent-deadlocks>false</prevent-deadlocks>
    </options>)
};

declare function ts:add-roles-to-user(
  $role as xs:string+
) as empty-sequence()
{
  ts:add-roles-to-user($role, xdmp:get-current-user())
};

declare function ts:add-roles-to-user(
  $role as xs:string+,
  $user as xs:string
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  xdmp:invoke-function(
    function () { sec:user-add-roles($user, $role) },
    <options xmlns="xdmp:eval">
      <database>{xdmp:security-database()}</database>
      <isolation>different-transaction</isolation>
      <transaction-mode>update-auto-commit</transaction-mode>
      <prevent-deadlocks>false</prevent-deadlocks>
    </options>)
};

(: This function is amped but is only supposed to be called by the
   update-configuration functions.
:)
declare function ts:remove-resources-from-resource-group(
  $type as xs:string,
  $deleted as xs:string+
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-data-internal",
                       "execute"),

  (: Now update all effected resource groups :)
  let $rgs := for $rg in rs:resource-groups-containing-resources($type, $deleted)
              return
                rs:resource-group-remove-resources($rg, $deleted)
  let $_   := trace("Updating " || count($rgs) || " resource group(s)", $t:trace-debug)
  where exists($rgs)
  return
    rs:resource-scopes-save($rgs)
};

declare function ts:insert-resource-group(
  $uri as xs:string,
  $role as xs:string,
  $group as object-node()
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  if (doc-available($uri))
  then error((), "Attempt to create duplicate resource group: " || $uri)
  else (),

  xdmp:invoke-function(
    function () {
      let $perms  := ($t:ADMIN-DBPERMS,
                      xdmp:permission($role, "update"))
      let $colls  := $t:COLLROOT || "resource-group"
      let $group  := t:assert-scope($group, $t:RESOURCE-GROUP)
      let $group  := t:replace-property($group, "opsdirDtstart", string(current-dateTime()))
      let $group  := t:replace-property($group, "opsdirDtend", "9999-12-31T23:59:59Z")
      let $group  := t:replace-property($group, "relatedRole", string(xdmp:role($role)))
      let $_ := trace("Insert: " || $uri, $t:trace-debug-insert)
      let $_ := trace("Ins: " || $uri, $t:trace-debug-uri)
      return
        xdmp:document-insert($uri, $group, $perms, $colls)
    },
    <options xmlns="xdmp:eval">
      <isolation>different-transaction</isolation>
      <transaction-mode>update-auto-commit</transaction-mode>
      <prevent-deadlocks>false</prevent-deadlocks>
    </options>)
};

declare function ts:resource-scope-save(
  $scope as object-node()
) as empty-sequence()
{
  (: If we're saving it, make sure we update the timestamp :)
  let $scope := t:assert-scope($scope)
  let $scope := t:replace-property($scope, "ts", string(current-dateTime()))
  return
    switch ($scope/opsdirKind)
    case $t:RESOURCE-GROUP
    return
      let $id     := $scope/opsdirId
      let $rg     := rs:resource-group($id)
      let $olduri := xdmp:node-uri($rg)
      let $newuri := "/resource-group/" || $id || "/" || t:timestamp() || ".json"
      let $perm   := ($t:ADMIN-DBPERMS,
                      for $gp in $scope/groupPermissions
                      let $role := xs:unsignedLong($gp/role)
                      let $exists := try { xdmp:role-name($role) } catch ($e) { () }
                      where exists($exists)
                      return
                        xdmp:permission($role, $gp/capability))
      let $coll   := $t:COLLROOT || "resource-group"
      let $_      := if ($scope/DELETE = "true")
                     then
                       trace("Deleting (not saving) resource scope: " || $newuri,
                             $t:trace-debug-insert)
                     else
                       let $_ := trace("Insert: " || $newuri, $t:trace-debug-insert)
                       return
                         xdmp:document-insert($newuri, $scope, $perm, $coll)
      let $_      := if ($olduri = $newuri)
                     then () (: created in same txn :)
                     else (xdmp:document-set-collections($olduri, ()),
                           xdmp:document-set-permissions($olduri, $perm))
      return
        if ($olduri = $newuri)
        then
          ()
        else
          let $updated := t:replace-property($rg, "opsdirDtend", string(current-dateTime()))
          return
            xdmp:node-replace($rg, $updated)
    case $t:RESOURCE-SCOPE
    return
      let $rsrc   := t:resource-scope($scope/opsdirClusterId, $scope/opsdirResourceType,
                                      $scope/opsdirId)
      let $_      := if (exists(xdmp:node-uri($rsrc)))
                     then
                       let $uprsrc := t:replace-property($rsrc, "opsdirDtend",
                                                         string(current-dateTime()))
                       return
                         xdmp:node-replace($rsrc, $uprsrc)
                     else
                       ()
      let $olduri := xdmp:node-uri($rsrc)
      let $newuri := "/resource-scope/" || $scope/opsdirResourceType
                     || "/" || $scope/opsdirClusterId
                     || "/" || $scope/opsdirId
                     || "/" || t:timestamp() || ".json"
      let $coll   := $t:COLLROOT || "resource-scope"
      let $_      := trace("Insert: " || $newuri, $t:trace-debug-insert)
      let $_      := trace("Ins: " || $newuri, $t:trace-debug-uri)
      let $_      := xdmp:document-insert($newuri, $scope, $t:DBPERMS, $coll)
      let $_      := if (empty($olduri))
                     then ()
                     else xdmp:document-set-collections($olduri, ())
      return
        ()
    case $t:DEFAULT-RESOURCE-SCOPE
    return
      let $cid    := if ($scope/opsdirClusterId = "all")
                     then ()
                     else $scope/opsdirClusterId
      let $type   := if ($scope/opsdirResourceType = "all")
                     then ()
                     else $scope/opsdirResourceType
      let $rsrc   := t:default-resource-scope($cid, $type)
      let $olduri := xdmp:node-uri($rsrc)
      let $newuri := "/resource-scope/" || $scope/opsdirResourceType
                     || "/" || $scope/opsdirClusterId
                     || "/" || $scope/opsdirId
                     || "/" || t:timestamp() || ".json"
      let $coll   := $t:COLLROOT || "resource-scope"
      let $_      := trace("Insert: " || $newuri, $t:trace-debug-insert)
      let $_      := trace("Ins: " || $newuri, $t:trace-debug-uri)
      let $_      := xdmp:document-insert($newuri, $scope, $t:DBPERMS, $coll)
      let $_      := if (empty($olduri))
                     then
                       ()
                     else
                       let $uprsrc := t:replace-property($rsrc, "opsdirDtend",
                                                         string(current-dateTime()))
                       let $_ := xdmp:document-set-collections($olduri, ())
                       return
                         xdmp:node-replace($rsrc, $uprsrc)
      return
        ()
    default
      return
        error((), "Invalid scope: " || $scope/opsdirKind)
};

declare function ts:resource-scope-create(
  $scope as object-node()
) as empty-sequence()
{
  switch ($scope/opsdirKind)
  case "resource-scope"
  return
    let $_      := try {
                     let $rsrc := t:resource-scope($scope/opsdirClusterId,
                                     $scope/opsdirResourceType,
                                     $scope/opsdirId)
                     return
                       error((), "Resource exists")
                   } catch ($e) {
                     ()
                   }

    let $newuri := "/resource-scope/" || $scope/opsdirResourceType
                   || "/" || $scope/opsdirClusterId
                   || "/" || $scope/opsdirId
                   || "/" || t:timestamp() || ".json"
    let $coll   := ($t:COLLROOT || "resource-scope",
                    $t:COLLROOT || "resource-scope" || "/" || $scope/opsdirClusterId
                                || "/" || $scope/opsdirResourceType,
                    $t:COLLROOT || "resource-scope" || "/" || $scope/opsdirClusterId
                                || "/" || $scope/opsdirResourceType
                                || "/" || $scope/opsdirId)
    let $_      := trace("Insert: " || $newuri, $t:trace-debug-insert)
    let $_      := trace("Ins: " || $newuri, $t:trace-debug-uri)
    let $_      := xdmp:document-insert($newuri, $scope, $t:DBPERMS, $coll)
    return
      ()
  default
    return
      error((), "Not a resource")
};

declare function ts:endpoint-authentication()
{
  ts:endpoint-authentication(true())
};

declare function ts:endpoint-authentication(
  $require-authentication as xs:boolean
) as xs:boolean
{
  let $cert := xdmp:get-request-client-certificate()
  return
    if (empty($cert))
    then
      if ($require-authentication)
      then
        error((), "Authentication required to use this endpoint")
      else
        (xdmp:log("Missing certificate; unauthenticated access to " || xdmp:get-original-url()
                  || ", running as user: " || xdmp:get-current-user()),
         false())
    else
      try {
        if (xdmp:login(""))
        then
          let $_ := trace("Certificate login, user: " || xdmp:get-request-username(),
                          $t:trace-debug)
          return
            true()
        else
          if ($require-authentication)
          then
            error((), "Authentication required to use this endpoint")
          else
            (xdmp:log("No login; unauthenticated access to " || xdmp:get-original-url()
                      || ", running as user: " || xdmp:get-current-user()),
             false())
      } catch ($e) {
        xdmp:log(("Certificate authentication failed!", $e)),
        false()
      }
};

declare function ts:acknowledge-alert(
  $uri as xs:string,
  $ack as xs:string
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  let $user-visible := xdmp:invoke-function(function () {
                         doc-available($uri)
                       }, <options xmlns="xdmp:eval">
                            <ignore-amps>true</ignore-amps>
                          </options>)
  let $valid-ack    := if ($ack = "yes") then string($ack) else "no"
  return
    if ($user-visible)
    then
      if (starts-with($uri, "/alerts/"))
      then
        let $doc := doc($uri)/object-node()
        let $ack := $doc/acknowledged
        let $new := object-node { "acknowledged": $valid-ack }
        return
          xdmp:node-replace($ack, $new/acknowledged)
      else
        error((), "OD-BADURI", ("Attempt to acknowledge document that isn't an alert: "
                                || $uri))
    else
      error((), "SEC-PERMDENIED", ("Cannot acknowledge alert: " || $uri))
};

declare function ts:reset-collections-on-uris(
  $uris as xs:string*
) as empty-sequence()
{
  xdmp:spawn-function(
    function () {
      ts:reset-collections-on-uris($uris, 1000)
    },
    <options xmlns="xdmp:eval">
      <update>true</update>
      <commit>auto</commit>
    </options>)
};

declare function ts:reset-collections-on-uris(
  $uris as xs:string*,
  $size as xs:unsignedLong
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                       "execute"),

  let $fix  := subsequence($uris, 1, $size)
  let $_    := $fix ! xdmp:document-set-collections(., ())
  let $uris := subsequence($uris, $size+1)
  where exists($uris)
  return
    xdmp:spawn-function(
      function () {
        ts:reset-collections-on-uris($uris, $size)
      },
      <options xmlns="xdmp:eval">
        <update>true</update>
        <commit>auto</commit>
      </options>)
};
