xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace p="http://marklogic.com/v1/opsdirector/polling";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace raw="http://marklogic.com/v1/opsdirector/rawdata"
    at "/common/rawdata.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare namespace host="http://marklogic.com/xdmp/status/host";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare function p:host-to-run(
) as xs:boolean
{
  (: If you want a scheduled tasks to run on only a single host. We
     need to pick a host that's in the right group, otherwise the
     tasks will never run! In the appropriate group, we pick the host
     that has the smallest host id that is not in an error state. :)

  (: FIXME: Go back to CJL's smallest-Security-forest-replica plan :)

  let $hosts     := xdmp:group-hosts()
  let $ok-hosts  := $hosts[empty(xdmp:host-status(.)/host:error)]
  return
    min($ok-hosts) = xdmp:host()
};

declare function p:poll-clusters(
  $type as xs:string,
  $poll as function(*)
)
{
  p:poll-clusters($type, $poll, 1)
};

declare function p:poll-clusters(
  $type as xs:string,
  $poll as function(*),
  $default-concurrency as xs:unsignedLong
)
{
  let $c-name      := $type || "-concurrency"
  let $t-name      := $type || "-time-limit"
  let $txn-name    := "opsdir-" || $type || "-poll"
  let $concurrency := xs:unsignedLong(config:get-atomic-property($c-name, 6))
  let $time-limit  := xs:unsignedLong(config:get-atomic-property($t-name, 300))
  let $txnlist     := xdmp:transaction($txn-name)
  return
    if (count($txnlist) ge $concurrency)
    then
      let $_    := trace("Host " || xdmp:host-name()
                         || " " || $type || " concurrency too high: "
                         || count($txnlist), $t:trace-polling)
      let $txns := xdmp:host-status(xdmp:host())/host:transactions
                      /host:transaction[host:transaction-name = $txn-name]
      for $txn in $txns
      let $time := current-dateTime() - $txn/host:start-time
      return
        trace("Txn " || $txn/host:transaction-id
              || " polling " || $type || " for " || $time, $t:trace-polling)
    else
      let $cluster := p:poll-cluster($type)
      let $_       := if (empty($cluster))
                      then
                        trace("Host " || xdmp:host-name()
                              || " has no cluster to poll for " || $type, $t:trace-polling)
                      else
                        ()
      let $hosts  := if (exists($cluster))
                     then raw:resource-ids($cluster, "hosts")
                     else ()
      let $_       := if (empty($hosts))
                      then
                        trace("Host " || xdmp:host-name()
                              || " finds no hosts to poll for cluster " || $cluster,
                              $t:trace-polling)
                      else
                        ()
      where exists($hosts)
      return
        (: Choose the least recently polled host on this cluster :)
        let $dquery := cts:directory-query("/config/host-status/", "infinity")
        let $cquery := cts:json-property-value-query("opsdirClusterId", $cluster)
        let $order  := cts:index-order(cts:element-reference(fn:QName("", "ts")), "ascending")
        let $count  := xdmp:estimate(
                          cts:search(/, cts:and-query(($dquery, $cquery)), $order))
        let $oldest := cts:search(/, cts:and-query(($dquery, $cquery)), $order)[1]
        let $hostid := $oldest/opsdirHostId
        let $host   := raw:resource($cluster, "hosts", $hostid)
        let $_      := trace("Picked host: " || $host/opsdirName
                             || " (least recently polled of " || $count || " hosts)"
                             || " for " || $type,
                             $t:trace-pselect)
        let $_      := if (empty($hostid))
                       then
                         trace("Host " || xdmp:host-name()
                               || " has no hosts to poll for " || $type, $t:trace-polling)
                       else
                         trace("Host " || xdmp:host-name()
                               || " poll "
                               || $host/opsdirClusterName || "/"
                               || $host/opsdirName
                               || " for " || $type, $t:trace-polling)
        where exists($hostid)
        return
          xdmp:spawn-function(function () {
              p:poll-host($txn-name, $poll, $host, $type)
            },
            <options xmlns="xdmp:eval">
              <time-limit>{$time-limit}</time-limit>
              <update>true</update>
              <commit>auto</commit>
            </options>)
};

declare %private function p:poll-host(
  $txn-name as xs:string,
  $poll as function(object-node()) as empty-sequence(),
  $host as object-node(),
  $type as xs:string
)
{
  let $_    := xdmp:set-transaction-name($txn-name)
  let $_    := $poll($host)
  return
    let $txn  := xdmp:host-status(xdmp:host())/host:transactions
                   /host:transaction[host:transaction-id = xdmp:transaction()]
    let $_    := trace("Host " || xdmp:host-name()
                       || " poll "
                       || $host/opsdirClusterName || "/"
                       || $host/opsdirName
                       || " for " || $type || ": " || string(xdmp:elapsed-time()),
                       $t:trace-polling)
    return
      ()
};

(: ====================================================================== :)

declare %private function p:poll-cluster(
  $type as xs:string
) as xs:string?
{
  let $clusters := raw:resource-ids("clusters")
  let $hosts    := for $host in xdmp:group-hosts()
                   let $key := "suppress-polling-from-" || $host
                   let $value := config:get-atomic-property($key, false())
                   where string($value) ne "true"
                   return
                     $host
  where exists($clusters)
  return
    p:poll-cluster($type, current-dateTime(), $clusters, $hosts, xdmp:host())
};

declare %private function p:poll-cluster(
  $type as xs:string,
  $stamp as xs:dateTime,
  $clusters as xs:string+,
  $hosts-in-group as xs:unsignedLong+,
  $this-host as xs:unsignedLong
) as xs:string?
{
  (: ========================================================================

   This function determines what cluster should be polled next (for
   alerts, for example, or status information). This function is used
   by scheduled tasks. We assume that this function is called at the
   same time (to a granularity of 1 minute) on all hosts in the group.
   We assume all hosts are in the same timezone so TZ doesn't matter.

   It has the following goals:

   1. Every host in the group should poll a different cluster:
      * In this iteration
        - This assures that different tasks are never trying to update
          the same documents.
      * On each iteration
        - This assures that if one host is down, a particular cluster
          doesn't get starved for updates.
   2. We make this calculation without maintaining state in the
      database. This means that we can work out what to do without
      acquiring a lock on a document in the database.

   Solution: Take the list of clusters in numerical order by cluster id.
   Use the number of minutes past midnight to locate a starting place
   in the list. Distribute the clusters across the hosts.

   But be careful, if we just blindly iterated through in a strictly
   round-robin fashion, we could end up always polling the same
   clusters with the same hosts. Consider the simple case of three
   clusters and three hosts: you'd get the cycle 1,2,3; 1,2,3; etc.
   That would mean that if the first host was down, cluster 1 would
   never get polled. To combat this, after each complete cycle, we
   move forward. So we start the first cycle with the first host,
   the second cyle with the second, etc.

   For example:

   Consider the case of ten clusters (numbered 11-20) and three hosts
   (numbered 111-113). Iterations occur every minute:

    1: 111 polls 11, 112 polls 12, 113 polls 13
    2: 111 polls 14, 112 polls 15, 113 polls 16
    3: 111 polls 17, 112 polls 18, 113 polls 19
    4: 111 polls 20

   At this point, we wrap around.

    4:               112 polls 11, 113 polls 12

   Now we've done a full cycle, so we start over at 12:

    5: 111 polls 12, 112 polls 13, 113 polls 14
    6: 111 polls 15, 112 polls 16, 113 polls 17
    7: 111 polls 18, 112 polls 19, 113 polls 20
    8: 111 polls 11, 112 polls 12, 113 polls 13

   Now we've done a full cycle, so we start over at 13:

    9: 111 polls 13, 112 polls 14, 113 polls 15
   10: 111 polls 16, 112 polls 17, 113 polls 18
   11: 111 polls 19, 112 polls 20, 113 polls 11
   12: 111 polls 12, 112 polls 13, 113 polls 14

   Etc.

   You can see that there's a little bit of overlap between
   the end of one cycle and the beginning of the next. This little
   "ripple" rolls across the whole set over time, but doesn't seem
   significant enough to worry about.

   ======================================================================== :)

  let $minute := (hours-from-dateTime($stamp) * 60) + minutes-from-dateTime($stamp)

  let $_      := trace("Polling at " || $stamp || " (" || $minute || ") on host "
                       || xdmp:host-name($this-host) || " for " || $type,
                       $t:trace-pselect)
  let $_      := trace(count($clusters) || " clusters, "
                       || count($hosts-in-group) || " hosts for " || $type,
                       $t:trace-pselect)

  (: We need them in a stable order; could stash this in a server field
     or something, but do we really think the time it takes to sort a
     modest number of atomics is worth the effort for a task that runs
     no more frequently than once a minute? Nah. :)
  let $clusters       := for $cluster in $clusters order by $cluster return $cluster
  let $hosts-in-group := for $host in $hosts-in-group order by $host return $host

  let $ordinal        := for $host at $index in $hosts-in-group
                         where $host = $this-host
                         return $index

  (: We only have 1440 minutes. What if there are more than 1440 clusters?
     Then do them in chunks. :)
  let $clusters       := if (count($clusters) gt 1440)
                         then
                           let $partitions := (count($clusters) idiv 1440) + 1
                           let $partition  := $minute mod $partitions
                           return
                             subsequence($clusters, ($partition*1440)+1, 1440)
                         else
                           $clusters

  let $cluster-count  := count($clusters)
  let $host-count     := count($hosts-in-group)

  let $_              := trace($cluster-count || " clusters: "
                               || string-join($clusters, ", ") || " for " || $type,
                               $t:trace-pselect)
  let $_              := trace($host-count || " hosts: "
                               || string-join($hosts-in-group ! string(.), ", ")
                               || " for " || $type,
                               $t:trace-pselect)
  let $_              := trace("Ordinal: " || $ordinal || " for " || $type,
                               $t:trace-pselect)

  (: Divide the clusters into partitions by number of hosts. Pick a partition :)
  let $partitions     := ($cluster-count + $host-count - 1) idiv $host-count
  let $partition      := $minute mod $partitions

  (: Work out how many interations we have to go through before we've done
     an entire cycle. :)
  let $iterations     := if ($cluster-count mod $host-count = 0)
                         then $cluster-count idiv $host-count
                         else ($cluster-count idiv $host-count) + 1

  (: Work out the offset for each iteration :)
  let $offset         := if ($host-count = 1)
                         then 0
                         else ($minute idiv $iterations) mod $iterations

  (: Pick one of the clusters! :)
  let $pick           := ($partition * $host-count) + $ordinal + $offset
  let $pick           := if ($pick gt $cluster-count)
                         then $pick - $cluster-count
                         else $pick

  let $_              := trace("Partition: " || $partition
                               || ", iterations: " || $iterations
                               || ", offset: " || $offset
                               || ", pick: " || $pick
                               || " for " || $type, $t:trace-pselect)

  return
    if ($ordinal gt $cluster-count)
    then
      ()
    else
      let $_ := trace("Picked cluster: " || $clusters[$pick] || " for " || $type,
                      $t:trace-pselect)
      return
        $clusters[$pick]
};
