xquery version "1.0-ml";

(: Copyright 2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace raw="http://marklogic.com/v1/opsdirector/rawdata";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare %private variable $Z := xs:dayTimeDuration("PT0S");

declare %private variable $QOPTS
  := ("collation=http://marklogic.com/collation/");

(: ====================================================================== :)

declare function raw:archive-resource(
  $node as object-node()?
) as empty-sequence()
{
  if (empty($node))
  then
    ()
  else
    (: Must be called with a node that is a reference to the in-database node :)
    let $updated := t:replace-property($node, "opsdirDtend", string(current-dateTime()))
    let $_ := trace("Archiving " || xdmp:node-uri($node), $t:trace-debug)
    return
      (xdmp:node-replace($node, $updated),
       xdmp:document-set-collections(xdmp:node-uri($node), ()))
};

(: ====================================================================== :)

declare function raw:resource-id(
  $cluster-id as xs:string?,
  $resource-type as xs:string,
  $resource-name as xs:string
) as xs:string?
{
  if ($resource-type = $t:RESOURCE-TYPES)
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  (: Hack; this substring happens to work today, might not always :)
  let $s     := substring($resource-type, 1, string-length($resource-type) - 1)
  let $pname := fn:QName("", $s || "Name")

  for $id in raw:resource-ids($cluster-id, $resource-type)
  let $obj  := raw:resource($cluster-id, $resource-type, $id)
  let $name := $obj/node()[node-name(.) = $pname]/string()
  where $name = $resource-name
  return
    $obj/opsdirId
};

declare function raw:resource-ids(
  $resource-type as xs:string?
) as xs:string*
{
  raw:resource-ids((), $resource-type)
};

declare function raw:resource-ids(
  $cluster-id as xs:string?,
  $resource-type as xs:string?
) as xs:string*
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  let $type-query := if (empty($resource-type))
                     then ()
                     else cts:element-range-query(
                              xs:QName("opsdirResourceType"), "=", $resource-type, $QOPTS)
  let $dt-query   := cts:element-range-query(
                         xs:QName("opsdirDtend"), ">", current-dateTime())
  let $clus-query := if (empty($cluster-id))
                     then
                       ()
                     else
                       cts:element-range-query(
                           xs:QName("opsdirClusterId"), "=", $cluster-id, $QOPTS)
  let $coll-query := cts:or-query((cts:collection-query($t:COLLROOT||"resource"),
                                   cts:collection-query($t:COLLROOT||"stats")))
  return
    cts:element-values(
      xs:QName("opsdirId"), (), (),
      cts:and-query(($type-query, $dt-query, $clus-query, $coll-query)))
};

declare function raw:resources-of-type(
  $cid as xs:string?,
  $type as xs:string
) as object-node()*
{
  let $ids := raw:resource-ids($cid, $type)
  return
    raw:resources($type, $ids)
};

(:
declare function raw:resources(
  $ids as xs:string*
) as object-node()*
{
  raw:resources((), $ids)
};
:)

declare function raw:resources(
  $resource-type as xs:string?,
  $ids as xs:string*
) as object-node()*
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  $ids ! raw:resource((), $resource-type, .)
};

declare function raw:resource(
  $cluster-id as xs:string?,
  $resource-type as xs:string?,
  $id as xs:string
) as object-node()?
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  let $type-query := if (empty($resource-type))
                     then
                       ()
                     else
                       cts:json-property-range-query(
                           "opsdirResourceType", "=", $resource-type, $QOPTS)

  let $dt-query  := cts:json-property-range-query("opsdirDtend", ">", current-dateTime())

  let $c-query   := if (empty($cluster-id))
                    then
                      ()
                    else
                      cts:json-property-range-query(
                          "opsdirClusterId", "=", $cluster-id, $QOPTS)

  let $id-query  := cts:json-property-range-query("opsdirId", "=", $id, $QOPTS)

  let $coll-query := cts:or-query((cts:collection-query($t:COLLROOT||"resource"),
                                   cts:collection-query($t:COLLROOT||"stats")))

  let $doc       := cts:search(collection(),
                               cts:and-query(($type-query, $dt-query, $c-query, $id-query, $coll-query)),
                               "unfiltered")
  let $_ := if (count($doc) gt 1)
            then xdmp:log(("Error: multiple resource documents for ('"
                          || $cluster-id || "','" || $resource-type || "','"
                          || $id || " (" || count($doc) || ")",
                          $doc ! xdmp:node-uri(.)))
            else ()
  return
    ($doc)[1]/node()
};
