xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace status="http://marklogic.com/v1/opsdirector/status";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare namespace host="http://marklogic.com/xdmp/status/host";
declare namespace pki="http://marklogic.com/xdmp/x509";
declare namespace html="http://www.w3.org/1999/xhtml";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $SESSION_COLLECTION  := "http://marklogic.com/opsdirector/v1/sessions";
declare variable $RESOURCE_COLLECTION := "http://marklogic.com/opsdirector/v1/resource";
declare variable $RGROUP_COLLECTION   := "http://marklogic.com/opsdirector/v1/resource-group";
declare variable $LOGS_COLLECTION     := "http://marklogic.com/opsdirector/v1/logs";
declare variable $ALERTS_COLLECTION   := "http://marklogic.com/opsdirector/v1/alerts";
declare variable $TAB := "&#9;";
declare variable $PT0S := xs:dayTimeDuration("PT0S");

declare variable $show-perms := (exists(xdmp:get-request-field("show-perms"))
                                 and string(xdmp:get-request-field("show-perms")) ne "false");

declare function status:html(
)
{
  if (string(config:get-atomic-property("status-enabled")) = "true")
  then
    status:amped-html()
  else
    (xdmp:set-response-code(403, "Forbidden"),
     <html xmlns="http://www.w3.org/1999/xhtml">
       <head>
         <title>Forbidden</title>
       </head>
       <body>
         <p>Request forbidden</p>
       </body>
     </html>)
};

declare function status:amped-html(
)
{
  let $uri := xdmp:get-request-field("uri")
  return
    if ($uri and doc-available($uri))
    then
      (xdmp:set-response-content-type("application/json"),
       doc($uri))
    else
      <html xmlns="http://www.w3.org/1999/xhtml">
        <head>
          <title>OpsDirector Status</title>
          <link rel="stylesheet" type="text/css" href="/status.css" />
          <script src="https://code.jquery.com/jquery-3.2.1.min.js"
                  integrity="sha256-hwg4gsxgFZhOsEEamdOYGBf13FyQuiTwlAQgxVSNgt4="
                  crossorigin="anonymous"></script>
          <script src="/status.js"></script>
        </head>
        <body>
          <h1>OpsDirector Status</h1>
          <p>
            {format-dateTime(current-dateTime(),
                             "[D01] [MNn,*-3] [Y0001] [h01]:[m01]:[s01][P]")}
          </p>

          <p>
            { if ($show-perms)
              then <a href="/status?show-perms=false">hide document permissions</a>
              else <a href="/status?show-perms=true">show document permissions</a>
            }
          </p>

          { status:html-data() }

          <p>End of status report</p>
        </body>
      </html>
};

declare function status:html-alerts(
  $alerts
) as element(html:tr)*
{
  let $rsrcids := $alerts/resourceId
  let $rsrcs   := t:managed-resources((), $rsrcids)
  let $sorted := for $alert in $alerts
                 order by $alert/opsdirDtstart descending
                 return $alert
  let $rg     := collection($RGROUP_COLLECTION)/object-node()
  return
    if (empty($alerts))
    then
      ()
    else
      (<tr xmlns="http://www.w3.org/1999/xhtml">
         <th>Ack</th>
         <th>Status</th>
         <th>State</th>
         <th>Code</th>
         <th>Type</th>
         <th>Resource</th>
         <th>Cluster</th>
         <th>Message</th>
       </tr>,
       <tr xmlns="http://www.w3.org/1999/xhtml" class="select-alerts">
         <td>
           <select id="alert-ack">
             <option value="all"> </option>
             <option value="no">☐</option>
             <option value="yes">🗹</option>
           </select>
         </td>
         <td>
           <select id="alert-status">
             <option value="all"> </option>
             { for $status in distinct-values($alerts/*:status)
               order by $status
               return
                 <option value="{$status}">{$status}</option>
             }
           </select>
         </td>
         <td>
           <select id="alert-state">
             <option value="all"> </option>
             { for $state in distinct-values($alerts/*:state)
               order by $state
               return
                 <option value="{$state}">{$state}</option>
             }
           </select>
         </td>
         <td>
           <select id="alert-code">
             <option value="all"> </option>
             { for $code in distinct-values($alerts/*:code)
               order by $code
               return
                 <option value="{$code}">{$code}</option>
             }
           </select>
         </td>
         <td>
           <select id="alert-type">
             <option value="all"> </option>
             { for $value in distinct-values($alerts/*:resourceType)
               order by $value
               return
                 <option value="{$value}">{$value}</option>
             }
           </select>
         </td>
         <td>
           <select id="alert-resource">
             <option value="all"> </option>
             { for $value in distinct-values($rsrcs/*:opsdirName)
               order by $value
               return
                 <option value="{$value}">{$value}</option>
             }
           </select>
         </td>
         <td>
           <select id="alert-cluster">
             <option value="all"> </option>
             { for $value in distinct-values($rsrcs/*:opsdirClusterName)
               order by $value
               return
                 <option value="{$value}">{$value}</option>
             }
           </select>
         </td>
         <td></td>
       </tr>,
       for $alert at $index in $sorted
       let $cid := string($alert/opsdirClusterId)
       let $type := string($alert/resourceType) || "s"
       let $code := string($alert/code)
       let $status := string($alert/status)
       let $ack := if (string($alert/acknowledged) = "no")
                   then " "
                   else "🗹"
       let $state := string($alert/state)
       let $message := string($alert/message)
       let $id := string($alert/resourceId)
       let $rsrc := t:managed-resource((), $type, $id)
       let $class := if (empty($rsrc)) then "error" else status:access-classes($rsrc, $rg)
       let $name := $rsrc/opsdirName
       return
         (<tr xmlns="http://www.w3.org/1999/xhtml">
            { attribute { fn:QName("","class") }
              { string-join(("alert", $class, "status-"||$status,
                             "ack-"||string($alert/*:acknowledged),
                             "state-"||$state, "code-"||$code,
                             "type-"||$type, "resource-"||$name ), " ") } }
            <td align="center">
              <span>
                { if ($alert/*:acknowledged = "yes")
                  then
                    attribute { fn:QName("", "title") } { $alert/*:opsdirDtend }
                  else
                    ()
                }
                {$ack}
              </span>
            </td>
            <td>
              <span title="{ $alert/*:opsdirDtstart }">
                {$status}
              </span>
            </td>
            <td>{$state}</td>
            <td>{$code}</td>
            <td>{$type}</td>
            <td>{status:link-by-id($type, $id)}</td>
            <td>{status:link-by-id("clusters", $cid)}</td>
            <td>{$message}</td>
          </tr>,
          if ($show-perms)
          then
            <tr xmlns="http://www.w3.org/1999/xhtml">
              <td colspan="3"> </td>
              <td colspan="5">
                <a href="/status?uri={xdmp:node-uri($alert)}"
                   style="text-decoration: none">•</a>
                {status:readable(xdmp:node-uri($alert))}
              </td>
            </tr>
          else
            ())
       )
};

declare %private function status:html-data(
) as element(html:div)
{
  <div xmlns="http://www.w3.org/1999/xhtml">
    <h2>Active sessions</h2>

    <table cellpadding="5" cellspacing="0" border="1">
      <tr>
        <th>Cluster</th>
        <th>Host</th>
        <th>Version</th>
        <th>Duration</th>
        <th>Expires</th>
      </tr>
      { status:html-sessions() }
    </table>

    <h2>Resources</h2>
    { status:html-resources() }

    <h2>Config</h2>
    <table cellpadding="5" cellspacing="0" border="1">
      <tr><th>Property</th><th>Value</th></tr>
      { for $prop in doc("/config/opsdirector.xml")/*/*
        return
          <tr>
            <td>{local-name($prop)}</td>
            <td>{string($prop)}</td>
          </tr>
      }
    </table>

    <h2>Credentials</h2>
    <table cellpadding="5" cellspacing="0" border="1">
      {
        try { xdmp:invoke-function(
                function () { status:html-credentials() },
                <options xmlns="xdmp:eval">
                  <database>{xdmp:security-database()}</database>
                </options>)
             } catch ($e) {
               if ($e/error:code eq 'SEC-PRIV')
               then
                 "Insufficient privileges to show credentials"
               else
                 xdmp:rethrow()
             }
      }
    </table>

    { let $alerts := cts:search(/, cts:collection-query($ALERTS_COLLECTION))
      where exists($alerts)
      return
        (<h2>Alerts</h2>,
         <table cellspacing="0" cellpadding="5" border="1">
           <tbody>
             { status:html-alerts($alerts) }
           </tbody>
         </table>)
    }
  </div>
};

declare function status:html-credentials(
) as element(html:tr)*
{
  let $creds := for $id in sec:get-credential-ids()
                let $cred := sec:get-credential-by-id($id)
                let $name := $cred/sec:credential-name
                where (contains($name, "OpsDirector") or contains($name, "opsdir"))
                order by $name
                return
                  $cred
  return
    if (empty($creds))
    then
      <tr xmlns="http://www.w3.org/1999/xhtml">
        <td>No sessions</td>
      </tr>
    else
      (<tr xmlns="http://www.w3.org/1999/xhtml">
         <th>Credential</th>
         <th>Common Name</th>
         <th>Valid from</th>
         <th>Valid to</th>
       </tr>,
       for $cred in $creds
       let $x509    := xdmp:x509-certificate-extract($cred/sec:credential-certificate)
       let $name    := string($cred/sec:credential-name)
       let $common  := string($x509/pki:issuer/pki:commonName)
       let $dtstart := xs:dateTime($x509/pki:validity/pki:notBefore)
       let $dtend   := xs:dateTime($x509/pki:validity/pki:notAfter)
       let $istart  := if ($dtstart lt current-dateTime())
                       then attribute { fn:QName("","style") }
                                      { "text-decoration: line-through
                                         background-color: #ffaaaa;" }
                       else ()
       let $iend    := if ($dtend lt current-dateTime())
                       then attribute { fn:QName("","style") }
                                      { "text-decoration: line-through;
                                         background-color: #ffaaaa;" }
                       else ()
       return
         <tr xmlns="http://www.w3.org/1999/xhtml">
           <td>{$name}</td>
           <td>{$common}</td>
           <td>{$istart}{$dtstart}</td>
           <td>{$iend}{$dtend}</td>
        </tr>)
};

declare function status:html-sessions(
) as element(html:tr)*
{
  let $sessions := cts:search(/, cts:collection-query($SESSION_COLLECTION))
  return
    if (empty($sessions))
    then
      ()
    else
      for $session at $index in $sessions
      let $cid := $session/clusterId
      let $hid := $session/hostId
      let $ver := $session/serverVersion
      let $expires := xs:dateTime($session/expires)
      let $ete := $expires - current-dateTime()
      let $ete := if ($ete lt $PT0S)
                  then $PT0S - $ete
                  else $ete
      let $h   := hours-from-duration($ete)
      let $m   := minutes-from-duration($ete)
      let $dur := (config:get-atomic-property("session-duration-" || $cid),
                   config:get-atomic-property("session-duration"),
                   "“default”")[1]
      let $time := if (($expires - current-dateTime()) lt $PT0S)
                   then "EXPIRED"
                   else
                     if ($h = 0)
                     then format-number($m, "   ~00m")
                     else format-number($h, "~00h") || format-number($m, "00m")
      return
        <tr xmlns="http://www.w3.org/1999/xhtml">
          <td>{ status:link-by-id("clusters", $cid) }</td>
          <td>{ status:link-by-id("hosts", $hid) }</td>
          <td>{ $ver }</td>
          <td>{ $dur }</td>
          <td>{ $time }</td>
        </tr>
};

declare function status:html-resources(
) as element(html:div)*
{
  let $cluster-ids := t:managed-resource-ids((), "clusters")
  let $rg          := collection($RGROUP_COLLECTION)/object-node()
  return
    (if (exists($rg))
     then
       <div xmlns="http://www.w3.org/1999/xhtml" class="resource-groups">
         <h2>{count($rg)} Resource Groups</h2>
         { status:html-resource-groups($rg) }
      </div>
     else
       (),
     for $cluster-id in $cluster-ids
     let $cluster := t:managed-resource($cluster-id, "clusters", $cluster-id)
     let $rcount  := xdmp:estimate(cts:search(collection($RESOURCE_COLLECTION),
                         cts:json-property-value-query("opsdirClusterId", $cluster-id)))
     order by $cluster/opsdirClusterName
     return
        <div xmlns="http://www.w3.org/1999/xhtml" id="{$cluster-id}"
             class="cluster {if (empty($cluster)) then 'error' else ''}" >
          <h2>Cluster {$cluster/*:opsdirName}
              (<code>{$cluster-id}</code>; {$rcount} resources)</h2>
          { if (empty($cluster))
            then
              <p>No cluster resource exists in the database!</p>
            else
              <div>
                <ul>
                  <li>Related resources: {status:related-resources($cluster-id)}</li>
                </ul>
                { status:html-groups($cluster) }
                { status:html-hosts($cluster, $rg) }
                { status:html-servers($cluster, $rg) }
                { status:html-databases($cluster, $rg) }
                { status:html-forests($cluster, $rg) }
              </div>
           }
         </div>)
};

declare function status:html-groups(
  $cluster as object-node()
) as element(html:div)
{
  let $group-ids := t:managed-resource-ids($cluster/opsdirClusterId, "groups")
  let $groups    := t:managed-resources($group-ids)
  return
    <div xmlns="http://www.w3.org/1999/xhtml" class="groups">
      <h3>{count($groups) || " " || (if (count($groups) = 1) then "Group" else "Groups")}</h3>
      { for $group in $groups
        let $id := $group/*:opsdirId
        order by $group/*:opsdirName
        return
          <div class="group" id="{$id}">
            <h4>Group <span class="name">{$group/*:opsdirName}</span>
                (<code>{$group/*:opsdirId}</code>)</h4>
            <ul>
              <li>Config timestamp: {status:duration($group/*:opsdirDtstart)}</li>
              <li>Cluster: <a href="#{$cluster/*:opsdirId}">{$cluster/*:opsdirName}</a></li>
              <li>
                { let $hosts := cts:search(collection($RESOURCE_COLLECTION),
                                  cts:and-query(
                                    (cts:json-property-value-query("opsdirClusterId",
                                                                   $group/*:opsdirClusterId),
                                     cts:json-property-value-query("opsdirResourceType",
                                                                   "hosts"),
                                     cts:json-property-value-query("opsdirGroupName",
                                                                   $group/*:opsdirName))))
                                /object-node()
                  return
                    (count($hosts) || " "
                     || (if (count($hosts) = 1) then "host" else "hosts"),
                     if (exists($hosts))
                     then
                       <ul>
                         { for $host in $hosts
                           order by $host/*:opsdirName
                           return
                             <li>
                               <a href="#{$host/*:opsdirId}">{$host/*:opsdirName}</a>
                             </li>
                         }
                      </ul>
                     else
                       ())
                }
              </li>
              <li>Related resources: {status:related-resources($id)}</li>
            </ul>
          </div>

      }
    </div>
};

declare function status:html-hosts(
  $cluster as object-node(),
  $resgroup as object-node()*
) as element(html:div)
{
  let $host-ids := t:managed-resource-ids($cluster/opsdirClusterId, "hosts")
  let $hosts    := t:managed-resources($host-ids)
  return
    <div xmlns="http://www.w3.org/1999/xhtml" class="hosts">
      <h3>{count($hosts) || " " || (if (count($hosts) = 1) then "Host" else "Hosts")}</h3>
      { for $rsrc in $hosts
        let $id := $rsrc/*:opsdirId
        let $id   := string($rsrc/*:opsdirId) (: I HATE XQUERY DEFAULT NS HANDLING :)
        let $cid  := string($rsrc/*:opsdirClusterId)
        let $dt   := string($rsrc/*:opsdirDtstart)
        let $suri := "/config/host-status/" || $cid || "/" || $id || "/snapshot.json"
        let $stat := doc($suri)/object-node()
        let $logs := xdmp:estimate(cts:search(collection($LOGS_COLLECTION),
                        cts:and-query(
                           (cts:json-property-value-query("opsdirClusterId", $cid),
                            cts:json-property-value-query("opsdirHostId", $id)))))
        order by $rsrc/*:opsdirName
        return
          <div class="resource host {status:access-classes($rsrc,$resgroup)}
                      {if (not(starts-with($stat/*:status, 'ok'))) then 'error' else ''}"
               id="{$id}">
            <h4>Host <span class="name">{$rsrc/*:opsdirName}</span>
                (<code>{$rsrc/*:opsdirId}</code>)</h4>
            <ul>
              <li>Config timestamp: {status:duration($rsrc/*:opsdirDtstart)}</li>
              <li>Cluster: <a href="#{$cluster/*:opsdirId}">{$cluster/*:opsdirName}</a></li>
              <li>Status timestamp:
                { if ($stat/*:ts)
                  then
                    <span>{string($stat/*:status)}:
                      <span title="{$stat/*:ts}">
                        { string($stat/*:ts) }
                        { " (" }
                        { status:duration($stat/*:ts) }
                        { " ago)" }
                      </span>
                    </span>
                  else
                    <span>-</span>
                }
              </li>
              { status:access($rsrc, $resgroup) }
              <li>Logs: {$logs}
                 { if ($logs gt 0)
                   then ", the most recent 10:"
                   else ""
                 }
                 { let $messages
                     := cts:search(collection($LOGS_COLLECTION),
                          cts:and-query(
                             (cts:json-property-value-query("opsdirClusterId", $cid),
                              cts:json-property-value-query("opsdirHostId", $id))),
                          cts:index-order(
                             cts:element-reference(fn:QName("","hostTs")), "descending"))
                        [1 to 10]
                   return
                     <ul>
                       { for $msg in $messages
                         let $text := $msg/object-node()/*:message
                         let $text := if (string-length($text) gt 80)
                                      then substring($text,1,80)||"…"
                                      else $text
                         let $date := $msg/object-node()/*:hostTs
                         return
                           <li>
                             { format-dateTime(xs:dateTime($date),
			                    "[Y0001]-[M01]-[D01] [H01]:[m01] ") }
                             <a href="/status?uri={xdmp:node-uri($msg)}">
                               {string($text)}
                             </a>
                             {status:readable(xdmp:node-uri($msg))}
                           </li>
                       }
                     </ul>
                 }
              </li>
              <li>Related resources: {status:related-resources($id)}</li>
              <li>Tasks: {status:tasks($rsrc)}</li>
            </ul>
          </div>
      }
    </div>
};

declare function status:html-databases(
  $cluster as object-node(),
  $resgroup as object-node()*
) as element(html:div)
{
  let $ids       := t:managed-resource-ids($cluster/opsdirClusterId, "databases")
  let $rsrcs     := t:managed-resources($ids)
  return
    <div xmlns="http://www.w3.org/1999/xhtml" class="databases">
      <h3>{count($rsrcs) || " "
           || (if (count($rsrcs) = 1) then "Database" else "Databases")}</h3>
      { for $rsrc in $rsrcs
        let $id := $rsrc/*:opsdirId
        let $id   := string($rsrc/*:opsdirId) (: I HATE XQUERY DEFAULT NS HANDLING :)
        let $cid  := string($rsrc/*:opsdirClusterId)
        let $dt   := string($rsrc/*:opsdirDtstart)
        let $forests := cts:search(collection($RESOURCE_COLLECTION),
                          cts:and-query(
                            (cts:json-property-value-query("opsdirClusterId",
                                                           $cluster/*:opsdirId),
                             cts:json-property-value-query("opsdirResourceType",
                                                           "forests"),
                             cts:json-property-value-query("database",
                                                           $rsrc/*:opsdirName))))
                        /object-node()
        order by $rsrc/*:opsdirName
        return
          <div class="resource database {status:access-classes($rsrc,$resgroup)}" id="{$id}">
            <h4>Database <span class="name">{$rsrc/*:opsdirName}</span>
                (<code>{$rsrc/*:opsdirId}</code>)</h4>
            <ul>
              <li>Config timestamp: {status:duration($rsrc/*:opsdirDtstart)}</li>
              <li>Cluster: <a href="#{$cluster/*:opsdirId}">{$cluster/*:opsdirName}</a></li>
              { status:access($rsrc, $resgroup) }
              <li>{count($forests) || " "
                   || (if (count($forests) = 1) then "forest" else "forests")}
                { if (exists($forests))
                  then
                    <ul>
                      { for $forest in $forests
                        return
                          <li>
                            <a href="#{$forest/*:opsdirId}">
                              {$forest/*:opsdirName}
                            </a>
                            (on host {status:link-by-name($cluster, "hosts", $forest/*:host)})
                          </li>
                      }
                    </ul>
                  else
                    ()
                }
              </li>
              <li>Related resources: {status:related-resources($id)}</li>
            </ul>
          </div>
      }
    </div>
};

declare function status:html-forests(
  $cluster as object-node(),
  $resgroup as object-node()*
) as element(html:div)
{
  let $ids   := t:managed-resource-ids($cluster/opsdirClusterId, "forests")
  let $rsrcs := t:managed-resources($ids)
  return
    <div xmlns="http://www.w3.org/1999/xhtml" class="forests">
      <h3>{count($rsrcs) || " "
           || (if (count($rsrcs) = 1) then "Forest" else "Forests")}</h3>
      { for $rsrc in $rsrcs
        let $id := $rsrc/*:opsdirId
        let $id   := string($rsrc/*:opsdirId) (: I HATE XQUERY DEFAULT NS HANDLING :)
        let $cid  := string($rsrc/*:opsdirClusterId)
        let $dt   := string($rsrc/*:opsdirDtstart)
        order by $rsrc/*:opsdirName
        return
          <div class="resource forest {status:access-classes($rsrc,$resgroup)}" id="{$id}">
            <h4>Forest <span class="name">{$rsrc/*:opsdirName}</span>
                (<code>{$rsrc/*:opsdirId}</code>)</h4>
            <ul>
              <li>Config timestamp: {status:duration($rsrc/*:opsdirDtstart)}</li>
              <li>Cluster: <a href="#{$cluster/*:opsdirId}">{$cluster/*:opsdirName}</a></li>
              { status:access($rsrc, $resgroup) }
              <li>Host: {status:link-by-name($cluster, "hosts", $rsrc/*:host)}</li>
              <li>Database:
                { if ($rsrc/*:database)
                  then status:link-by-name($cluster, "databases", $rsrc/*:database)
                  else <i>None</i>
                }
              </li>
              <li>Storage:
                {
                  let $storage := cts:search(collection($RESOURCE_COLLECTION),
                                    cts:and-query(
                                       (cts:json-property-value-query("opsdirId", $id),
                                        cts:json-property-value-query("opsdirResourceType",
                                                                      "forest-storage"))))
                                  /object-node()
                  let $locs := ($storage/*:storageListItems/*:storageHost
                                  /*:locations/*:location
                                |$storage/*:storageListItems/*:storageHost/*:location)
                  let $fids := ($locs/*:locationForests/*:locationForest
                                |$locs/*:locationForest)
                  return
                    if (empty($locs) or empty($fids))
                    then "-"
                    else
                      <span>
                        { count($fids) }
                        { if (count($fids) ne 1) then " hosts" else " host" }
                      </span>
                }
              </li>
              <li>Related resources: {status:related-resources($id)}</li>
            </ul>
          </div>
      }
    </div>
};

declare function status:html-servers(
  $cluster as object-node(),
  $resgroup as object-node()*
) as element(html:div)
{
  let $ids       := t:managed-resource-ids($cluster/opsdirClusterId, "servers")
  let $rsrcs     := t:managed-resources($ids)
  let $group-ids := t:managed-resource-ids($cluster/opsdirClusterId, "groups")
  let $groups    := t:managed-resources($group-ids)
  return
    <div xmlns="http://www.w3.org/1999/xhtml" class="servers">
      <h3>{count($rsrcs) || " "
           || (if (count($rsrcs) = 1) then "Server" else "Servers")}</h3>
      { for $rsrc in $rsrcs
        let $id := $rsrc/*:opsdirId
        let $id   := string($rsrc/*:opsdirId) (: I HATE XQUERY DEFAULT NS HANDLING :)
        let $cid  := string($rsrc/*:opsdirClusterId)
        let $dt   := string($rsrc/*:opsdirDtstart)
        let $group := $groups[*:opsdirName = $rsrc/*:opsdirGroupName]
        order by $rsrc/*:opsdirName
        return
          <div class="resource server {status:access-classes($rsrc,$resgroup)}" id="{$id}">
            <h4>Server {$rsrc/*:opsdirName} (<code>{$rsrc/*:opsdirId}</code>)</h4>
            <ul>
              <li>Config timestamp: {status:duration($rsrc/*:opsdirDtstart)}</li>
              <li>Cluster: <a href="#{$cluster/*:opsdirId}">{$cluster/*:opsdirName}</a></li>
              <li>
                Group: <a href="#{$group/*:opsdirId}">{$group/*:opsdirName}</a>
              </li>
              { status:access($rsrc, $resgroup) }
              <li>Related resources: {status:related-resources($id)}</li>
            </ul>
          </div>
      }
    </div>
};

declare function status:html-resource-groups(
  $resgroup as object-node()*
) as element(html:div)+
{
  for $rsrc in $resgroup
  let $id     := $rsrc/opsdirId
  let $access := for $rid in $rsrc/*:groupPermissions
                 let $name := status:role-name(xs:unsignedLong($rid/*:role))
                 where not(starts-with($name, "opsdir-rg-internal-"))
                 return
                   <li>
                     {$name} ({ $rid/*:capability })
                   </li>
  order by $rsrc/name
  return
    <div xmlns="http://www.w3.org/1999/xhtml" id="{$id}"
         class="resource-group {if (empty($access)) then 'warning' else ''}">
      <h4>Resource Group <span class="name">{$rsrc/*:name}</span> (<code>{$id}</code>)
        <input type="checkbox" id="rg-{$id}"/>
      </h4>
      <ul>
        <li>Kind: {$rsrc/*:opsdirResourceType}</li>
        <li>Role: {status:role-name(xs:unsignedLong($rsrc/*:relatedRole))}</li>
        <li>Size: {string($rsrc/*:size)}</li>
        <li>Access:
          <ul>
            { for $rid in $rsrc/*:groupPermissions
              let $name := status:role-name(xs:unsignedLong($rid/*:role))
              where not(starts-with($name, "opsdir-rg-internal-"))
              return
                <li>
                  {$name} ({ $rid/*:capability })
                </li>
            }
          </ul>
        </li>
        <li>Resources:
          { if ($rsrc/*:size gt 0)
            then
              <ul>
                { for $rid in $rsrc/*:resource
                  let $r := t:managed-resource((), $rsrc/*:opsdirResourceType, $rid)
                  return
                    <li>
                      { status:link-by-id($rsrc/*:opsdirResourceType, $rid) }
                      on
                      { status:link-by-id("clusters", $r/*:opsdirClusterId) }
                    </li>
                }
              </ul>
            else
              ()
          }
        </li>
      </ul>
    </div>
};

declare %private function status:access-classes(
  $rsrc as object-node(),
  $resgroup as object-node()*
) as xs:string*
{
  let $uri  := xdmp:node-uri($rsrc)
  let $perm := xdmp:document-get-permissions($uri)
  let $groles := for $id in $perm/*:role-id
                 let $role := status:role-name($id)
                 where starts-with($role, "opsdir-rg-internal-")
                 return string($id)
  let $groups := for $group in $resgroup
                 where $group/*:relatedRole = $groles
                 return
                   $group/*:name
  where exists($groups)
  return
    let $groups := for $group in distinct-values($groups)
                   order by $group return $group
    for $group at $index in $groups
    for $rg in $resgroup[*:name = $group]
    return
      "rg-"||$rg/*:opsdirId
};

declare %private function status:access(
  $rsrc as object-node(),
  $resgroup as object-node()*
) as element(html:li)+
{
  let $uri  := xdmp:node-uri($rsrc)
  let $perm := xdmp:document-get-permissions($uri)
  let $groles := for $id in $perm/*:role-id
                 let $role := status:role-name($id)
                 where starts-with($role, "opsdir-rg-internal-")
                 return string($id)
  let $groups := for $group in $resgroup
                 where $group/*:relatedRole = $groles
                 return
                   $group/*:name
  return
    (<li xmlns="http://www.w3.org/1999/xhtml">
       { "Resource groups: " }
       { if (empty($groups))
         then "-"
         else
           let $groups := for $group in distinct-values($groups)
                          order by $group return $group
           for $group at $index in $groups
           let $rg := $resgroup[*:name = $group]
           return
             (if ($index gt 1) then ", " else "",
              <a href="#{$rg/*:opsdirId}">{$group}</a>)
       }
     </li>,
     if (empty($groups))
     then
       ()
     else
       <li xmlns="http://www.w3.org/1999/xhtml">
         { "Access roles: " }
         { let $roles := for $role in $perm ! status:role-name(./*:role-id)
                         where not($role = ("opsdir-data-internal", "opsdir-admin"))
                               and not(starts-with($role, "opsdir-rg-internal-"))
                         return $role
           return
             if (empty($roles))
             then "-"
             else
               string-join(
                  distinct-values(for $role in $roles order by $role return $role),
                  ", ")
         }
       </li>)
};

declare function status:duration(
  $stamp as xs:string
) as xs:string
{
  let $dt  := xs:dateTime($stamp)
  let $dur := current-dateTime() - $dt
  let $d   := days-from-duration($dur)
  let $h   := hours-from-duration($dur)
  let $m   := minutes-from-duration($dur)
  let $s   := math:trunc(seconds-from-duration($dur))
  return
    if ($d gt 0)
    then
      $d || " days " || format-number($h, "00") || ":" || format-number($m, "00")
      || ":" || format-number($s, "00")
    else
      format-number($h, "00") || ":" || format-number($m, "00")
      || ":" || format-number($s, "00")
};

declare function status:link-by-name(
  $cluster as object-node(),
  $kind as xs:string,
  $name as xs:string
) as element()
{
  let $rsrc := cts:search(collection($RESOURCE_COLLECTION),
                 cts:and-query(
                   (cts:json-property-value-query("opsdirClusterId", $cluster/*:opsdirId),
                    cts:json-property-value-query("opsdirResourceType", $kind),
                    cts:json-property-value-query("opsdirName", $name))))
  return
    if (count($rsrc) = 0)
    then
      <span xmlns="http://www.w3.org/1999/xhtml">-</span>
    else if (count($rsrc) gt 1)
    then
      <span xmlns="http://www.w3.org/1999/xhtml">(not unique?)</span>
    else
      <a xmlns="http://www.w3.org/1999/xhtml" href="#{$rsrc/*:opsdirId}">
        {$rsrc/*:opsdirName}
      </a>
};

declare function status:link-by-id(
  $kind as xs:string,
  $id as xs:string
) as element()
{
  let $rsrc := cts:search(collection($RESOURCE_COLLECTION),
                 cts:and-query(
                   (cts:json-property-value-query("opsdirResourceType", $kind),
                    cts:json-property-value-query("opsdirId", $id))))
  return
    if (count($rsrc) = 0)
    then
      <span xmlns="http://www.w3.org/1999/xhtml">-</span>
    else if (count($rsrc) gt 1)
    then
      <span xmlns="http://www.w3.org/1999/xhtml">(not unique?)</span>
    else
      <a xmlns="http://www.w3.org/1999/xhtml" href="#{$rsrc/*:opsdirId}">
        {$rsrc/*:opsdirName}
      </a>
};

declare function status:related-resources(
  $id as xs:string
) as element(html:ul)
{
  <ul xmlns="http://www.w3.org/1999/xhtml" class="urilist">
    { for $uri in cts:uris((), ("ascending"),
                           cts:and-query(
                             (cts:collection-query($RESOURCE_COLLECTION),
                              cts:json-property-value-query("opsdirId", $id))))
      order by $uri
      return
        <li>
          <a href="/status?uri={$uri}">{$uri}</a>
          {status:readable($uri)}
        </li>
    }
  </ul>
};

declare function status:tasks(
  $host as object-node()
) as element()
{
  let $cluster := $host/opsdirClusterId
  let $gname   := $host/opsdirGroupName 
  let $group   := t:managed-resource-ids($cluster, "groups")
                    ! t:managed-resource($cluster, "groups", string(.))[opsdirName = $gname]
  let $gid     := $group/opsdirId
  let $tasks   := t:managed-resource-ids($cluster, "tasks")
                    ! t:managed-resource($cluster, "tasks", string(.))[opsdirGroupId = $gid]
  return
    if ($tasks)
    then
      <ul xmlns="http://www.w3.org/1999/xhtml" class="urilist">
        { for $task in $tasks
          return
            <li>
              <a href="/status?uri={xdmp:node-uri($task)}">
                {xdmp:node-uri($task)}
              </a>
              {status:readable(xdmp:node-uri($task))}
            </li>
        }
      </ul>
    else
      <i xmlns="http://www.w3.org/1999/xhtml">None</i>
};

declare function status:readable(
  $uri as xs:string
) as element()
{
  let $perm := xdmp:document-get-permissions($uri)[./sec:capability = "read"]
  where $show-perms and $perm
  return
    <span xmlns="http://www.w3.org/1999/xhtml">
      { " (" }
      { string-join(distinct-values(for $name in $perm/sec:role-id ! status:role-name(.)
                                    order by $name
                                    return $name), ", ") }
      { ")" }
    </span>
};

declare function status:role-name(
  $id as xs:unsignedLong
) as xs:string
{
  try {
    xdmp:role-name($id)
  } catch ($e) {
    "ERR-RDN-" || $id
  }
};
