xquery version "1.0-ml";

(: Copyright 2018 MarkLogic Corporation.  All Rights Reserved. :)

(: This library contains upgrade functions :)

module namespace upgrade="http://marklogic.com/v1/opsdirector/upgrade";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace admin="http://marklogic.com/xdmp/admin"
    at "/MarkLogic/admin.xqy";

import module namespace manage="http://marklogic.com/manage"
    at "/MarkLogic/manage/manage.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace j="http://marklogic.com/v1/opsdirector/joins"
    at "/common/joins.xqy";

import module namespace odi="http://marklogic.com/opsdir/admin/lib"
    at "/MarkLogic/Admin/lib/opsdir-info.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare variable $LOCK     := "/config/upgrade-lock";
declare variable $LOCKJS   := "/config/upgrade.js";
declare variable $LOCKBODY := object-node { "message": "Upgrade in progress" };
declare variable $LOCKPERM :=
  (xdmp:permission("opsdir-user", "read"),
   xdmp:permission("opsdir-admin", "update"));

declare option xdmp:mapping "false";

declare function upgrade:upgrade(
)
{
  xdmp:lock-for-update($LOCK),
  if (upgrade:check-finished())
  then
    xdmp:log("Ops Director Upgrade finished")
  else
    let $_ := xdmp:log("Starting Ops Director upgrade")
    return
      xdmp:spawn-function(
        function () {
          (: There's only one possibility in 9.0-5 :)
          upgrade:from-904()
        },
        <options xmlns="xdmp:eval">
          <transaction-mode>update-auto-commit</transaction-mode>
        </options>)
};

declare function upgrade:in-progress(
) as xs:boolean
{
  xdmp:lock-for-update($LOCK),
  if (upgrade:ready())
  then
    (if (doc-available($LOCKJS))
     then
       upgrade:complete-upgrade()
     else
       (),
    false())
  else
    doc-available($LOCKJS)
};

declare %private function upgrade:check-finished(
) as xs:boolean
{
  let $ready := upgrade:ready()
  let $_ := xdmp:log("Ops Director upgrade: ready="
                     ||$ready||" lockjs="||doc-available($LOCKJS))
  return
    if (doc-available($LOCKJS))
    then
      if ($ready)
      then
        (upgrade:complete-upgrade(),
         true())
      else
        error((), "OD-UPGINP", "Ops Director upgrade in progress")
  else
    if ($ready)
    then
      true()
    else
      (xdmp:document-insert($LOCKJS, $LOCKBODY, $LOCKPERM),
       false())
};

declare %private function upgrade:complete-upgrade(
) as empty-sequence()
{
  let $_ := config:set-atomic-property("version", $t:OPSDIRECTOR-VERSION)
  let $_ := xdmp:document-delete($LOCKJS)
  return
    j:update-joins()
};

declare %private function upgrade:ready(
) as xs:boolean
{
  try {
    let $_ := xdmp:estimate(
                cts:search(/,
                  cts:json-property-range-query(
                    "activeFragmentCount", "=", 1)))
    return
      true()
  } catch ($e) {
    false()
  }
};

declare %private function upgrade:from-904()
{
  let $_ := xdmp:log("Upgrading Ops Director from 1.0 to 1.1")
  let $config := admin:get-configuration()
  let $dbid   := xdmp:database()
  let $config := fold-left(
                   function($config, $index) {
                     admin:database-delete-range-element-index($config, $dbid, $index)
                   }, $config, admin:database-get-range-element-indexes($config, $dbid))

  (: Let's generate the indexes, because it's clearer ... :)
  let $reidx := <range-element-indexes xmlns="http://marklogic.com/manage">
                  { for $name in ("created", "hostTs",
                                  "opsdirDtend", "opsdirDtstart",
                                  "timestamp", "ts", "lastStartup", "lastBackup")
                    return
                      <range-element-index>
                        <collation></collation>
                        <localname>{$name}</localname>
                        <invalid-values>ignore</invalid-values>
                        <namespace-uri></namespace-uri>
                        <range-value-positions>false</range-value-positions>
                        <scalar-type>dateTime</scalar-type>
                      </range-element-index>
                  }

                  { for $name in ("acknowledged", "availability", "clusterId",
                                  "contentDatabase", "database", "databaseName",
                                  "databaseReplication", "description", "effectiveVersion",
                                  "enabled", "encryption",
                                  "failoverEnable",
                                  "forest",
                                  "forestName", "group", "groupName", "ha", "hostId",
                                  "hostName", "hostNames",
                                  "keystoreKmsType",
                                  "level", "licenseKey",
                                  "maintenanceMode", "message", "modulesDatabase", "name",
                                  "opsdirClusterId", "opsdirClusterName",
                                  "opsdirGroupName", "opsdirHostId",
                                  "opsdirHostName", "opsdirId", "opsdirName",
                                  "opsdirResourceName", "opsdirResourceType",
                                  "opsdirUid", "osVersion", "path", "priority", "resourceId",
                                  "resourceName", "resourceType", "root",
                                  "security", "securityDatabase", "schemaDatabase",
                                  "serverName", "serverType", "sslHostname",
                                  "code", "state", "status", "triggersDatabase",
                                  "type", "user", "version", "zone")
                    return
                      <range-element-index>
                        <collation>http://marklogic.com/collation/</collation>
                        <localname>{$name}</localname>
                        <namespace-uri></namespace-uri>
                        <range-value-positions>true</range-value-positions>
                        <invalid-values>ignore</invalid-values>
                        <scalar-type>string</scalar-type>
                      </range-element-index>
                  }

                  { for $name in ("activeFragmentCount",
                                  "deletedFragmentCount", "documentCount", "fragmentCount",
                                  "databaseCount", "forestCount", "groupCount", "serverCount",
                                  "hostCount",
                                  "encryptedSize", "fastDataMaxSize", "diskSpace", "diskSize",
                                  "encryptedDiskSize",
                                  "freeSpace", "largeDataSize", "period", "port", "size",
                                  "standCount")
                    return
                      <range-element-index>
                        <collation/>
                        <localname>{$name}</localname>
                        <namespace-uri></namespace-uri>
                        <range-value-positions>false</range-value-positions>
                        <invalid-values>ignore</invalid-values>
                        <scalar-type>unsignedLong</scalar-type>
                      </range-element-index>
                  }

                  <!-- meters below here -->
                  <range-element-index>
                    <scalar-type>unsignedLong</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <collation></collation>
                    <localname>cluster-id</localname>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>unsignedLong</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/usage-meters</namespace-uri>
                    <collation></collation>
                    <localname>cluster-id</localname>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>unsignedLong</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>group-id</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>unsignedLong</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>host-id</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>unsignedLong</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>database-id</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>unsignedLong</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>forest-id</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>unsignedLong</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>server-id</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>dateTime</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>start-time</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>dateTime</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>end-time</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>dateTime</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>period-start-time</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>string</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters</namespace-uri>
                    <localname>period</localname>
                    <collation>http://marklogic.com/collation/</collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>dateTime</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters/label</namespace-uri>
                    <localname>start_dt</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>dateTime</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters/label</namespace-uri>
                    <localname>end_dt</localname>
                    <collation></collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                  <range-element-index>
                    <scalar-type>string</scalar-type>
                    <namespace-uri>http://marklogic.com/manage/meters/label</namespace-uri>
                    <localname>name</localname>
                    <collation>http://marklogic.com/collation/</collation>
                    <range-value-positions>true</range-value-positions>
                    <invalid-values>reject</invalid-values>
                  </range-element-index>
                </range-element-indexes>

  let $config := fold-left(
                   function ($config, $rei) {
                     let $idx := admin:database-range-element-index(
                                   $rei/manage:scalar-type,
                                   $rei/manage:namespace-uri,
                                   $rei/manage:localname,
                                   $rei/manage:collation,
                                   $rei/manage:range-value-positions,
                                   $rei/manage:invalid-values)
                     return
                       admin:database-add-range-element-index($config, $dbid, $idx)
                   }, $config, $reidx/manage:range-element-index)

  let $fields := <fields xmlns="http://marklogic.com/manage">
            <field>
              <field-name>servers-search</field-name>
              <field-path>
                <path>/opsdirClusterName</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/opsdirName</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/serverType</path>
                <weight>1.0</weight>
              </field-path>
            </field>
            <field>
              <field-name>clusters-search</field-name>
              <field-path>
                <path>/opsdirName</path>
                <weight>1.0</weight>
              </field-path>
            </field>
            <field>
              <field-name>databases-search</field-name>
              <field-path>
                <path>/opsdirClusterName</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/opsdirName</path>
                <weight>1.0</weight>
              </field-path>
            </field>
            <field>
              <field-name>forests-search</field-name>
              <field-path>
                <path>/opsdirName</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/hostNames</path>
                <weight>1.0</weight>
              </field-path>
            </field>
            <field>
              <field-name>hosts-search</field-name>
              <field-path>
                <path>/opsdirClusterName</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/opsdirName</path>
                <weight>1.0</weight>
              </field-path>
            </field>
            <field>
              <field-name>logs-search</field-name>
              <field-path>
                <path>/message</path>
                <weight>1.0</weight>
              </field-path>
            </field>
            <field>
              <field-name>tasks-search</field-name>
              <field-path>
                <path>/opsdirClusterName</path>
                <weight>1.0</weight>
              </field-path>
            </field>
            <field>
              <field-name>resource-groups-search</field-name>
              <field-path>
                <path>/name</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/size</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/description</path>
                <weight>1.0</weight>
              </field-path>
            </field>
            <field>
              <field-name>alerts-search</field-name>
              <field-path>
                <path>/resourceName</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/state</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/opsdirClusterName</path>
                <weight>1.0</weight>
              </field-path>
              <field-path>
                <path>/message</path>
                <weight>1.0</weight>
              </field-path>
            </field>

            <field>
              <field-name>rg-resources</field-name>
              <field-path>
                <path>/resource</path>
                <weight>1.0</weight>
              </field-path>
              <field-value-searches>true</field-value-searches>
            </field>
            <field>
              <field-name>rg-roles</field-name>
              <field-path>
                <path>/groupPermissions/role</path>
                <weight>1.0</weight>
              </field-path>
              <field-value-searches>true</field-value-searches>
            </field>
          </fields>

  let $oldFields := admin:database-get-fields($config, xdmp:database())
  let $oldNames  := $oldFields[string(*:field-name) != ""]/*:field-name/string()
  let $config    := admin:database-delete-field($config, $dbid, $oldNames)

  let $newFields := fold-left(
                      function ($fields, $field) {
                        let $paths := fold-left(
                                        function ($paths, $path) {
                                          let $fp := $path/manage:path/string()
                                          let $w  := xs:double($path/manage:weight)
                                          return
                                            ($paths, admin:database-field-path($fp,$w))
                                        }, (), $field/manage:field-path)
                        let $dbf := admin:database-path-field($field/manage:field-name, $paths)
                        return
                          ($fields, $dbf)
                      }, (), $fields/manage:field)

  let $config    := admin:database-add-field($config, $dbid, $newFields)

  (: Find the OpsDirectorSystem application :)
  let $appid := for $id in admin:get-appserver-ids($config)
                where admin:appserver-get-name($config, $id) = $odi:OPSDIRECTOR-UI-SERVER
                return $id
  let $modid := admin:appserver-get-modules-database($config, $appid)

  let $config := admin:appserver-set-authentication($config, $appid, "digest")

  let $amps  := <amps xmlns="http://marklogic.com/manage">
                  <amp>
                    <local-name>poll-clusters</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/polling</namespace>
                    <document-uri>/common/polling.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>poll-host</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/polling</namespace>
                    <document-uri>/common/polling.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>do-update-joins</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/joins</namespace>
                    <document-uri>/common/joins.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>do-update-joins-txn</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/joins</namespace>
                    <document-uri>/common/joins.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>add-managed-cluster</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/setup</namespace>
                    <document-uri>/common/setup-opsdir.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>remove-managed-cluster</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/setup</namespace>
                    <document-uri>/common/setup-opsdir.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>do-reset-collections-on-uris</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/setup</namespace>
                    <document-uri>/common/setup-opsdir.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>can-connect-to-managed-cluster</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/setup</namespace>
                    <document-uri>/common/setup-opsdir.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>get-role-names</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-user-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>role-exists</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-user-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>get-role-description</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-user-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>set-role-description</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-user-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>get-role-subroles</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-user-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>set-role-subroles</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-user-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>create-role</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>admin</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>remove-role</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>admin</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>add-roles-to-user</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>admin</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>remove-resources-from-resource-group</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>admin</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>insert-resource-group</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>admin</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>resource-scope-save</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>admin</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>resource-scope-create</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>admin</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>acknowledge-alert</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>reset-collections-on-uris</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/security</namespace>
                    <document-uri>/common/opsec.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>gzip-reply</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/jslib</namespace>
                    <document-uri>/common/jslib.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>server-authentication</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/jslib</namespace>
                    <document-uri>/common/jslib.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>roles-and-subroles</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/jslib</namespace>
                    <document-uri>/common/jslib.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>resource-metrics</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/jslib</namespace>
                    <document-uri>/common/jslib.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>resource-metrics-periods</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/jslib</namespace>
                    <document-uri>/common/jslib.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-data-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>resource-time-series</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/jslib</namespace>
                    <document-uri>/common/jslib.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-data-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>client-session-status</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/jslib</namespace>
                    <document-uri>/common/jslib.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>upgrade</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/upgrade</namespace>
                    <document-uri>/common/upgrade.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>get-logs</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/info</namespace>
                    <document-uri>/common/info.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>get-logs-text</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/info</namespace>
                    <document-uri>/common/info.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>amped-html</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/status</namespace>
                    <document-uri>/common/status.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>get-atomic-property</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/config</namespace>
                    <document-uri>/common/config.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                  <amp>
                    <local-name>update-joins</local-name>
                    <namespace>http://marklogic.com/v1/opsdirector/joins</namespace>
                    <document-uri>/common/joins.xqy</document-uri>
                    <modules-database/>
                    <roles>
                      <role>opsdir-admin-internal</role>
                    </roles>
                  </amp>
                </amps>
  (: No amps were removed between 1.0.0 and 1.1.0 :)
  let $_     := fold-left(function($acc, $amp) {
                  upgrade:amp-create($modid, $amp)
                }, (), $amps/manage:amp)

  (: Add get/set server field to opsdir-data-internal :)
  let $_     := upgrade:amend-opsdir-data-internal()


  return
    admin:save-configuration($config)
};

declare %private function upgrade:amp-create(
  $ampdb as xs:unsignedLong,
  $amp as element(manage:amp)
) as empty-sequence()
{
  let $id := try {
               xdmp:amp($amp/manage:namespace, $amp/manage:local-name,
                        $amp/manage:document-uri, $ampdb)
             } catch ($e) {
               xdmp:invoke-function(
                 function () {
                   sec:create-amp($amp/manage:namespace, $amp/manage:local-name,
                                  $amp/manage:document-uri, $ampdb,
                                  $amp/manage:roles/manage:role)
                 },
               <options xmlns="xdmp:eval">
                <database>{xdmp:security-database()}</database>
               </options>)
             }
  return
    ()
};

declare %private function upgrade:amend-opsdir-data-internal(
) as empty-sequence()
{
  xdmp:invoke-function(
      function () {
        sec:privilege-add-roles("http://marklogic.com/xdmp/privileges/xdmp-get-server-field",
                                "execute", "opsdir-data-internal"),
        sec:privilege-add-roles("http://marklogic.com/xdmp/privileges/xdmp-set-server-field",
                                "execute", "opsdir-data-internal")
      },
    <options xmlns="xdmp:eval">
     <database>{xdmp:security-database()}</database>
    </options>)
};
