xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace up="http://marklogic.com/v1/opsdirector/upload";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace la="http://marklogic.com/v1/opsdirector/log/alerts"
    at "/common/logalerts.xqy";

import module namespace j="http://marklogic.com/v1/opsdirector/joins"
    at "/common/joins.xqy";

import module namespace raw="http://marklogic.com/v1/opsdirector/rawdata"
    at "/common/rawdata.xqy";

import module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes"
    at "/common/rscopes.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare %private function up:response(
  $code as xs:integer,
  $message as xs:string
) as object-node()
{
  let $_ := xdmp:set-response-code($code, $message)
  return
    object-node { "code": $code, "message": $message }
};

declare function up:log(
  $session as object-node(),
  $cluster-id as xs:string,
  $host-id as xs:string,
  $date as xs:date,
  $stream-file as xs:string,
  $json as object-node()
)
{
  let $_ := trace("PUT " || count($json/logfile) || " log records", $t:trace-debug)

  (: N.B. When a cluster first joins, it's possible there will be a short
     period of time when the cluster and host names are not yet known.
     For those occasions, we use '?' as the value. :)
  let $cname := (raw:resource($cluster-id, "clusters", $cluster-id)/opsdirName, $cluster-id)[1]
  let $hname := (raw:resource($cluster-id, "hosts", $host-id)/opsdirName, $host-id)[1]

  for $msg at $index in $json/logfile
  let $data := $json/hostId || "/" || $msg/ts || "/" || $msg/message
  let $hash := xdmp:hash64($data)
  let $_    := trace("LOG " || $index || ": "
                     || string($msg/level) || ": " || string($msg/message),
                     $t:trace-debug-log-messages)
  let $uri := "/logs/" || $cluster-id || "/" || $host-id || "/" || $hash || ".json"
  let $log := object-node {
                "opsdirClusterId": string($json/clusterId),
                "opsdirClusterName": $cname,
                "opsdirHostId": string($json/hostId),
                "opsdirHostName": $hname,
                "hostTs": string($msg/ts),
                "ts": string(current-dateTime()),
                "level": string($msg/level),
                "uid": string($hash),
                "message": string($msg/message)
              }
  let $log := la:augment($log)
  let $_ := trace("Insert: " || $uri, $t:trace-debug-insert)
  let $_ := trace("Ins: " || $uri, $t:trace-debug-uri)

  let $p := let $r := t:managed-resource(string($json/clusterId),
                                         "hosts", string($json/hostId))
            return
              if (exists($r))
              then
                rs:effective-resource-permissions($r)
              else
                $t:DBPERMS
  return
    xdmp:document-insert($uri, $log, $p, ($t:COLLROOT||"logs")),

  up:response(200, "Ok")
};

declare function up:config(
  $session as object-node(),
  $cluster-id as xs:string,
  $host-id as xs:string,
  $date as xs:date,
  $stream-file as xs:string,
  $json as object-node()
)
{
  let $_     := trace("PUT 1 config timestamp record", $t:trace-debug)
  let $ts    := t:timestamp()
  let $uri   := "/config/timestamps/"
                || $cluster-id || "/" || $host-id || "/" || $ts || ".json"

  let $cur   := t:cluster-resource-timestamps($cluster-id, $host-id)
  let $json  := xdmp:to-json(
                  map:new(
                    (map:entry("ts", string(current-dateTime())),
                     if ($cur/opsdirDtstart)
                     then map:entry("opsdirDtstart", $cur/opsdirDtstart)
                     else map:entry("opsdirDtstart", string(current-dateTime())),
                     map:entry("opsdirDtend", "9999-12-31T23:59:59Z"),
                     map:entry("opsdirResourceType", "timestamps"),
                     map:entry("opsdirClusterId", concat(string($cluster-id))),
                     map:entry("opsdirHostId", concat(string($host-id))),
                     for $property in ($json/node() except ($json/clusterId | $json/hostId))
                     return
                       map:entry(local-name-from-QName(node-name($property)), $property)
                    )))/node()

  let $_   := raw:archive-resource($cur)
  let $_   := trace("Ins: " || $uri, $t:trace-debug-uri)
  return
    (xdmp:document-insert($uri, $json, $t:DBPERMS, ($t:COLLROOT||"config/timestamps")),
     (: Put into a different transaction so that we have better isolation; was
        getting deadlocks and this is a stab at a possible workaround. :)
     xdmp:invoke("/common/invoked/update-configs.xqy",
                 map:new((map:entry("json", $json),
                          map:entry("cluster-id", $cluster-id),
                          map:entry("host-id", $host-id))),
                 <options xmlns="xdmp:eval">
                   <isolation>different-transaction</isolation>
                   <transaction-mode>update-auto-commit</transaction-mode>
                 </options>),
     up:response(200, "Ok"))
};

declare function up:meters(
  $session as object-node(),
  $cluster-id as xs:string,
  $host-id as xs:string,
  $date as xs:date,
  $stream-file as xs:string,
  $doc as node()
)
{
  let $_ := trace("PUT 1 meters record", $t:trace-debug)
  let $xml  := $doc/*
  let $uri  := "/meters" || string($doc/@url)
  let $coll := ($t:COLLROOT || "meters", "meters") (: bare meters for b/w compat w/lib :)
  let $_    := trace("Insert: " || $uri, $t:trace-debug-insert)
  let $_    := trace("Ins: " || $uri, $t:trace-debug-uri)
  let $_    := xdmp:document-insert($uri, $xml, $t:DBPERMS, $coll)
  return
     up:response(200, "Ok")
};
