xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

(: Note: This module has been written so that it can be shared between
   XQuery and JavaScript. Consequently it uses camelCase naming conventions
   and passes ID values around as strings instead of xs:unsignedLongs.
   Because Javascript. Also, the resource scope XML is not in any namespace.
   Because JSON.
:)

module namespace t="http://marklogic.com/v1/opsdirector";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace ts="http://marklogic.com/v1/opsdirector/security"
    at "/common/opsec.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $t:CA-NAME := "OpsDirector-CA";
declare variable $t:CRED-NAME := "OpsDirector-Credential";

declare variable $t:DBPERMS :=
  (xdmp:permission("opsdir-admin", "read"),
   xdmp:permission("opsdir-admin", "update"),
   xdmp:permission("opsdir-data-internal", "read"),
   xdmp:permission("opsdir-data-internal", "update"));

declare variable $t:ADMIN-DBPERMS :=
  (xdmp:permission("opsdir-admin", "read"),
   xdmp:permission("opsdir-admin", "update"));

declare variable $t:COLLROOT := "http://marklogic.com/opsdirector/v1/";

declare variable $t:trace-status       := "opsdir-status";
declare variable $t:trace-upload       := "opsdir-upload";
declare variable $t:trace-debug        := "opsdir-debug";
declare variable $t:trace-debug-io     := "opsdir-debug-io";
declare variable $t:trace-debug-insert := "opsdir-insert";
declare variable $t:trace-debug-uri    := "opsdir-uri";
declare variable $t:trace-debug-perm   := "opsdir-debug-perm";
declare variable $t:trace-debug-priv   := "opsdir-debug-priv";
declare variable $t:trace-debug-role   := "opsdir-debug-role";
declare variable $t:trace-debug-rgrp   := "opsdir-debug-rgrp";
declare variable $t:trace-debug-http   := "opsdir-debug-http";
declare variable $t:trace-sessions     := "opsdir-sessions";
declare variable $t:trace-session-expires    := "opsdir-session-expires";
declare variable $t:trace-debug-scopes       := "opsdir-debug-scopes";
declare variable $t:trace-debug-log-messages := "opsdir-debug-log-messages";
declare variable $t:trace-debug-alerts       := "opsdir-debug-alerts";
declare variable $t:trace-debug-cluster-offline := "opsdir-debug-cluster-offline";
declare variable $t:trace-debug-configs := "opsdir-debug-configs";
declare variable $t:trace-disconnect   := "opsdir-disconnect";
declare variable $t:trace-polling      := "opsdir-polling";
declare variable $t:trace-pselect      := "opsdir-polling-selection";
declare variable $t:trace-poll-info    := "opsdir-polling-info";
declare variable $t:trace-expires      := "opsdir-expires";
declare variable $t:trace-joins        := "opsdir-joins";
declare variable $t:trace-debug-joins  := "opsdir-debug-joins";
declare variable $t:trace-debug-csv    := "opsdir-debug-csv";
declare variable $t:trace-search       := "opsdir-search";
declare variable $t:trace-app-endpoint := "opsdir-app-endpoint";

declare variable $t:trace-log-filter       := "opsdir-log-filter";
declare variable $t:trace-log-filter-debug := "opsdir-log-filter-debug";
declare variable $t:trace-alert-disable    := "opsdir-alert-disable";

declare variable $t:DEFAULT-RESOURCE-SCOPE := "default-resource-scope";
declare variable $t:RESOURCE-SCOPE         := "resource-scope";
declare variable $t:RESOURCE-GROUP         := "resource-group";
declare variable $t:RESOURCE               := "resource";
declare variable $t:RESOURCE-SCOPES-LIST   := ($t:DEFAULT-RESOURCE-SCOPE,
                                               $t:RESOURCE-SCOPE,
                                               $t:RESOURCE-GROUP);

declare variable $t:SHARED-SECRET
     := "It doesn't matter as long as it's shared with OpsDirector";

(: I've allowed this to become polluted with types of resource files in the db :)
declare variable $t:RESOURCE-TYPES
  := ("clusters", "groups", "hosts", "databases", "servers", "forests", "tasks",
      "counts", "roles", "privileges", "security",
      "host-status", "forest-status", "forest-storage", "database-status", "server-status","local-cluster-status",
      "resource-list");

declare variable $t:SERVER-RESOURCE-TYPES
  := ("clusters", "groups", "hosts", "databases", "servers", "forests");

declare %private variable $Z := xs:dayTimeDuration("PT0S");
declare %private variable $QOPTS
  := ("collation=http://marklogic.com/collation/");

(: ====================================================================== :)

declare function t:timestamp(
) as xs:string
{
  t:timestamp(current-dateTime())
};

declare function t:timestamp(
  $dt as xs:dateTime
) as xs:string
{
  let $stamp := adjust-dateTime-to-timezone($dt, $Z)
  return
    format-dateTime($stamp, "[Y0001][M01][D01]-[H01][m01][s01].[f000001]")
};

(: ====================================================================== :)

declare function t:archive-resource(
  $node as object-node()?
) as empty-sequence()
{
  if (empty($node))
  then
    ()
  else
    (: Must be called with a node that is a reference to the in-database node :)
    let $updated := t:replace-property($node, "opsdirDtend", string(current-dateTime()))
    let $_ := trace("Archiving " || xdmp:node-uri($node), $t:trace-debug)
    return
      (xdmp:node-replace($node, $updated),
       xdmp:document-set-collections(xdmp:node-uri($node), ()))
};

(: ====================================================================== :)

declare function t:add-opsdir-properties(
  $obj as object-node(),
  $rsrc as object-node()?,
  $type as xs:string,
  $id as xs:string?
) as object-node()
{
  t:add-opsdir-properties($obj, $rsrc, $type, $id, ())
};

declare function t:add-opsdir-properties(
  $obj as object-node(),
  $rsrc as object-node()?,
  $type as xs:string,
  $id as xs:string?,
  $other as map:map?
) as object-node()
{
  t:add-opsdir-properties($obj, $rsrc, $type, $id, $other, ())
};

declare function t:add-opsdir-properties(
  $obj as object-node(),
  $rsrc as object-node()?,
  $type as xs:string,
  $id as xs:string?,
  $other as map:map?,
  $exclude as node()*
) as object-node()
{
  (: Get the name property; beware that there are two possible forms depending
     on whether the JSON has been camel cased or not. :)
  let $name  := switch ($type)
                case "forests"    return ($obj/forestName, $obj/forest-name)[1]
                case "clusters"   return ($obj/clusterName, $obj/cluster-name)[1]
                case "databases"  return ($obj/databaseName, $obj/database-name)[1]
                case "groups"     return ($obj/groupName, $obj/group-name)[1]
                case "servers"    return ($obj/serverName, $obj/server-name)[1]
                case "hosts"      return ($obj/hostName, $obj/host-name)[1]
                case "roles"      return ($obj/roleName, $obj/role-name)[1]
                case "privileges" return ($obj/privilegeName, $obj/privilege-name)[1]
                case "tasks"      return $id
                case "host-status"     return ($obj/hostStatus/name,
                                               $obj/host-status/name)[1]
                case "database-status" return ($obj/databaseStatus/name,
                                               $obj/database-status/name)[1]
                case "forest-status"   return ($obj/forestStatus/name,
                                               $obj/forest-status/name)[1]
                case "server-status"   return ($obj/serverStatus/name,
                                               $obj/server-status/name)[1]
                case "counts"          return ($obj/databaseCounts/name,
                                               $obj/database-counts/name)[1]
                default return ()

  let $other := if (empty($other)) then map:map() else $other

  let $props := map:new(
                  ($other,

                   if (map:contains($other, 'ts'))
                   then ()
                   else map:entry("ts", string(current-dateTime())),

                   if (map:contains($other, 'opsdirName') or empty($name))
                   then ()
                   else map:entry("opsdirName", $name),

                   if (map:contains($other, 'opsdirResourceType'))
                   then ()
                   else map:entry("opsdirResourceType", $type),

                   if (map:contains($other, 'opsdirId') or empty($id))
                   then ()
                   else map:entry("opsdirId", $id),

                   if (map:contains($other, 'opsdirDtstart'))
                   then ()
                   else
                     if ($rsrc/opsdirDtstart)
                     then map:entry("opsdirDtstart", $rsrc/opsdirDtstart)
                     else map:entry("opsdirDtstart", string(current-dateTime())),

                   if (map:contains($other, 'opsdirDtend'))
                   then ()
                   else map:entry("opsdirDtend", "9999-12-31T23:59:59Z"),

                   for $property in $obj/node() except $exclude
                   let $lname := local-name-from-QName(node-name($property))
                   where not(map:contains($other, $lname))
                   return
                     map:entry($lname, $property)))
  return
    xdmp:to-json($props)/object-node()
};

(: ====================================================================== :)

declare function t:replace-property(
  $obj as object-node(),
  $property as xs:string,
  $value as item()*
) as object-node()
{
  xdmp:to-json(
    map:new(
      (map:entry($property, $value),
       for $prop in $obj/node()
       let $name := local-name-from-QName(node-name($prop))
       where $name ne $property
       return
         map:entry($name, $prop)
    ))
  )/object-node()
};

(: ====================================================================== :)

declare function t:managed-resource-ids(
  $resource-type as xs:string?
) as xs:string*
{
  t:managed-resource-ids((), $resource-type)
};

declare function t:managed-resource-ids(
  $cluster-id as xs:string?,
  $resource-type as xs:string?
) as xs:string*
{
  t:managed-resource-ids($cluster-id, $resource-type, ())
};

declare function t:managed-resource-ids(
  $cluster-id as xs:string?,
  $resource-type as xs:string?,
  $updated-since as xs:string?
) as xs:string*
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  if (empty($updated-since) or ($updated-since castable as xs:dateTime))
  then ()
  else error((), "OD-INVDATE", ("Invalid date: " || $updated-since)),

  let $type-query := if (empty($resource-type))
                     then ()
                     else cts:element-range-query(
                              xs:QName("opsdirResourceType"), "=", $resource-type, $QOPTS)
  let $dt-query   := cts:element-range-query(
                         xs:QName("opsdirDtend"), ">", current-dateTime())
  let $clus-query := if (empty($cluster-id))
                     then
                       ()
                     else
                       cts:element-range-query(
                           xs:QName("opsdirClusterId"), "=", $cluster-id, $QOPTS)
  let $coll-query := cts:collection-query($t:COLLROOT || "resource")
  let $time-query := if (empty($updated-since))
                     then ()
                     else cts:element-range-query(
                              xs:QName("ts"), ">", xs:dateTime($updated-since))
  return
    cts:element-values(
      xs:QName("opsdirId"), (), (),
      cts:and-query(($type-query, $dt-query, $clus-query, $coll-query, $time-query)))
};

(: ====================================================================== :)

declare function t:managed-resources(
  $ids as xs:string*
) as object-node()*
{
  t:managed-resources((), $ids)
};

declare function t:managed-resources(
  $resource-type as xs:string?,
  $ids as xs:string*
) as object-node()*
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  $ids ! t:managed-resource((), $resource-type, .)
};

declare function t:managed-resource(
  $cluster-id as xs:string?,
  $resource-type as xs:string?,
  $id as xs:string
) as object-node()?
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  let $type-query := if (empty($resource-type))
                     then
                       (: exclude stats, counts, etc. by default :)
                       cts:or-query((
                         for $type in $t:SERVER-RESOURCE-TYPES
                         return
                           cts:json-property-range-query(
                              "opsdirResourceType", "=", $type, $QOPTS)))
                     else
                       cts:json-property-range-query(
                           "opsdirResourceType", "=", $resource-type, $QOPTS)

  let $dt-query  := cts:json-property-range-query("opsdirDtend", ">", current-dateTime())

  let $c-query   := if (empty($cluster-id))
                    then
                      ()
                    else
                      cts:json-property-range-query(
                          "opsdirClusterId", "=", $cluster-id, $QOPTS)

  let $id-query  := cts:json-property-range-query("opsdirId", "=", $id, $QOPTS)

  let $doc       := cts:search(collection($t:COLLROOT || "resource"),
                               cts:and-query(($type-query, $dt-query, $c-query, $id-query)),
                               "unfiltered")
  let $_ := if (count($doc) gt 1)
            then xdmp:log(("Error: multiple resource documents for ('"
                          || $cluster-id || "','" || $resource-type || "','"
                          || $id || " (" || count($doc) || ")",
                          $doc ! xdmp:node-uri(.)))
            else ()
  return
    ($doc)[1]/node()
};

(: ====================================================================== :)

declare function t:managed-resource-id(
  $cluster-id as xs:string?,
  $resource-type as xs:string,
  $resource-name as xs:string
) as xs:string?
{
  if ($resource-type = $t:RESOURCE-TYPES)
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  (: Hack; this substring happens to work today, might not always :)
  let $s     := substring($resource-type, 1, string-length($resource-type) - 1)
  let $pname := fn:QName("", $s || "Name")

  for $id in t:managed-resource-ids($cluster-id, $resource-type)
  let $obj  := t:managed-resource($cluster-id, $resource-type, $id)
  let $name := $obj/node()[node-name(.) = $pname]/string()
  where $name = $resource-name
  return
    $obj/opsdirId
};

(: ====================================================================== :)

declare function t:resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string,
  $resource-id as xs:string
) as object-node()
{
  t:resource-scope($cluster, $resource-type, $resource-id, true())
};

declare function t:resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string,
  $resource-id as xs:string,
  $create as xs:boolean
) as object-node()?
{
  if (empty($cluster) or ($cluster = t:managed-resource-ids("clusters")))
  then ()
  else error((), "OD-NOCLUSTER", ("No such cluster: " || $cluster)),

  let $obj  := t:managed-resource($cluster, $resource-type, $resource-id)
  let $_    := if (empty($obj))
               then error((), "OD-NORESOURCE", ("Invalid scope, no such resource ("
                              || $resource-type || "/" || $resource-id || ")"))
               else ()

  let $cid  := string($obj/opsdirClusterId)
  let $rid  := string($obj/opsdirId)

  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:json-property-range-query("opsdirClusterId", "=", $cid, $QOPTS)
  let $id-query  := cts:json-property-range-query("opsdirId", "=", $rid, $QOPTS)
  let $k-query   := cts:json-property-value-query("opsdirKind", "resource")

  let $doc       := cts:search(collection($t:COLLROOT || "resource-scope"),
                               cts:and-query(($dt-query, $c-query, $id-query, $k-query)),
                               "unfiltered")

  let $_    := if (count($doc) gt 1)
               then error((), "OD-INTERR" ("Internal error: multiple resources for "
                              || "resource-scope(" || $cluster || ", "
                              || $resource-type || ", " || $resource-id || ")"))
               else ()
  return
    if (empty($doc) and $create)
    then
      object-node {
        "opsdirKind": $t:RESOURCE-SCOPE,
        "opsdirClusterId": $cid,
        "opsdirResourceType": $resource-type,
        "opsdirId": $rid,
        "opsdirDtstart": string(current-dateTime()),
        "ts": string(current-dateTime()),
        "opsdirDtend": "9999-12-31T23:59:59Z"
      }
    else
      $doc/object-node()
};

declare function t:default-resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string?
) as object-node()
{
  t:default-resource-scope($cluster, $resource-type, true())
};

declare %private function t:default-resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string?,
  $create as xs:boolean
) as object-node()?
{
  if (empty($resource-type) or ($resource-type = ("all", $t:RESOURCE-TYPES)))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  if (empty($cluster) or ($cluster = ("all", t:managed-resource-ids("clusters"))))
  then ()
  else error((), "OD-NOCLUSTER", ("No such cluster: " || $cluster)),

  let $cid   := if (empty($cluster)) then "all" else $cluster
  let $rtype := if (empty($resource-type)) then "all" else $resource-type

  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:element-range-query(
                        xs:QName("opsdirClusterId"), "=", $cid, $QOPTS)
  let $t-query   := cts:element-range-query(
                        xs:QName("opsdirResourceType"), "=", $rtype, $QOPTS)
  let $id-query  := cts:element-range-query(xs:QName("opsdirId"), "=", "all", $QOPTS)
  let $doc       := cts:search(collection($t:COLLROOT || "resource-scope"),
                               cts:and-query(($dt-query, $c-query, $t-query, $id-query)),
                               "unfiltered")

  let $_    := if (count($doc) gt 1)
               then error((), "OD-INTERR", ("Internal error: multiple resources for "
                              || "default-resource-scope(" || $cluster || ", "
                              || $resource-type || ")"))
               else ()
  return
    if (empty($doc) and $create)
    then
      object-node {
        "opsdirKind": $t:DEFAULT-RESOURCE-SCOPE,
        "opsdirClusterId": $cid,
        "opsdirResourceType": $resource-type,
        "opsdirId": "all",
        "opsdirDtstart": string(current-dateTime()),
        "ts": string(current-dateTime()),
        "opsdirDtend": "9999-12-31T23:59:59Z"
      }
    else
      $doc/object-node()
};

declare function t:resource-scope-resources(
  $scope as object-node()
) as object-node()*
{
  switch ($scope/opsdirKind)
  case $t:DEFAULT-RESOURCE-SCOPE
  return
    t:default-resource-resources($scope)
  case $t:RESOURCE-SCOPE
  return
    t:managed-resource($scope/opsdirClusterId, $scope/opsdirResourceType,
                       $scope/opsdirId)
  case $t:RESOURCE-GROUP
  return
    $scope/resource ! t:managed-resource((), $scope/opsdirResourceType, .)
  default
  return
    error((), "OD-BADSCOPE", ("Unexpected scope: " || local-name($scope)))
};

declare %private function t:default-resource-resources(
  $scope as object-node()
) as object-node()*
{
  let $cluster  := if ($scope/opsdirClusterId = "all")
                   then ()
                   else $scope/opsdirClusterId
  let $type     := if ($scope/opsdirResourceType = "all")
                   then ()
                   else $scope/opsdirResourceType

  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())

  let $c-query   := if ($cluster)
                    then
                      cts:element-range-query(
                          xs:QName("opsdirClusterId"), "=", $cluster, $QOPTS)
                    else
                      ()

  let $t-query   := if ($type)
                    then
                      cts:element-range-query(
                          xs:QName("opsdirResourceType"), "=", $type, $QOPTS)
                    else
                      ()
  return
     cts:search(collection($t:COLLROOT || "resource"),
                cts:and-query(($dt-query, $c-query, $t-query)),
                "unfiltered")/node()
};

(: ====================================================================== :)

declare %private function t:assert-roles(
  $roles as xs:string*
) as xs:string*
{
  for $role in $roles
  return
    try {
       string(xdmp:role($role))
    } catch ($e) {
      if ($role castable as xs:unsignedLong)
      then
        if (empty(ts:get-role-names($role cast as xs:unsignedLong)))
        then error((), "OD-BADROLE", ("Bad role: " || $role))
        else $role
      else
        error((), "OD-BADROLE", ("Bad role: " || $role))
   }
};

declare function t:assert-scope(
  $scope as object-node()
) as object-node()
{
  t:assert-scope($scope, $t:RESOURCE-SCOPES-LIST)
};

declare function t:assert-scope(
  $scope as object-node(),
  $scopes as xs:string+
) as object-node()
{
  if ($scope/opsdirKind = $scopes)
  then
    $scope
  else
    error((), "OD-BADSCOPE", ("Unexpected scope: " || $scope/opsdirKind))
};

(: ====================================================================== :)

declare function t:resource-add-permissions(
  $scope as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $roles  := t:assert-roles($roles)
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node { "role": string($role), "capability": "read" },
                  for $perm in $scope/permissions
                  where not($perm/role = $roles)
                  return
                    $perm)
  return
    t:replace-property(t:assert-scope($scope), "permissions", array-node { $update })
};

declare function t:resource-remove-permissions(
  $scope as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $roles  := t:assert-roles($roles)
  let $update := (for $perm in $scope/permissions
                  where not($perm/role = $roles)
                  return
                    $perm)
  return
    t:replace-property(t:assert-scope($scope), "permissions", array-node { $update })
};

declare function t:resource-set-permissions(
  $scope as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $roles  := t:assert-roles($roles)
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node { "role": $role, "capability": "read" })
  return
    t:replace-property(t:assert-scope($scope), "permissions", array-node { $update })
};

(: ====================================================================== :)

declare function t:resource-add-privileges(
  $scope as object-node(),
  $roles as xs:string*,
  $actions as xs:string*
) as object-node()
{
  let $roles  := t:assert-roles($roles)
  let $update := (for $priv in $scope/privileges
                  return
                    object-node {
                       "role": $priv/role,
                       "actions": array-node {
                                     if ($priv/role = $roles)
                                     then
                                       distinct-values(($priv/actions, $actions))
                                     else
                                       $priv/actions
                                  }
                    },
                  for $role in distinct-values($roles)
                  where empty($scope/privileges[role = $role])
                  return
                    object-node {
                       "role": $role,
                       "actions": array-node { distinct-values($actions) }
                    })
  return
    t:replace-property(t:assert-scope($scope), "privileges", array-node { $update })
};

declare function t:resource-remove-privileges(
  $scope as object-node(),
  $roles as xs:string*,
  $actions as xs:string*
) as object-node()
{
  let $roles  := t:assert-roles($roles)
  let $update := (for $priv in $scope/privileges
                  return
                    object-node {
                       "role": $priv/role,
                       "actions": array-node {
                                     if ($priv/role = $roles)
                                     then
                                       for $action in $priv/actions
                                       where not($action = $actions)
                                       return
                                         $action
                                     else
                                       $priv/actions
                                  }
                    })
  let $update := for $priv in $update
                 where exists($priv/actions)
                 return
                   $priv
  return
    t:replace-property(t:assert-scope($scope), "privileges", array-node { $update })
};

declare function t:resource-set-privileges(
  $scope as object-node(),
  $roles as xs:string*,
  $actions as xs:string*
) as object-node()
{
  let $roles  := t:assert-roles($roles)
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node {
                       "role": $role,
                       "actions": array-node { distinct-values($actions) }
                    })
  let $update := for $priv in $update
                 where exists($priv/actions)
                 return
                   $priv
  return
    t:replace-property(t:assert-scope($scope), "privileges", array-node { $update })
};

(: ====================================================================== :)

declare function t:resource-add-roles(
  $scope as object-node(),
  $roles as xs:string*,
  $add-roles as xs:string*
) as object-node()
{
  let $roles     := t:assert-roles($roles)
  let $add-roles := t:assert-roles($add-roles)

  let $update := (for $robj in $scope/roles
                  return
                    object-node {
                       "role": $robj/role,
                       "addRoles": array-node {
                                     if ($robj/role = $roles)
                                     then
                                       distinct-values(($robj/addRoles, $add-roles))
                                     else
                                       $robj/addRoles
                                  }
                    },
                  for $role in distinct-values($roles)
                  where empty($scope/roles[role = $role])
                  return
                    object-node {
                       "role": $role,
                       "addRoles": array-node { distinct-values($add-roles) }
                    })
  return
    t:replace-property(t:assert-scope($scope), "roles", array-node { $update })
};

declare function t:resource-remove-roles(
  $scope as object-node(),
  $roles as xs:string*,
  $add-roles as xs:string*
) as object-node()
{
  let $roles     := t:assert-roles($roles)
  let $add-roles := t:assert-roles($add-roles)

  let $update := (for $robj in $scope/roles
                  return
                    object-node {
                       "role": $robj/role,
                       "addRoles": array-node {
                                     if ($robj/role = $roles)
                                     then
                                       for $role in $robj/addRoles
                                       where not($role = $add-roles)
                                       return
                                         $role
                                     else
                                       $robj/addRoles
                                  }
                    })
  let $update := for $robj in $update
                 where exists($robj/addRoles)
                 return
                   $robj
  return
    t:replace-property(t:assert-scope($scope), "roles", array-node { $update })
};

declare function t:resource-set-roles(
  $scope as object-node(),
  $roles as xs:string*,
  $add-roles as xs:string*
) as object-node()
{
  let $roles     := t:assert-roles($roles)
  let $add-roles := t:assert-roles($add-roles)

  let $update := (for $role in distinct-values($roles)
                  return
                    object-node {
                       "role": $role,
                       "addRoles": array-node { distinct-values($add-roles) }
                    })
  let $update := for $robj in $update
                 where exists($robj/addRoles)
                 return
                   $robj
  return
    t:replace-property(t:assert-scope($scope), "roles", array-node { $update })
};

(: ====================================================================== :)

declare function t:resource-groups(
) as object-node()*
{
  t:resource-groups((), ())
};

declare function t:resource-groups(
  $resource-type as xs:string?
) as object-node()*
{
  t:resource-groups($resource-type, ())
};

declare function t:resource-groups(
  $resource-type as xs:string?,
  $search-text as xs:string?
) as object-node()*
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  let $type-query := if (empty($resource-type))
                     then
                       ()
                     else
                       cts:element-range-query(
                         xs:QName("opsdirResourceType"), "=", $resource-type, $QOPTS)
  let $dt-query   := cts:element-range-query(
                       xs:QName("opsdirDtend"), ">", current-dateTime())
  let $text-query := if (exists($search-text))
                     then
                       cts:word-query(normalize-space($search-text)||"*",
                                      ('case-insensitive', 'punctuation-insensitive'))
                     else
                       ()

  return for $group in cts:search(collection($t:COLLROOT || "resource-group"),
               cts:and-query(($type-query, $dt-query, $text-query)),
               "unfiltered")
         order by $group/*:name
         return $group/object-node()
};

declare function t:resource-groups-containing-resources(
  $resource-type as xs:string,
  $ids as xs:string+
) as object-node()*
{
  let $rg-ids := for $id in $ids
                 return
                   t:resource-groups($resource-type)[$id = resource]/opsdirId
  return
    distinct-values($rg-ids) ! t:resource-group(.)
};

(: ====================================================================== :)

declare function t:applicable-scopes(
  $resource as object-node()
) as object-node()*
{
  let $cluster  := $resource/opsdirClusterId
  let $type     := $resource/opsdirResourceType
  let $id       := $resource/opsdirId

  (: Applicable defaults:
     all resources on all clusters
     {$type} resources on all clusters
     all resources on this cluster
     {$type} resources on this cluster
  :)
  let $all_all   := t:default-resource-scope((), (), false())
  let $all_type  := t:default-resource-scope((), $type, false())

  (: The cluster may not exist yet, so don't worry if these fail. :)
  let $clus_all  := try { t:default-resource-scope($cluster, (), false())    }
                    catch ($e) { () }
  let $clus_type := try { t:default-resource-scope($cluster, $type, false()) }
                    catch ($e) { () }

  (: This resource may not exist yet, so don't worry if this fails. :)
  let $this_rsrc := try { t:resource-scope((), $type, $id) }
                    catch ($e) { () }

  let $groups    := if (exists($id))
                    then t:resource-groups-containing-resources($type, $id)
                    else ()
  return
    ($all_all, $all_type, $clus_all, $clus_type, $this_rsrc, $groups)
};

(: ====================================================================== :)

declare function t:effective-resource-permissions(
  $resource as object-node()
) as element()*
{
  let $scopes := t:applicable-scopes($resource)
  return
    t:effective-resource-permissions($resource, $scopes)
};

declare function t:effective-resource-permissions(
  $resource as object-node(),
  $scopes as object-node()*
) as element(sec:permission)+
{
  let $cluster := $resource/opsdirClusterId
  let $type    := $resource/opsdirResourceType
  let $id      := $resource/opsdirId
  let $scopes  := $scopes ! t:assert-scope(.)

  let $dscopes := $scopes[opsdirKind = $t:DEFAULT-RESOURCE-SCOPE
                          and (opsdirClusterId = "all" or opsdirClusterId = $cluster)
                          and opsdirResourceType = $type]
  let $rscopes := $scopes[opsdirKind = $t:RESOURCE-SCOPE
                          and (opsdirClusterId = "all" or opsdirClusterId = $cluster)
                          and opsdirResourceType = $type
                          and opsdirId = $id]
  let $gscopes := $scopes[opsdirKind = $t:RESOURCE-GROUP
                          and opsdirResourceType = $type
                          and resource = $id]

  let $roles   := ($dscopes/permissions/role ! xs:unsignedLong(.),
                   $rscopes/permissions/role ! xs:unsignedLong(.),
                   $gscopes/permissions/role ! xs:unsignedLong(.),
                   $gscopes/groupPermissions/role ! xs:unsignedLong(.))

  let $roles := distinct-values($roles)

  return
    ($t:DBPERMS,
     for $roleid in $roles
     let $role := ts:get-role-names($roleid)
     let $gcap := $gscopes/groupPermissions[role=string($roleid)]/capability
     let $cap  := "read"

     let $_   := trace("Resource permission: "
                       || $role || " can " || string-join($cap,",") || " "
                       || $type || " " || $id || " on cluster "
                       || $cluster, $t:trace-debug-perm)
     return
       xdmp:permission($role, $cap))
};

declare function t:effective-resource-privileges(
  $resource as object-node(),
  $scopes as object-node()*
) as xs:string*
{
  t:effective-resource-privileges($resource, $scopes,
                                  xdmp:user-roles(xdmp:get-current-user()) ! string(.))
};

declare function t:effective-resource-privileges(
  $resource as object-node(),
  $scopes as object-node()*,
  $roles as xs:string*
) as xs:string*
{
  let $scopes  := $scopes ! t:assert-scope(.)
  let $roles   := t:assert-roles($roles)

  let $cluster := $resource/opsdirClusterId
  let $type    := $resource/opsdirResourceType
  let $id      := $resource/opsdirId
  let $scopes  := t:effective-resource-scopes($resource, $scopes)

  let $dscopes := $scopes[opsdirKind = $t:DEFAULT-RESOURCE-SCOPE
                          and (opsdirClusterId = "all" or opsdirClusterId = $cluster)
                          and opsdirResourceType = $type]
  let $rscopes := $scopes[opsdirKind = $t:RESOURCE-SCOPE
                          and (opsdirClusterId = "all" or opsdirClusterId = $cluster)
                          and opsdirResourceType = $type
                          and opsdirId = $id]
  let $gscopes := $scopes[opsdirKind = $t:RESOURCE-GROUP
                          and opsdirResourceType = $type
                          and resource = $id]

  (:
  let $_ := xdmp:log(("D", $dscopes, "R", $rscopes, "G", $gscopes, $roles))
  let $_ := xdmp:log($rscopes/privileges[role = $roles])
  :)

  return
    distinct-values(($dscopes/privileges[role = $roles]/actions,
                     $rscopes/privileges[role = $roles]/actions,
                     $gscopes/privileges[role = $roles]/actions))
};

(: ====================================================================== :)

declare function t:effective-resource-roles(
  $resource as object-node(),
  $scopes as object-node()*
) as xs:string*
{
  t:effective-resource-privileges($resource, $scopes,
                                  xdmp:user-roles(xdmp:get-current-user()) ! string(.))
};

declare function t:effective-resource-roles(
  $resource as object-node(),
  $scopes as object-node()*,
  $roles as xs:string*
) as xs:string*
{
  let $scopes  := $scopes ! t:assert-scope(.)
  let $roles   := t:assert-roles($roles)

  let $cluster := $resource/opsdirClusterId
  let $type    := $resource/opsdirResourceType
  let $id      := $resource/opsdirId
  let $scopes  := t:effective-resource-scopes($resource, $scopes)

  let $dscopes := $scopes[opsdirKind = $t:DEFAULT-RESOURCE-SCOPE
                          and (opsdirClusterId = "all" or opsdirClusterId = $cluster)
                          and opsdirResourceType = $type]
  let $rscopes := $scopes[opsdirKind = $t:RESOURCE-SCOPE
                          and (opsdirClusterId = "all" or opsdirClusterId = $cluster)
                          and opsdirResourceType = $type
                          and opsdirId = $id]
  let $gscopes := $scopes[opsdirKind = $t:RESOURCE-GROUP
                          and opsdirResourceType = $type
                          and resource = $id]

  (:
  let $_ := xdmp:log(("D", $dscopes, "R", $rscopes, "G", $gscopes, $roles))
  let $_ := xdmp:log($rscopes/privileges[role = $roles])
  :)

  return
    distinct-values(($dscopes/roles[role = $roles]/addRoles,
                     $rscopes/roles[role = $roles]/addRoles,
                     $gscopes/roles[role = $roles]/addRoles))
};

(: ====================================================================== :)

declare function t:create-resource-group-role(
  $name as xs:string
) as xs:string
{
  let $secret := "d72e95f66c03b31d855e7546854f2882afd339eb"
  let $seed   := xdmp:get-current-user() || "/" || $name
  let $hash   := xdmp:hmac-sha1($secret, $seed)

  (: If the role exists, so will the document. :)
  let $hash   := if (ts:role-exists("opsdir-rg-internal-" || $hash))
                 then
                   let $_ := trace("Hash collision for " || $seed, $t:trace-status)
                   return
                     string(xdmp:random())
                 else $hash
  return
    "opsdir-rg-internal-" || $hash
};

(: ====================================================================== :)

declare function t:resource-group-create(
  $name as xs:string,
  $description as xs:string,
  $resource-type as xs:string
) as object-node()
{
  (: N.B. There's no option to create a resource group with an initial
     set of resources because (1) the app doesn't need that today and
     (2) the existing resource documents wouldn't get their permissions
     updated. Fixing the latter doesn't seem worth the effort right now,
     given the former :)

  if (normalize-space($name) = "")
  then error((), "OD-BADNAME", ("Name must not be empty"))
  else (),

  let $resources := t:assert-resource-type((), $resource-type)

  let $role   := t:create-resource-group-role($name)
  let $roleid := ts:create-role($role)
  let $_      := xdmp:sleep(1000) (: Give the first transactions a chance to commit :)
  let $_      := ts:add-roles-to-user($role)
  let $_      := xdmp:sleep(1000) (: Give the second transaction a chance to commit :)

  let $id     := string(xdmp:random())
  let $uri    := "/resource-group/" || $id || "/" || t:timestamp() || ".json"

  let $group  := object-node {
                   "opsdirKind": $t:RESOURCE-GROUP,
                   "opsdirResourceType": $resource-type,
                   "opsdirId": $id,
                   "opsdirDtstart": string(current-dateTime()),
                   "opsdirDtend": "9999-12-31T23:59:59Z",
                   "ts": string(current-dateTime()),
                   "name": $name,
                   "groupPermissions": array-node {
                      (object-node {
                        "role": string($roleid),
                        "capability": "update"
                       })
                   },
                   "description": $description,
                   "size": 0,
                   "resource": array-node { () }
                 }
  return
    (
     (: This next bit has to be amped and has to be in a diff txn :)
     ts:insert-resource-group($uri, $role, $group),
     $group
    )
};

declare function t:resource-group-delete(
  $id as xs:string
) as empty-sequence()
{
  let $rg := t:resource-group($id)
  where exists($rg)
  return
    let $roleid  := $rg/relatedRole
    let $role    := ts:get-role-names($roleid)

    (: Remove all roles and resources; then save it to update the permissions :)
    let $newrg := t:resource-group-set-group-permissions($rg, ())
    let $newrg := t:resource-group-set-resources($newrg, ())
    let $newrg := t:replace-property($newrg, "opsdirDtend", string(current-dateTime()))
    let $newrg := t:replace-property($newrg, "DELETE", "true")
    return
      (t:resource-scopes-save($newrg),
       ts:remove-role($role))
};

declare function t:resource-group(
  $id as xs:string
) as object-node()?
{
  let $id-query  := cts:element-range-query(xs:QName("opsdirId"), "=", $id, $QOPTS)
  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  return
    cts:search(collection($t:COLLROOT || "resource-group"),
               cts:and-query(($id-query, $dt-query)), ("unfiltered"))/object-node()
};

declare function t:resource-group-set-name(
  $group as object-node(),
  $name as xs:string
) as object-node()
{
  t:replace-property(t:assert-scope($group, $t:RESOURCE-GROUP), "name", $name)
};

declare function t:resource-group-set-description(
  $group as object-node(),
  $description as xs:string
) as object-node()
{
  let $group := t:assert-scope($group, $t:RESOURCE-GROUP)
  return
    t:replace-property($group, "description", $description)
};

declare function t:resource-group-add-group-permissions(
  $group as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $group  := t:assert-scope($group, $t:RESOURCE-GROUP)
  let $roles  := t:assert-roles($roles)
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node { "role": $role, "capability": "read" },
                  for $perm in $group/groupPermissions
                  where not($perm/role = $roles)
                  return
                    $perm)
  return
    t:replace-property($group, "groupPermissions", array-node { $update })
};

declare function t:resource-group-remove-group-permissions(
  $group as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $group  := t:assert-scope($group, $t:RESOURCE-GROUP)
  let $roles  := t:assert-roles($roles)
  let $update := (for $perm in $group/groupPermissions
                  where not($perm/role = $roles)
                  return
                    $perm)
  return
    t:replace-property($group, "groupPermissions", array-node { $update })
};

declare function t:resource-group-set-group-permissions(
  $group as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $group  := t:assert-scope($group, $t:RESOURCE-GROUP)
  let $roles  := t:assert-roles($roles)
  let $keep   := $group/groupPermissions[capability = "update"]
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node { "role": $role, "capability": "read" })
  return
    t:replace-property($group, "groupPermissions", array-node { ($keep, $update) })
};

declare function t:resource-group-add-resources(
  $group as object-node(),
  $rsrcs as xs:string*
) as object-node()
{
  (: FIXME: check if rsrcs are valid :)
  let $group   := t:assert-scope($group, $t:RESOURCE-GROUP)
  let $current := $group/resource
  let $update  := distinct-values(($current, $rsrcs))
  let $obj     := t:replace-property($group, "resource", array-node { $update })
  return
    t:replace-property($obj, "size", count($update))
};

declare function t:resource-group-remove-resources(
  $group as object-node(),
  $rsrcs as xs:string*
) as object-node()
{
  let $group  := t:assert-scope($group, $t:RESOURCE-GROUP)
  let $update := for $rsrc in $group/resource
                 where not($rsrc = $rsrcs)
                 return
                   $rsrc
  let $obj     := t:replace-property($group, "resource", array-node { $update })
  return
    t:replace-property($obj, "size", count($update))
};

declare function t:resource-group-set-resources(
  $group as object-node(),
  $rsrcs as xs:string*
) as object-node()
{
  let $group  := t:assert-scope($group, $t:RESOURCE-GROUP)
  let $update := distinct-values($rsrcs)
  let $obj    := t:replace-property($group, "resource", array-node { $update })
  return
    t:replace-property($obj, "size", count($update))
};

declare function t:effective-resource-scopes(
  $rsrc as object-node()?,
  $scopes as object-node()*
) as object-node()+
{
  (: Step 0: If there are any resource groups in $scopes, then
     all of the other resource groups that also contain the
     same resource(s) have to be considered as well...
  :)
  let $groups := $scopes[opsdirKind = "resource-group"]
  let $others := $scopes except $groups

  let $gids   := $groups/opsdirId
  let $allids := for $scope in $groups
                 for $id in $scope/resource
                 return
                   t:resource-groups($scope/opsdirResourceType)[$id = resource]/opsdirId
  let $addin  := for $id in distinct-values($allids)
                 where not($id = $gids)
                 return
                   t:resource-group($id)
  let $scopes := ($others, $groups, $addin)

  (: Step 1: What resource scopes need to be considered? :)
  let $rscopes := if (empty($rsrc))
                  then $scopes ! t:resource-scope-closure(.)
                  else $rsrc ! t:resource-scope-closure(.)

  (: Step 2: Remove duplicates favoring passed in scopes! :)
  (:         (is there a faster way?) :)
  let $map   := map:map()
  let $_     := for $scope in ($rscopes, $scopes)
                return
                map:put($map,
                        $scope/opsdirClusterId
                        ||$scope/opsdirResourceType
                        ||$scope/opsdirId,
                        $scope)
  let $rscopes := map:keys($map) ! map:get($map, .)
  return
    $rscopes
};

declare function t:resource-scopes-save(
  $scopes as object-node()+
) as empty-sequence()
{
  let $_       := trace(("Resource scopes save:", $scopes), $t:trace-debug-scopes)
  let $rscopes := t:effective-resource-scopes((), $scopes)

  let $_       := trace(("Effective scopes:", $rscopes), $t:trace-debug-scopes)

  (: Find all the effected resources; avoid dups! :)
  let $map   := map:map()
  let $rsrcs := for $scope in $rscopes
                for $rsrc in t:resource-scope-resources($scope)
                return
                  map:put($map, xdmp:node-uri($rsrc), $rsrc)

  (: What about resources that have been removed from a scope? :)
  (: I think this can only happen with resource groups... :)
  let $rmvd  := for $scope in $scopes
                where $scope/opsdirKind = "resource-group"
                return
                  let $oscope := t:resource-group($scope/opsdirId)
                  for $id in $oscope/resource
                  where not($id = $scope/resource)
                  return
                    let $rsrc := t:managed-resources($scope/opsdirResourceType, $id)
                    return
                      map:put($map, xdmp:node-uri($rsrc), $rsrc)

  let $rsrcs := map:keys($map) ! map:get($map, .)

  let $_     := trace(("Resources:", count($rsrcs), $rsrcs/opsdirId), $t:trace-debug-scopes)

  (: For each resource, update the permissions, if necessary :)
  let $taskmap := map:map()
  let $_     := for $rsrc in $rsrcs
                let $newperm := t:effective-resource-permissions($rsrc, $rscopes)

                (: All of the URIs (stats, counts, etc.) related to this resource
                   must have the same permissions as the underlying resource. :)
                let $all-resources :=
                       (cts:search(
                          collection(($t:COLLROOT||"resource", $t:COLLROOT||"join")),
                             cts:json-property-value-query("opsdirId",
                                                           $rsrc/opsdirId))
                           ! xdmp:node-uri(.),
                        if ($rsrc/opsdirResourceType = "databases")
                        then t:find-forest-resources($rsrc)
                        else ())

                (: Try to update open alerts about the resource. :)
                let $all-alerts :=
                       cts:search(collection($t:COLLROOT||"alerts"),
                                  cts:and-query(
                                     (cts:json-property-value-query("resourceId",
                                                                    $rsrc/opsdirId),
                                      cts:json-property-value-query("status", "open"))))
                          ! xdmp:node-uri(.)

                let $_ := if ($rsrc/opsdirResourceType = "hosts")
                          then t:update-task-permissions($taskmap, $rsrc, $rscopes, $newperm)
                          else ()

                for $uri in ($all-resources, $all-alerts)
                let $curperm := xdmp:document-get-permissions($uri)
                return
                  if (t:same-permissions($curperm, $newperm))
                  then
                    let $_ := trace(("Same permissions: " || $uri), $t:trace-debug)
                    return ()
                  else
                    let $_ := trace(("Updating permissions: " || $uri), $t:trace-debug)
                    let $_ := trace(("Current:", t:show-perms($curperm)), $t:trace-debug-scopes)
                    let $_ := trace(("New:", t:show-perms($newperm)), $t:trace-debug-scopes)
                    return
                      xdmp:document-set-permissions($uri, $newperm)

  let $_     := for $uri in map:keys($taskmap)
                return
                  xdmp:document-set-permissions($uri, map:get($taskmap, $uri))

  (: For each resource and resource-group, save it if it's been changed :)
  return
    for $scope in $scopes
    where empty(xdmp:node-uri($scope))
    return
      let $_ := trace(("Saving resource scope:", $scope), $t:trace-debug-scopes)
      return
        ts:resource-scope-save($scope)
};

declare %private function t:find-forest-resources(
  $database as object-node()
) as xs:string*
{
  let $names := $database/forest ! string(.)
  where exists($names)
  return
    let $forests := cts:search(collection($t:COLLROOT||"resource"),
                                  cts:and-query(
                                     (cts:json-property-value-query("opsdirClusterId",
                                                                    $database/opsdirClusterId),
                                      cts:json-property-value-query("opsdirResourceType",
                                                                    "forests"),
                                      cts:json-property-value-query("opsdirName", $names))))
    let $fids    := $forests/opsdirId
    return
      cts:search(collection(($t:COLLROOT||"resource",$t:COLLROOT||"join")),
                   cts:json-property-value-query("opsdirId", $fids))
        ! xdmp:node-uri(.)
};

declare %private function t:update-task-permissions(
  $taskmap as map:map,
  $host as object-node(),
  $rscopes as object-node()*,
  $newperm as element()*
) as map:map
{
  let $cid      := $host/opsdirClusterId
  let $gname    := $host/opsdirGroupName
  let $newperm  := t:compute-task-permissions($host, $rscopes, $newperm)

  let $gid   := t:group-id-for-host($host)
  let $qtype := cts:json-property-value-query("opsdirResourceType", "tasks")
  let $qclus := cts:json-property-value-query("opsdirClusterId", $cid)
  let $qgrp  := cts:json-property-value-query("opsdirGroupId", $gid)
  let $tasks := cts:search(collection($t:COLLROOT||"resource"),
                           cts:and-query(($qtype, $qclus, $qgrp))) ! xdmp:node-uri(.)
  let $_     := for $uri in $tasks
                return
                  let $curperm := xdmp:document-get-permissions($uri)
                  return
                    if (t:same-permissions($curperm, $newperm))
                    then
                      let $_ := trace(("Same permissions: " || $uri), $t:trace-debug)
                      return ()
                    else
                      let $_ := trace(("Updating permissions: " || $uri),
                                      $t:trace-debug)
                      let $_ := trace(("Current:", t:show-perms($curperm)),
                                      $t:trace-debug-scopes)
                      let $_ := trace(("New:", t:show-perms($newperm)),
                                      $t:trace-debug-scopes)
                      return
                        if (map:contains($taskmap, $uri))
                        then
                          map:put($taskmap, $uri, (map:get($taskmap, $uri), $newperm))
                        else
                          map:put($taskmap, $uri, $newperm)
  return
    $taskmap
};

declare function t:compute-task-permissions(
  $host as object-node(),
  $rscopes as object-node()*,
  $newperm as element()*
)
{
  let $cid      := $host/opsdirClusterId
  let $gname    := $host/opsdirGroupName
  let $qtype    := cts:json-property-value-query("opsdirResourceType", "hosts")
  let $qclus    := cts:json-property-value-query("opsdirClusterId", $cid)
  let $qgrp     := cts:json-property-value-query("opsdirGroupName", $gname)
  let $hosts    := cts:search(collection($t:COLLROOT||"resource"),
                              cts:and-query(($qtype, $qclus, $qgrp)))/object-node()

  let $hostids  := $hosts/opsdirId ! string(.)
  let $qkind    := cts:json-property-value-query("opsdirKind", "resource-group")
  let $qhosts   := cts:json-property-value-query("resource", $hostids)
  let $groups   := cts:search(collection($t:COLLROOT||"resource-group"),
                              cts:and-query(($qkind, $qhosts)))/object-node()

  (: Swap in any that are currently being updated :)
  let $groups   := for $group in $groups
                   let $id := $group/opsdirId
                   return
                     if ($id = $rscopes/opsdirId)
                     then
                       $rscopes[$id = opsdirId]
                     else
                       $group

  let $allscopes := fold-left(
                      function ($scopes, $host) {
                        t:effective-resource-scopes($host, $scopes)
                      }, $groups, $hosts)

  (: Now drop the ones that no longer contain this host :)
  let $allscopes := $allscopes[resource = $hostids]

  let $merged   := for $curscope in $allscopes
                   let $scope := if ($rscopes[opsdirId = $curscope/opsdirId])
                                 then $rscopes[opsdirId = $curscope/opsdirId]
                                 else $curscope
                   let $perms := $scope/groupPermissions
                                   ! xdmp:permission(xs:unsignedLong(./role), ./capability)
                   (:let $_ := xdmp:log(($scope/opsdirId, t:show-perms($perms))):)
                   return
                     $perms

  (: Remove duplicates so that t:same-permissions works :)
  let $map       := map:map()
  let $_         := for $perm in ($newperm, $merged)
                    let $key := $perm/sec:role-id || "," || $perm/sec:capability
                    return
                      map:put($map, $key, $perm)
  return
    map:keys($map) ! map:get($map, .)
};

declare function t:group-id-for-host(
  $host as object-node()
)
{
  let $cid   := $host/opsdirClusterId
  let $name  := $host/opsdirGroupName
  let $qtype := cts:json-property-value-query("opsdirResourceType", "groups")
  let $qname := cts:json-property-value-query("opsdirName", $name)
  let $qclus := cts:json-property-value-query("opsdirClusterId", $cid)
  let $group := cts:search(collection($t:COLLROOT||"resource"),
                           cts:and-query(($qtype, $qname, $qclus)))[1]/object-node()
  return
    $group/opsdirId
};

declare %private function t:show-perms(
  $perms as element(sec:permission)*
) as xs:string*
{
  $perms ! t:show-perm(.)
};

declare %private function t:show-perm(
  $perm as element(sec:permission)
) as xs:string
{
  let $role := ts:get-role-names($perm/sec:role-id)
  return
    "Permission: " || $role || " " || $perm/sec:capability
};

(: ====================================================================== :)

declare %private function t:resource-scope-closure(
  $scope as object-node()
) as object-node()*
{
  switch ($scope/opsdirKind)
  case $t:RESOURCE-GROUP
  return
    (: Check the version in the db because resources may have been removed :)
    (t:resource-group-closure($scope),
     (: ??? t:resource-group-closure(t:resource-group($scope/opsdirId)), :)
     t:resource-scope-resource-scopes($scope))
  case $t:RESOURCE-SCOPE
  return
    ($scope,
     t:resource-scope-resource-scopes($scope))
  case $t:DEFAULT-RESOURCE-SCOPE
  return
    (t:resource-default-closure($scope),
     t:resource-scope-resource-scopes($scope))
  case $t:RESOURCE  (: handle the special case where what we pass in is a resource :)
  return
    let $scope := t:resource-scope($scope/opsdirClusterId, $scope/opsdirResourceType,
                                   $scope/opsdirId)
    return
      t:resource-scope-closure($scope)
  default return
    error((), "OD-BADSCOPE", ("Unexpected scope: " || $scope/opsdirKind))
};

declare %private function t:resource-scope-resource-scopes(
  $scope as object-node()
) as object-node()*
{
  let $clusters := if ($scope/opsdirKind = $t:RESOURCE-GROUP)
                   then
                     if ($scope/opsdirResourceType = "clusters")
                     then $scope/resource
                     else t:managed-resource-ids("clusters")
                   else
                     if ($scope/opsdirClusterId = "all")
                     then t:managed-resource-ids("clusters")
                     else $scope/opsdirClusterId/string()
  let $types    := if ($scope/opsdirResourceType = "all")
                   then $t:RESOURCE-TYPES
                   else $scope/opsdirResourceType/string()

  return
    (
     (: The default for all resources on all clusters always applies :)
     t:default-resource-scope((), (), false()),

     (: The default for all resources of this type on all clusters always applies :)
     t:default-resource-scope((), $scope/opsdirResourceType, false()),

     (: The default for all resources on this clusters always applies :)
     t:default-resource-scope($scope/opsdirClusterId, (), false()),

     (: The default for all resources of this tpye on this clusters always applies :)
     t:default-resource-scope($scope/opsdirClusterId, $scope/opsdirResourceType, false()),

     (: Any resource scopes covered by this scope apply :)
     if ($scope/opsdirKind = $t:RESOURCE-GROUP)
     then
       $scope/resource ! t:resource-scope((), $scope/opsdirResourceType, ., false())
     else
       if ($scope/opsdirId = "all")
       then
         for $cluster in $clusters
         for $type in $types
         for $id in t:managed-resource-ids($cluster, $type)
         return
           t:resource-scope($cluster, $type, $id, false())
       else
         t:resource-scope($scope/opsdirClusterId, $scope/opsdirResourceType,
                          $scope/opsdirId, false())
      )
};

declare (:%private:) function t:resource-group-closure(
  $scope as object-node()
) as object-node()*
{
  for $id in $scope/resource
  return
    t:resource-scope((), $scope/opsdirResourceType, $id, false())
};

declare %private function t:resource-default-closure(
  $scope as object-node()
) as object-node()*
{
  let $scope    := t:assert-scope($scope, $t:DEFAULT-RESOURCE-SCOPE)
  let $clusters := if ($scope/opsdirClusterId = "all")
                   then t:managed-resource-ids("clusters")
                   else $scope/opsdirClusterId
  let $types    := if ($scope/opsdirResourceType = "all")
                   then $t:RESOURCE-TYPES
                   else $scope/opsdirResourceType

  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:or-query(
                      for $cid in $clusters
                      return
                        cts:element-range-query(
                            xs:QName("opsdirClusterId"), "=", $cid, $QOPTS))
  let $t-query   := cts:or-query(
                      for $tid in $types
                      return
                        cts:element-range-query(
                            xs:QName("opsdirResourceType"), "=", $tid, $QOPTS))
  return
     cts:search(collection($t:COLLROOT || "resource-scope"),
                cts:and-query(($dt-query, $c-query, $t-query)),
                "unfiltered")/object-node()
};

(: ====================================================================== :)

declare %private function t:resources-in-scope(
  $scope as object-node()
) as object-node()*
{
  switch ($scope/opsdirKind)
  case $t:RESOURCE-GROUP
  return
    (: Check the version in the db because resources may have been removed :)
    (t:resource-group-resources-in-scope($scope),
     t:resource-group-resources-in-scope(t:resource-group($scope/opsdirId)))
  case $t:RESOURCE-SCOPE
  return
    t:resource-resources-in-scope($scope)
  case $t:DEFAULT-RESOURCE-SCOPE
  return
    t:resource-resources-in-scope($scope)
  default return
    error((), "OD-BADSCOPE", ("Unexpected scope: " || $scope/opsdirKind))
};

declare %private function t:resource-resources-in-scope(
  $scope as object-node()
) as object-node()*
{
  if ($scope/opsdirKind = $t:DEFAULT-RESOURCE-SCOPE)
  then
    t:default-resource-resources($scope)
  else
    let $coll := $t:COLLROOT || "config/" || $scope/opsdirClusterId
                 || "/" || $scope/opsdirResourceType || "/" || $scope/opsdirId
    let $doc  := collection($coll)
    let $_    := if (count($doc) ne 1)
                 then error((), "OD-BADRSRC", ("Resource doesn't exactly identify a single config"))
                 else ()
    return
      $scope
};

declare %private function t:resource-group-resources-in-scope(
    $scope as element(resourceGroup)
) as element(resourceScope)*
{
  for $id in $scope/resource
  let $rsrc := t:managed-resource((), $scope/opsdirResourceType, xs:unsignedLong($id))
  let $_    := if (count($rsrc) ne 1)
               then error((), "OD-BADRSRC", ("Resource doesn't exactly identify a single config"))
               else ()
  return
    t:resource-scope((), $scope/opsdirResourceType, $id)
};

declare %private function t:resource-get-config-uri(
    $scope as element(resourceScope)
) as xs:string
{
  let $coll := $t:COLLROOT || "config/" || $scope/opsdirClusterId
               || "/" || $scope/opsdirResourceType || "/" || $scope/opsdirId
  let $doc  := collection($coll)
  let $_    := if (count($doc) ne 1)
               then error((), "OD-BADRSRC", ("Resource doesn't exactly identify a single config: " || $coll))
               else ()
  return
    xdmp:node-uri($doc)
};

declare %private function t:same-permissions(
  $curperm as element(sec:permission)*,
  $newperm as element(sec:permission)*
) as xs:boolean
{
  let $a := for $perm in $curperm
            return
              if ($newperm[sec:capability = $perm/sec:capability
                           and sec:role-id = $perm/sec:role-id])
              then ()
              else ("notsame")
  let $b := for $perm in $newperm
            return
              if ($curperm[sec:capability = $perm/sec:capability
                           and sec:role-id = $perm/sec:role-id])
              then ()
              else ("notsame")
  return
    empty($a) and empty($b)
};

(: ====================================================================== :)

declare function t:cluster-resource-timestamps(
  $cluster-id as xs:string,
  $host-id as xs:string
) as object-node()?
{
  let $t-query   := cts:element-range-query(
                        xs:QName("opsdirResourceType"), "=", "timestamps", $QOPTS)
  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:element-range-query(
                        xs:QName("opsdirClusterId"), "=", $cluster-id, $QOPTS)
  let $h-query   := cts:element-range-query(
                        xs:QName("opsdirHostId"), "=", $host-id, $QOPTS)
  let $doc       := cts:search(collection(),
                               cts:and-query(($t-query, $dt-query, $c-query, $h-query)),
                               ("unfiltered",
                                cts:index-order(
                                  cts:element-reference(fn:QName("","ts")), "descending")))

  let $_ := if (count($doc) gt 1)
            then xdmp:log(("Multiple timestamps:", $doc ! xdmp:node-uri(.)))
            else ()

  return
    ($doc/node())[1]
};

(: ====================================================================== :)

declare %private function t:assert-resource-type(
  $resources as xs:string*,
  $type as xs:string
) as xs:string*
{
  let $_ := if ($type = $SERVER-RESOURCE-TYPES)
            then
              ()
            else
              error((), "OD-BADRTYPE", ("Bad resource type: " || $type))
  return
    if (empty($resources))
    then
      ()
    else
      let $ids       := t:managed-resource-ids((), $type)
      let $resources := for $resource in $resources
                        return
                          if ($resource = $ids)
                          then $resource
                          else
                            let $_ := trace($resource || " is not a " || $type || ": "
                                            || string-join($ids ! string(.), ", "),
                                            $t:trace-debug-rgrp)
                            return
                              error((), "OD-BADRSRC", ("Bad resource; not a " || $type || ": " || $resource))
      return
        $resources
};

(: ====================================================================== :)

declare function t:poll-cluster(
  $type as xs:string
) as xs:string?
{
  let $clusters := t:managed-resource-ids("clusters")
  let $hosts    := for $host in xdmp:group-hosts()
                   let $key := "suppress-polling-from-" || $host
                   let $value := config:get-atomic-property($key, false())
                   where string($value) ne "true"
                   return
                     $host
  where exists($clusters)
  return
    t:poll-cluster($type, current-dateTime(), $clusters, $hosts, xdmp:host())
};

declare function t:poll-cluster(
  $type as xs:string,
  $stamp as xs:dateTime,
  $clusters as xs:string+,
  $hosts-in-group as xs:unsignedLong+,
  $this-host as xs:unsignedLong
) as xs:string?
{
  (: ========================================================================

   This function determines what cluster should be polled next (for
   alerts, for example, or status information). This function is used
   by scheduled tasks. We assume that this function is called at the
   same time (to a granularity of 1 minute) on all hosts in the group.
   We assume all hosts are in the same timezone so TZ doesn't matter.

   It has the following goals:

   1. Every host in the group should poll a different cluster:
      * In this iteration
        - This assures that different tasks are never trying to update
          the same documents.
      * On each iteration
        - This assures that if one host is down, a particular cluster
          doesn't get starved for updates.
   2. We make this calculation without maintaining state in the
      database. This means that we can work out what to do without
      acquiring a lock on a document in the database.

   Solution: Take the list of clusters in numerical order by cluster id.
   Use the number of minutes past midnight to locate a starting place
   in the list. Distribute the clusters across the hosts.

   But be careful, if we just blindly iterated through in a strictly
   round-robin fashion, we could end up always polling the same
   clusters with the same hosts. Consider the simple case of three
   clusters and three hosts: you'd get the cycle 1,2,3; 1,2,3; etc.
   That would mean that if the first host was down, cluster 1 would
   never get polled. To combat this, after each complete cycle, we
   move forward. So we start the first cycle with the first host,
   the second cyle with the second, etc.

   For example:

   Consider the case of ten clusters (numbered 11-20) and three hosts
   (numbered 111-113). Iterations occur every minute:

    1: 111 polls 11, 112 polls 12, 113 polls 13
    2: 111 polls 14, 112 polls 15, 113 polls 16
    3: 111 polls 17, 112 polls 18, 113 polls 19
    4: 111 polls 20

   At this point, we wrap around.

    4:               112 polls 11, 113 polls 12

   Now we've done a full cycle, so we start over at 12:

    5: 111 polls 12, 112 polls 13, 113 polls 14
    6: 111 polls 15, 112 polls 16, 113 polls 17
    7: 111 polls 18, 112 polls 19, 113 polls 20
    8: 111 polls 11, 112 polls 12, 113 polls 13

   Now we've done a full cycle, so we start over at 13:

    9: 111 polls 13, 112 polls 14, 113 polls 15
   10: 111 polls 16, 112 polls 17, 113 polls 18
   11: 111 polls 19, 112 polls 20, 113 polls 11
   12: 111 polls 12, 112 polls 13, 113 polls 14

   Etc.

   You can see that there's a little bit of overlap between
   the end of one cycle and the beginning of the next. This little
   "ripple" rolls across the whole set over time, but doesn't seem
   significant enough to worry about.

   ======================================================================== :)

  let $minute := (hours-from-dateTime($stamp) * 60) + minutes-from-dateTime($stamp)

  let $_      := trace("Polling at " || $stamp || " (" || $minute || ") on host "
                       || xdmp:host-name($this-host) || " for " || $type,
                       $t:trace-pselect)
  let $_      := trace(count($clusters) || " clusters, "
                       || count($hosts-in-group) || " hosts for " || $type,
                       $t:trace-pselect)

  (: We need them in a stable order; could stash this in a server field
     or something, but do we really think the time it takes to sort a
     modest number of atomics is worth the effort for a task that runs
     no more frequently than once a minute? Nah. :)
  let $clusters       := for $cluster in $clusters order by $cluster return $cluster
  let $hosts-in-group := for $host in $hosts-in-group order by $host return $host

  let $ordinal        := for $host at $index in $hosts-in-group
                         where $host = $this-host
                         return $index

  (: We only have 1440 minutes. What if there are more than 1440 clusters?
     Then do them in chunks. :)
  let $clusters       := if (count($clusters) gt 1440)
                         then
                           let $partitions := (count($clusters) idiv 1440) + 1
                           let $partition  := $minute mod $partitions
                           return
                             subsequence($clusters, ($partition*1440)+1, 1440)
                         else
                           $clusters

  let $cluster-count  := count($clusters)
  let $host-count     := count($hosts-in-group)

  let $_              := trace($cluster-count || " clusters: "
                               || string-join($clusters, ", ") || " for " || $type,
                               $t:trace-pselect)
  let $_              := trace($host-count || " hosts: "
                               || string-join($hosts-in-group ! string(.), ", ")
                               || " for " || $type,
                               $t:trace-pselect)
  let $_              := trace("Ordinal: " || $ordinal || " for " || $type,
                               $t:trace-pselect)

  (: Divide the clusters into partitions by number of hosts. Pick a partition :)
  let $partitions     := ($cluster-count + $host-count - 1) idiv $host-count
  let $partition      := $minute mod $partitions

  (: Work out how many interations we have to go through before we've done
     an entire cycle. :)
  let $iterations     := if ($cluster-count mod $host-count = 0)
                         then $cluster-count idiv $host-count
                         else ($cluster-count idiv $host-count) + 1

  (: Work out the offset for each iteration :)
  let $offset         := if ($host-count = 1)
                         then 0
                         else ($minute idiv $iterations) mod $iterations

  (: Pick one of the clusters! :)
  let $pick           := ($partition * $host-count) + $ordinal + $offset
  let $pick           := if ($pick gt $cluster-count)
                         then $pick - $cluster-count
                         else $pick

  let $_              := trace("Partition: " || $partition
                               || ", iterations: " || $iterations
                               || ", offset: " || $offset
                               || ", pick: " || $pick
                               || " for " || $type, $t:trace-pselect)

  return
    if ($ordinal gt $cluster-count)
    then
      ()
    else
      let $_ := trace("Picked cluster: " || $clusters[$pick] || " for " || $type,
                      $t:trace-pselect)
      return
        $clusters[$pick]
};

(: ====================================================================== :)
(: N.B. These are common, they're called from info and disconn, but the
   aren't amped so they're not likely to work from random places.
:)

declare function t:open-local-alert(
  $cid as xs:string,
  $hid as xs:string,
  $code as xs:string,
  $state as xs:string,
  $resourceType as xs:string,
  $resourceId as xs:string,
  $resourceName as xs:string,
  $message as xs:string
) as empty-sequence()
{
  (: Is there an existing open alert with this $cid and $code?
     There should never be more than one, but ... :)
  let $colls := ("http://marklogic.com/opsdirector/v1/alerts/" || $cid,
                 "http://marklogic.com/opsdirector/v1/alerts")
  let $qclus := cts:json-property-value-query("opsdirClusterId", $cid)
  let $qrsrc := cts:json-property-value-query("opsdirId", $resourceId)
  let $qstat := cts:json-property-value-query("status", "open")
  let $alert := cts:search(collection($colls[1]),
                           cts:and-query(($qclus, $qrsrc, $qstat)))/object-node()

  let $alert := $alert[code = $code]
  return
    if (empty($alert))
    then
      (: We're going to generate a log message; let's make sure it has the
         same permissions as the relevant resource... :)
      let $rsrc  := t:managed-resource($cid, $resourceType||"s", $resourceId)
      let $perms := if (exists($rsrc))
                    then xdmp:document-get-permissions(xdmp:node-uri($rsrc))
                    else
                      (xdmp:log("Generating local alert for non-existant resource: "
                                || $resourceId || " (" || $resourceType
                                || " on " || $cid || ")"),
                       $t:DBPERMS)
      let $cluster := t:managed-resource($cid, "clusters", $cid)
      let $host    := t:managed-resource($cid, "hosts", $hid)
      (: Generate a log message :)
      let $_       := xdmp:log($code || ": " || $host/*:opsdirName || ": " || $message)
      let $uid     := string(xdmp:random())
      let $alert   := object-node {
        "opsdirClusterId": $cid,
        "opsdirClusterName": $cluster/*:opsdirName,
        "opsdirHostId": $hid,
        "duration": "PT0S",
        "resourceType": $resourceType,
        "opsdirDtstart": string(current-dateTime()),
        "code": $code,
        "message": $message,
        "state": $state,
        "local": true(),
        "resourceName": $resourceName,
        "status": "open",
        "id": string(xdmp:random()),
        "uid": $uid,
        "ts": string(current-dateTime()),
        "acknowledged": "no",
        "opsdirId": $resourceId,
        "resourceId": $resourceId }
      return
        xdmp:document-insert("/alerts/" || $cid || "/" || $uid || ".json",
                             $alert, $perms, $colls)
    else
      let $uri := xdmp:node-uri($alert)
      (: Oh for the love of ... this is a hack :)
      let $map := map:new(for $key in $alert/* except ($alert/duration)
                          return map:entry(
                                    local-name-from-QName(node-name($key)),
                                    $alert/*[node-name(.) = node-name($key)]))
      let $duration := current-dateTime() - xs:dateTime($alert/dtstart)
      let $obj      := object-node { "duration": string($duration) }
      let $newalert := xdmp:to-json($map + $obj)/node()
      return
        xdmp:node-replace($alert, $newalert)
};

declare function t:close-local-alert(
  $cid as xs:string,
  $resourceId as xs:string,
  $code as xs:string
) as empty-sequence()
{
  let $colls := ("http://marklogic.com/opsdirector/v1/alerts/" || $cid,
                 "http://marklogic.com/opsdirector/v1/alerts")
  let $qclus := cts:json-property-value-query("opsdirClusterId", $cid)
  let $qrsrc := cts:json-property-value-query("opsdirId", $resourceId)
  let $qstat := cts:json-property-value-query("status", "open")
  let $alert := cts:search(collection($colls[1]),
                           cts:and-query(($qclus, $qrsrc, $qstat)))/object-node()
  for $alert in $alert[code = $code]
  return
    (: Oh for the love of ... this is a hack :)
    let $map := map:new(for $key in $alert/*
                        except ($alert/duration|$alert/opsdirDtend|$alert/status)
                        return map:entry(local-name-from-QName(node-name($key)),
                                         $alert/*[node-name(.) = node-name($key)]))
    let $duration := current-dateTime() - xs:dateTime($alert/dtstart)
    let $obj      := object-node { "duration": string($duration),
                                   "opsdirDtend": string(current-dateTime()),
                                   "status": "closed" }
    let $newalert := xdmp:to-json($map + $obj)/node()
    return
      xdmp:node-replace($alert, $newalert)
};
