xquery version "1.0-ml";

(: Copyright 2018 MarkLogic Corporation.  All Rights Reserved. :)

import module namespace setup="http://marklogic.com/v1/opsdirector/setup"
    at "/common/setup-opsdir.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $secret := "It doesn't matter as long as it's shared with OpsDirector";
declare variable $reqip  := xdmp:get-request-client-address();

declare function local:setup(
  $cluster-id as xs:string,
  $hostname as xs:string,
  $port as xs:string,
  $credname as xs:string
)
{
  setup:add-managed-cluster($hostname, $port, $cluster-id, $credname)
};

declare function local:connect(
  $body as element()?
)
{
  let $hash := xdmp:hmac-sha256($secret, $body/hostname||$body/cluster-id, "hex")
  let $bad  := if (empty($body))
               then true()
               else
                 if (not($body/hostname) or not($body/hash) or not($body/cluster-id))
                 then
                   (xdmp:log(("Warning: " || $reqip || " sent invalid payload:", $body)),
                    true())
                 else
                   if (string($body/hash) eq $hash)
                   then
                     false()
                   else
                     (xdmp:log(("Warning: " || $reqip || " sent incorrect hash",
                                "Expected: " || $hash || " received:", $body)),
                      true())
  return
    if ($bad)
    then
      (xdmp:set-response-content-type("text/plain"),
       xdmp:set-response-code(400, "Malformed request"),
       "Malformed request")
    else
      local:setup($body/cluster-id, $body/hostname, $body/management-api-port,
                  $body/credential-name)
};

let $body := try {
               xdmp:get-request-body("xml")/*
             } catch ($e) {
               xdmp:log("Warning: " || $reqip || " sent unparsable payload")
             }
return
  if ($body/self::hello)
  then
    <charmed/>
  else
    local:connect($body)
