xquery version "1.0-ml";

(: Copyright 2018 MarkLogic Corporation.  All Rights Reserved. :)

import module namespace setup="http://marklogic.com/v1/opsdirector/setup"
    at "/common/setup-opsdir.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $reqip  := xdmp:get-request-client-address();

declare function local:disconnect(
  $cluster-id as xs:string,
  $hostname as xs:string
)
{
  let $_ := xdmp:log("Unmanaging " || $cluster-id || " from " || $hostname)
  return
    setup:remove-managed-cluster($hostname, $cluster-id)
};

let $body := try {
               xdmp:get-request-body("xml")/*
             } catch ($e) {
               xdmp:log("Warning: " || $reqip || " sent unparsable payload")
             }
let $hash := xdmp:hmac-sha256($t:SHARED-SECRET, $body/hostname||$body/cluster-id, "hex")
let $bad  := if (empty($body))
             then true()
             else
               if (not($body/hostname) or not($body/hash) or not($body/cluster-id))
               then
                 (xdmp:log(("Warning: " || $reqip || " sent invalid payload:", $body)),
                  true())
               else
                 if (string($body/hash) eq $hash)
                 then
                   false()
                 else
                   (xdmp:log(("Warning: " || $reqip || " sent incorrect hash",
                              "Expected: " || $hash || " received:", $body)),
                    true())
return
  if ($bad)
  then
    (xdmp:set-response-content-type("text/plain"),
     xdmp:set-response-code(400, "Malformed request"),
     "Malformed request")
  else
    local:disconnect($body/cluster-id, $body/hostname)
