xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

declare default function namespace "http://www.w3.org/2005/xpath-functions";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace ts="http://marklogic.com/v1/opsdirector/security"
    at "/common/opsec.xqy";

import module namespace upload="http://marklogic.com/v1/opsdirector/upload"
    at "/common/upload.xqy";

import module namespace sess="http://marklogic.com/v1/opsdirector/session"
    at "/common/session.xqy";

declare option xdmp:mapping "false";

declare variable $threshold := xs:dayTimeDuration("PT10M");

declare function local:err(
  $code as xs:integer,
  $message as xs:string
) as xs:string
{
  (xdmp:set-response-code($code, $message),
   xdmp:set-response-content-type("text/plain"),
   $message)
};

ts:endpoint-authentication(false()),
let $contentType  := xdmp:get-request-header("content-type")
let $method       := xdmp:get-request-method()

let $cluster-id   := xdmp:get-request-field("cluster-id")
let $host-id      := xdmp:get-request-field("host-id")
let $type         := xdmp:get-request-field("stream-type")
let $year         := xdmp:get-request-field("year")
let $month        := xdmp:get-request-field("month")
let $day          := xdmp:get-request-field("day")
let $file         := xdmp:get-request-field("stream-file")

let $date         := $year || "-" || $month || "-" || $day

let $session      := sess:get($cluster-id, $host-id)

let $_ := trace("Upload: " || $cluster-id || "/" || $host-id || ": " || $type,
                $t:trace-upload)

let $_ := trace("Content-type: " || $contentType, $t:trace-debug-http)
let $_ := trace("Method:       " || $method, $t:trace-debug-http)
let $_ := trace("Cluster-id:   " || $cluster-id, $t:trace-debug-http)
let $_ := trace("Host-id:      " || $host-id, $t:trace-debug-http)
let $_ := trace("Date:         " || $date, $t:trace-debug-http)
let $_ := trace("Type:         " || $type, $t:trace-debug-http)
let $_ := trace("File:         " || $file, $t:trace-debug-http)

return
  if (empty($session) or sess:expired($session))
  then
    let $_ := if (empty($session))
              then
                trace("Refusing upload: no session for "|| $cluster-id || "/" || $host-id,
                      $t:trace-status)
              else
                trace("Refusing upload: expired session for "|| $cluster-id || "/" || $host-id,
                      $t:trace-status)
    return
      local:err(403, "Renegotiate session")
  else
    if ($method = "PUT" and ($date castable as xs:date)
        and ($type ne '') and ($file ne ''))
    then
      let $remain  := sess:remaining-time($session)
      let $_       := if ($remain lt $threshold)
                      then trace("Session " || $session/sessionId
                                 || " expires in " || $remain, $t:trace-session-expires)
                      else ()
      let $cid     := string($cluster-id)
      let $hid     := string($host-id)
      let $date    := $date cast as xs:date
      let $payload := if ($type = "meters")
                      then xdmp:get-request-body("xml")/node()
                      else xdmp:get-request-body("json")/node()
      let $_       := if (empty($payload))
                      then trace("Failed to parse payload: " || xdmp:get-request-body("text"),
                                 $t:trace-status)
                      else trace(("Uploaded payload:", $payload), $t:trace-debug-http)
      return
        switch ($type)
        case "errorLog" return upload:log($session, $cid,$hid,$date,$file,$payload)
        case "config"   return upload:config($session, $cid,$hid,$date,$file,$payload)
        case "meters"   return upload:meters($session, $cid,$hid,$date,$file,$payload)
        default         return local:err(400, "Bad request: unexpected type: " || $type)
    else
      let $_ := trace("Bad request  (incorrect method, date, type, or file)", $t:trace-status)
      return
        local:err(400, "Bad request")
